@@ -1,7 +1,7 @@
 /*
  * Copyright (C) 2016-2017 Intel, Ltd.
  * Copyright (C) 2016-2021 Canonical, Ltd.
- * Copyright (C) 2021-2022 Colin Ian King.
+ * Copyright (C) 2021-2023 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -20,35 +20,25 @@
  */
 #include "stress-ng.h"
 #include "core-arch.h"
+#include "core-cache.h"
 
 #if defined(HAVE_SYS_AUXV_H)
 #include <sys/auxv.h>
 #endif
 
-#if defined(__linux__)
-
 typedef struct {
 	const char	*name;			/* cache type name */
 	const stress_cache_type_t value;	/* cache type ID */
 } stress_generic_map_t;
 
 typedef enum {
-	CACHE_SIZE,
-	CACHE_LINE_SIZE,
-	CACHE_WAYS
+	STRESS_CACHE_SIZE,
+	STRESS_CACHE_LINE_SIZE,
+	STRESS_CACHE_WAYS
 } cache_size_type_t;
 
-#if defined(__linux__)
 #define SYS_CPU_PREFIX               "/sys/devices/system/cpu"
 #define SYS_CPU_CACHE_DIR            "cache"
-#endif
-
-static const stress_generic_map_t cache_type_map[] = {
-	{ "data",		CACHE_TYPE_DATA },
-	{ "instruction",	CACHE_TYPE_INSTRUCTION },
-	{ "unified",		CACHE_TYPE_UNIFIED },
-	{  NULL,		CACHE_TYPE_UNKNOWN }
-};
 
 /*
  * stress_cache_get_cpu()
@@ -61,6 +51,8 @@ static inline unsigned int stress_cache_get_cpu(const stress_cpus_t *cpus)
 	return (cpu >= cpus->count) ? 0 : cpu;
 }
 
+#if defined(__linux__) ||	\
+    defined(__APPLE__)
 /*
  * stress_get_string_from_file()
  * 	read data from file into a fixed size buffer
@@ -85,7 +77,437 @@ static int stress_get_string_from_file(
 
 	return 0;
 }
+#endif
 
+/*
+ * stress_get_cache_by_cpu()
+ * @cpu: cpu to consider.
+ * @cache_level: numeric cache level (1-indexed).
+ * Obtain the cpu cache indexed by @cache_level.
+ *
+ * POTENTIAL BUG: assumes only 1 data cache per CPU cache level.
+ *
+ * Returns: stress_cpu_cache_t, or NULL on error.
+ */
+static stress_cpu_cache_t * stress_get_cache_by_cpu(const stress_cpu_t *cpu, const int cache_level)
+{
+	uint32_t  i;
+
+	if (!cpu || !cache_level)
+		return NULL;
+
+	for (i = 0; i < cpu->cache_count; i++) {
+		stress_cpu_cache_t *p = &cpu->caches[i];
+
+		if (p->level != cache_level)
+			continue;
+
+		/* we want a data cache */
+		if (p->type != CACHE_TYPE_INSTRUCTION)
+			return p;
+	}
+	return NULL;
+}
+
+/*
+ * stress_get_max_cache_level()
+ * @cpus: array of cpus to query.
+ * Determine the maximum cache level available on the system.
+ *
+ * Returns: 1-index value denoting highest cache level, or 0 on error.
+ */
+uint16_t stress_get_max_cache_level(const stress_cpus_t *cpus)
+{
+	stress_cpu_t    *cpu;
+	uint32_t  i;
+	uint16_t  max = 0;
+
+	if (!cpus) {
+		pr_dbg("%s: invalid cpus parameter\n", __func__);
+		return 0;
+	}
+
+	cpu = &cpus->cpus[stress_cache_get_cpu(cpus)];
+
+	for (i = 0; i < cpu->cache_count; i++) {
+		const stress_cpu_cache_t *cache = &cpu->caches[i];
+
+		max = cache->level > max ? cache->level : max;
+	}
+
+	return max;
+}
+
+/*
+ * stress_get_cpu_cache()
+ * @cpus: array of cpus to query.
+ * @cache_level: numeric cache level (1-indexed).
+ * Obtain a cpu cache of level @cache_level.
+ *
+ * Returns: stress_cpu_cache_t pointer, or NULL on error.
+ */
+stress_cpu_cache_t *stress_get_cpu_cache(const stress_cpus_t *cpus, const uint16_t cache_level)
+{
+	stress_cpu_t *cpu;
+
+	if (!cpus) {
+		pr_dbg("%s: invalid cpus parameter\n", __func__);
+		return NULL;
+	}
+
+	if (!cache_level) {
+		pr_dbg("%s: invalid cache_level: %d\n",
+			__func__, cache_level);
+		return NULL;
+	}
+
+	cpu = &cpus->cpus[stress_cache_get_cpu(cpus)];
+
+	return stress_get_cache_by_cpu(cpu, cache_level);
+}
+
+#if defined(__linux__) &&	\
+    defined(STRESS_ARCH_SPARC)
+static int stress_get_cpu_cache_value(
+	const char *cpu_path,
+	const char *file,
+	uint64_t *value)
+{
+	char path[PATH_MAX];
+	char tmp[128];
+
+	(void)stress_mk_filename(path, sizeof(path), cpu_path, file);
+	if (stress_get_string_from_file(path, tmp, sizeof(tmp)) == 0) {
+		if (sscanf(tmp, "%" SCNu64, value) == 1)
+			return 0;
+	}
+	return -1;
+}
+#endif
+
+#if defined(__linux__) &&	\
+    defined(STRESS_ARCH_ALPHA)
+/*
+ *  stress_get_cpu_cache_alpha()
+ *	find cache information as provided by linux Alpha from
+ *	/proc/cpu. Assume cache layout for 1st CPU is same for
+ *	all CPUs.
+ */
+static int stress_get_cpu_cache_alpha(
+	stress_cpu_t *cpu,
+	const char *cpu_path)
+{
+	FILE *fp;
+	const size_t count = 4;
+	size_t idx = 0;
+
+	(void)cpu_path;
+
+	/*
+	 * parse /proc/cpu info in the form:
+	 * L1 Icache		: 64K, 2-way, 64b line
+	 * L1 Dcache		: 64K, 2-way, 64b line
+	 * L2 cache		: n/a
+	 * L3 cache		: n/a
+	 */
+	cpu->caches = calloc(count, sizeof(*(cpu->caches)));
+	if (!cpu->caches) {
+		pr_err("failed to allocate %zu bytes for cpu caches\n",
+			count * sizeof(*(cpu->caches)));
+		return 0;
+	}
+
+	fp = fopen("/proc/cpuinfo", "r");
+	if (fp) {
+		char buffer[4096];
+
+		while ((idx < count) && fgets(buffer, sizeof(buffer), fp)) {
+			stress_cache_type_t cache_type = CACHE_TYPE_UNKNOWN;
+			uint16_t cache_level = 0;
+			char *ptr;
+			int cache_size, cache_ways, cache_line_size, n;
+
+			if (!strncmp("L1 Icache", buffer, 9)) {
+				cache_type = CACHE_TYPE_INSTRUCTION;
+				cache_level = 1;
+			} else if (!strncmp("L1 Dcache", buffer, 9))  {
+				cache_type = CACHE_TYPE_DATA;
+				cache_level = 1;
+			} else if (!strncmp("L2 cache", buffer, 8)) {
+				cache_type = CACHE_TYPE_DATA;
+				cache_level = 2;
+			} else if (!strncmp("L3 cache", buffer, 8)) {
+				cache_type = CACHE_TYPE_DATA;
+				cache_level = 3;
+			} else {
+				continue;
+			}
+			ptr = strchr(buffer, ':');
+			if (!ptr)
+				continue;
+			ptr++;
+			cache_size = 0;
+			cache_ways = 0;
+			cache_line_size = 0;
+			n = sscanf(ptr, "%dK, %d-way, %db line",
+				&cache_size, &cache_ways, &cache_line_size);
+			if (n != 3)
+				continue;
+			cpu->caches[idx].type = cache_type;
+			cpu->caches[idx].level = cache_level;
+			cpu->caches[idx].size = cache_size * 1024;
+			cpu->caches[idx].ways = cache_ways;
+			cpu->caches[idx].line_size = cache_line_size;
+			idx++;
+		}
+		(void)fclose(fp);
+	}
+
+	if (idx == 0) {
+		free(cpu->caches);
+		cpu->caches = NULL;
+		cpu->cache_count = 0;
+
+		return 0;
+	}
+	cpu->cache_count = idx;
+
+	return idx;
+}
+#endif
+
+#if defined(__APPLE__)
+/*
+ *  stress_get_cpu_cache_apple()
+ *	find cache information as provided by BSD sysctl
+ */
+static int stress_get_cpu_cache_apple(stress_cpu_t *cpu)
+{
+	typedef struct {
+		const char *name;			/* sysctl name */
+		const stress_cache_type_t type;		/* cache type */
+		const uint16_t level;			/* cache level 1, 2 */
+		const cache_size_type_t size_type;	/* cache size field */
+		const size_t index;			/* map to cpu->cache array index */
+	} cache_info_t;
+
+	static const cache_info_t cache_info[] = {
+		{ "hw.cachelinesize",		CACHE_TYPE_DATA,	1, STRESS_CACHE_LINE_SIZE,	0 },
+		{ "hw.l1dcachesize",		CACHE_TYPE_DATA,	1, STRESS_CACHE_SIZE,		0 },
+		{ "hw.cachelinesize",		CACHE_TYPE_INSTRUCTION,	1, STRESS_CACHE_LINE_SIZE,	1 },
+		{ "hw.l1icachesize",		CACHE_TYPE_INSTRUCTION,	1, STRESS_CACHE_SIZE,		1 },
+		{ "hw.l2cachesize",		CACHE_TYPE_UNIFIED,	2, STRESS_CACHE_SIZE,		2 },
+		{ "hw.l3cachesize",		CACHE_TYPE_UNIFIED,	3, STRESS_CACHE_SIZE,		2 },
+	};
+
+	const size_t count = 3;
+	size_t i;
+	bool valid = false;
+
+	cpu->caches = calloc(count, sizeof(*(cpu->caches)));
+	if (!cpu->caches) {
+		pr_err("failed to allocate %zu bytes for cpu caches\n",
+			count * sizeof(*(cpu->caches)));
+		return 0;
+	}
+
+	for (i = 0; i < SIZEOF_ARRAY(cache_info); i++) {
+		const size_t idx = cache_info[i].index;
+		uint64_t value;
+
+		value = stress_bsd_getsysctl_uint64(cache_info[i].name);
+
+		cpu->caches[idx].type = cache_info[i].type;
+		cpu->caches[idx].level = cache_info[i].level;
+		switch (cache_info[i].size_type) {
+		case STRESS_CACHE_SIZE:
+			cpu->caches[idx].size = value;
+			valid = true;
+			break;
+		case STRESS_CACHE_LINE_SIZE:
+			cpu->caches[idx].line_size = (uint32_t)value;
+			valid = true;
+			break;
+		case STRESS_CACHE_WAYS:
+			cpu->caches[idx].size = (uint32_t)value;
+			valid = true;
+			break;
+		default:
+			break;
+		}
+	}
+
+	if (!valid) {
+		free(cpu->caches);
+		cpu->caches = NULL;
+		cpu->cache_count = 0;
+
+		return 0;
+	}
+	cpu->cache_count = count;
+
+	return count;
+}
+#endif
+
+#if defined(__linux__) &&	\
+    defined(STRESS_ARCH_SPARC)
+/*
+ *  stress_get_cpu_cache_sparc64()
+ *	find cache information as provided by linux SPARC64
+ *	/sys/devices/system/cpu/cpu0
+ */
+static int stress_get_cpu_cache_sparc64(
+	stress_cpu_t *cpu,
+	const char *cpu_path)
+{
+	typedef struct {
+		const char *filename;			/* /sys proc name */
+		const stress_cache_type_t type;		/* cache type */
+		const uint16_t level;			/* cache level 1, 2 */
+		const cache_size_type_t size_type;	/* cache size field */
+		const size_t index;			/* map to cpu->cache array index */
+	} cache_info_t;
+
+	static const cache_info_t cache_info[] = {
+		{ "l1_dcache_line_size",	CACHE_TYPE_DATA,	1, STRESS_CACHE_LINE_SIZE,	0 },
+		{ "l1_dcache_size",		CACHE_TYPE_DATA,	1, STRESS_CACHE_SIZE,		0 },
+		{ "l1_icache_line_size",	CACHE_TYPE_INSTRUCTION,	1, STRESS_CACHE_LINE_SIZE,	1 },
+		{ "l1_icache_size",		CACHE_TYPE_INSTRUCTION,	1, STRESS_CACHE_SIZE,		1 },
+		{ "l2_cache_line_size",		CACHE_TYPE_UNIFIED,	2, STRESS_CACHE_LINE_SIZE,	2 },
+		{ "l2_cache_size",		CACHE_TYPE_UNIFIED,	2, STRESS_CACHE_SIZE,		2 },
+	};
+
+	const size_t count = 3;
+	size_t i;
+	bool valid = false;
+
+	cpu->caches = calloc(count, sizeof(*(cpu->caches)));
+	if (!cpu->caches) {
+		pr_err("failed to allocate %zu bytes for cpu caches\n",
+			count * sizeof(*(cpu->caches)));
+		return 0;
+	}
+
+	for (i = 0; i < SIZEOF_ARRAY(cache_info); i++) {
+		const size_t idx = cache_info[i].index;
+		uint64_t value;
+
+		if (stress_get_cpu_cache_value(cpu_path, cache_info[i].filename, &value) < 0)
+			continue;
+
+		cpu->caches[idx].type = cache_info[i].type;
+		cpu->caches[idx].level = cache_info[i].level;
+		switch (cache_info[i].size_type) {
+		case STRESS_CACHE_SIZE:
+			cpu->caches[idx].size = value;
+			valid = true;
+			break;
+		case STRESS_CACHE_LINE_SIZE:
+			cpu->caches[idx].line_size = (uint32_t)value;
+			valid = true;
+			break;
+		case STRESS_CACHE_WAYS:
+			cpu->caches[idx].size = (uint32_t)value;
+			valid = true;
+			break;
+		default:
+			break;
+		}
+	}
+
+	if (!valid) {
+		free(cpu->caches);
+		cpu->caches = NULL;
+		cpu->cache_count = 0;
+
+		return 0;
+	}
+	cpu->cache_count = count;
+
+	return count;
+}
+#endif
+
+#if defined(__linux__) &&	\
+    defined(STRESS_ARCH_M68K)
+static int stress_get_cpu_cache_m68k(stress_cpu_t *cpu)
+{
+	FILE *fp;
+	char buffer[1024];
+	size_t i, count;
+	size_t cache_type[2] = { 0, 0 };
+	size_t cache_size[2] = { 0, 0 };
+	int cpu_id = -1;
+
+	cpu->caches = NULL;
+	cpu->cache_count = 0;
+
+	fp = fopen("/proc/cpuinfo", "r");
+	if (!fp)
+		return 0;
+
+	(void)memset(buffer, 0, sizeof(buffer));
+	while (fgets(buffer, sizeof(buffer), fp) != NULL) {
+		if (strncmp("CPU:", buffer, 4) == 0) {
+			if (sscanf(buffer + 4, "%d", &cpu_id) == 1)
+				break;
+		}
+	}
+	(void)fclose(fp);
+
+	switch (cpu_id) {
+	case 68020:
+		count = 1;
+		cache_type[0] = CACHE_TYPE_INSTRUCTION;
+		cache_size[0] = 256;
+		break;
+	case 68030:
+		count = 2;
+		cache_type[0] = CACHE_TYPE_INSTRUCTION;
+		cache_size[0] = 256;
+		cache_type[1] = CACHE_TYPE_DATA;
+		cache_size[1] = 256;
+		break;
+	case 68040:
+		count = 2;
+		cache_type[0] = CACHE_TYPE_INSTRUCTION;
+		cache_size[0] = 4096;
+		cache_type[1] = CACHE_TYPE_DATA;
+		cache_size[1] = 4096;
+		break;
+	case 68060:
+		count = 2;
+		cache_type[0] = CACHE_TYPE_INSTRUCTION;
+		cache_size[0] = 8192;
+		cache_type[1] = CACHE_TYPE_DATA;
+		cache_size[1] = 8192;
+		break;
+	default:
+		return 0;
+	}
+
+	cpu->caches = calloc(count, sizeof(*(cpu->caches)));
+	if (!cpu->caches) {
+		pr_err("failed to allocate %zu bytes for cpu caches\n",
+			count * sizeof(*(cpu->caches)));
+		return 0;
+	}
+	for (i = 0; i < count; i++) {
+		cpu->caches[i].type = cache_type[i];
+		cpu->caches[i].level = 1;
+		cpu->caches[i].size = cache_size[i];
+		cpu->caches[i].line_size = 64;	/* Assumption! */
+		cpu->caches[i].ways = cache_size[i] / 64;
+	}
+	cpu->cache_count = count;
+
+	return count;
+}
+#endif
+
+#if defined(__linux__) ||	\
+    defined(__APPLE__)
 /*
  * stress_size_to_bytes()
  * 	Convert human-readable integer sizes (such as "32K", "4M") into bytes.
@@ -142,6 +564,13 @@ static uint64_t stress_size_to_bytes(const char *str)
 	return bytes;
 }
 
+static const stress_generic_map_t cache_type_map[] = {
+	{ "data",		CACHE_TYPE_DATA },
+	{ "instruction",	CACHE_TYPE_INSTRUCTION },
+	{ "unified",		CACHE_TYPE_UNIFIED },
+	{  NULL,		CACHE_TYPE_UNKNOWN }
+};
+
 /*
  * stress_get_cache_type()
  * @name: human-readable cache type.
@@ -178,11 +607,9 @@ out:
  */
 static int stress_add_cpu_cache_detail(stress_cpu_cache_t *cache, const char *index_path)
 {
-	const size_t index_posn = index_path ? strlen(index_path) : 0;
-	const size_t path_len = index_posn + 32;
 	int ret = EXIT_FAILURE;
 	char tmp[2048];
-	char path[path_len];
+	char path[PATH_MAX];
 
 	(void)memset(path, 0, sizeof(path));
 	if (!cache)
@@ -212,9 +639,9 @@ static int stress_add_cpu_cache_detail(stress_cpu_cache_t *cache, const char *in
 	cache->line_size = (uint32_t)atoi(tmp);
 
 	(void)stress_mk_filename(path, sizeof(path), index_path, "ways_of_associativity");
-	if (stress_get_string_from_file(path, tmp, sizeof(tmp)) < 0)
+	if (stress_get_string_from_file(path, tmp, sizeof(tmp)) < 0) {
 		cache->ways = 0;
-	else {
+	} else {
 		if (sscanf(tmp, "%" SCNu32, &cache->ways) != 1)
 			cache->ways = 0;
 	}
@@ -224,113 +651,80 @@ out:
 }
 
 /*
- * stress_get_cache_by_cpu()
- * @cpu: cpu to consider.
- * @cache_level: numeric cache level (1-indexed).
- * Obtain the cpu cache indexed by @cache_level.
- *
- * POTENTIAL BUG: assumes only 1 data cache per CPU cache level.
- *
- * Returns: stress_cpu_cache_t, or NULL on error.
+ *  index_filter()
+ *	return 1 when filename is index followed by a digit
  */
-static stress_cpu_cache_t * stress_get_cache_by_cpu(const stress_cpu_t *cpu, const int cache_level)
+static int index_filter(const struct dirent *d)
 {
-	uint32_t  i;
-
-	if (!cpu || !cache_level)
-		return NULL;
-
-	for (i = 0; i < cpu->cache_count; i++) {
-		stress_cpu_cache_t *p;
-
-		p = &cpu->caches[i];
-		if (p->level != cache_level)
-			continue;
-
-		/* we want a data cache */
-		if (p->type != CACHE_TYPE_INSTRUCTION)
-			return p;
-	}
-	return NULL;
+	return ((strncmp(d->d_name, "index", 5) == 0) && isdigit(d->d_name[5]));
 }
 
 /*
- * stress_get_max_cache_level()
- * @cpus: array of cpus to query.
- * Determine the maximum cache level available on the system.
- *
- * Returns: 1-index value denoting highest cache level, or 0 on error.
+ *  index_sort()
+ *	sort by index number (digits 5 onwards)
  */
-uint16_t stress_get_max_cache_level(const stress_cpus_t *cpus)
+static int index_sort(const struct dirent **d1, const struct dirent **d2)
 {
-	stress_cpu_t    *cpu;
-	uint32_t  i;
-	uint16_t  max = 0;
+	const int i1 = atoi(&(*d1)->d_name[5]);
+	const int i2 = atoi(&(*d2)->d_name[5]);
 
-	if (!cpus) {
-		pr_dbg("%s: invalid cpus parameter\n", __func__);
+	return i1 - i2;
+}
+
+/*
+ *  stress_get_cpu_cache_index()
+ *	find cache information as provided by cache info indexes
+ *	in /sys/devices/system/cpu/cpu*
+ */
+static int stress_get_cpu_cache_index(
+	stress_cpu_t *cpu,
+	const char *cpu_path)
+{
+	struct dirent **namelist = NULL;
+	int n;
+	uint32_t i;
+	char path[PATH_MAX];
+
+	(void)stress_mk_filename(path, sizeof(path), cpu_path, SYS_CPU_CACHE_DIR);
+	n = scandir(path, &namelist, index_filter, index_sort);
+	if (n <= 0) {
+		cpu->caches = NULL;
 		return 0;
 	}
+	cpu->cache_count = (uint32_t)n;
+	cpu->caches = calloc(cpu->cache_count, sizeof(*(cpu->caches)));
+	if (!cpu->caches) {
+		size_t cache_bytes = cpu->cache_count * sizeof(*(cpu->caches));
 
-	cpu = &cpus->cpus[stress_cache_get_cpu(cpus)];
+		pr_err("failed to allocate %zu bytes for cpu caches\n",
+			cache_bytes);
+
+		cpu->caches = NULL;
+		cpu->cache_count = 0;
+		goto list_free;
+	}
 
 	for (i = 0; i < cpu->cache_count; i++) {
-		stress_cpu_cache_t *cache;
+		const char *name = namelist[i]->d_name;
+		char fullpath[PATH_MAX];
 
-		cache = &cpu->caches[i];
-		max = cache->level > max ? cache->level : max;
+		(void)memset(fullpath, 0, sizeof(fullpath));
+		(void)stress_mk_filename(fullpath, sizeof(fullpath), path, name);
+		if (stress_add_cpu_cache_detail(&cpu->caches[i], fullpath) != EXIT_SUCCESS) {
+			free(cpu->caches);
+			cpu->caches = NULL;
+			cpu->cache_count = 0;
+
+			goto list_free;
+		}
 	}
+list_free:
+	n = (int)cpu->cache_count;
+	stress_dirent_list_free(namelist, n);
 
-	return max;
+	return n;
 }
 
-/*
- * stress_get_cpu_cache()
- * @cpus: array of cpus to query.
- * @cache_level: numeric cache level (1-indexed).
- * Obtain a cpu cache of level @cache_level.
- *
- * Returns: stress_cpu_cache_t pointer, or NULL on error.
- */
-stress_cpu_cache_t *stress_get_cpu_cache(const stress_cpus_t *cpus, const uint16_t cache_level)
-{
-	stress_cpu_t *cpu;
-
-	if (!cpus) {
-		pr_dbg("%s: invalid cpus parameter\n", __func__);
-		return NULL;
-	}
-
-	if (!cache_level) {
-		pr_dbg("%s: invalid cache_level: %d\n",
-			__func__, cache_level);
-		return NULL;
-	}
-
-	cpu = &cpus->cpus[stress_cache_get_cpu(cpus)];
-
-	return stress_get_cache_by_cpu(cpu, cache_level);
-}
-
-#if defined(STRESS_ARCH_SPARC)
-static int stress_get_cpu_cache_value(
-	const char *cpu_path,
-	const char *file,
-	uint64_t *value)
-{
-	const size_t cpu_path_len = strlen(cpu_path);
-	char path[cpu_path_len + 128];
-	char tmp[128];
-
-	(void)stress_mk_filename(path, sizeof(path), cpu_path, file);
-	if (stress_get_string_from_file(path, tmp, sizeof(tmp)) == 0) {
-		if (sscanf(tmp, "%" SCNu64, value) == 1)
-			return 0;
-	}
-	return -1;
-}
-#endif
-
 /*
  *  stress_get_cpu_cache_auxval()
  *	find cache information as provided by getauxval
@@ -353,16 +747,16 @@ static int stress_get_cpu_cache_auxval(stress_cpu_t *cpu)
 
 	static const cache_auxval_info_t cache_auxval_info[] = {
 #if defined(AT_L1D_CACHESIZE)
-		{ AT_L1D_CACHESIZE,	CACHE_TYPE_DATA,	1, CACHE_SIZE,	0 },
+		{ AT_L1D_CACHESIZE,	CACHE_TYPE_DATA,	1, STRESS_CACHE_SIZE,	0 },
 #endif
 #if defined(AT_L1I_CACHESIZE)
-		{ AT_L1I_CACHESIZE,	CACHE_TYPE_INSTRUCTION,	1, CACHE_SIZE,	1 },
+		{ AT_L1I_CACHESIZE,	CACHE_TYPE_INSTRUCTION,	1, STRESS_CACHE_SIZE,	1 },
 #endif
 #if defined(AT_L2_CACHESIZE)
-		{ AT_L2_CACHESIZE,	CACHE_TYPE_UNIFIED,	2, CACHE_SIZE,	2 },
+		{ AT_L2_CACHESIZE,	CACHE_TYPE_UNIFIED,	2, STRESS_CACHE_SIZE,	2 },
 #endif
 #if defined(AT_L3_CACHESIZE)
-		{ AT_L3_CACHESIZE,	CACHE_TYPE_UNIFIED,	2, CACHE_SIZE,	3 },
+		{ AT_L3_CACHESIZE,	CACHE_TYPE_UNIFIED,	3, STRESS_CACHE_SIZE,	2 },
 #endif
 	};
 
@@ -387,13 +781,13 @@ static int stress_get_cpu_cache_auxval(stress_cpu_t *cpu)
 		cpu->caches[index].type = cache_auxval_info[i].type;
 		cpu->caches[index].level = cache_auxval_info[i].level;
 		switch (cache_auxval_info[i].size_type) {
-		case CACHE_SIZE:
+		case STRESS_CACHE_SIZE:
 			cpu->caches[index].size = value;
 			break;
-		case CACHE_LINE_SIZE:
+		case STRESS_CACHE_LINE_SIZE:
 			cpu->caches[index].line_size = (uint32_t)value;
 			break;
-		case CACHE_WAYS:
+		case STRESS_CACHE_WAYS:
 			cpu->caches[index].size = (uint32_t)value;
 			break;
 		default:
@@ -419,273 +813,6 @@ static int stress_get_cpu_cache_auxval(stress_cpu_t *cpu)
 #endif
 }
 
-#if defined(STRESS_ARCH_ALPHA)
-/*
- *  stress_get_cpu_cache_alpha()
- *	find cache information as provided by linux Alpha from
- *	/proc/cpu. Assume cache layout for 1st CPU is same for
- *	all CPUs.
- */
-static int stress_get_cpu_cache_alpha(
-	stress_cpu_t *cpu,
-	const char *cpu_path)
-{
-	FILE *fp;
-	const size_t count = 4;
-	size_t idx = 0;
-
-	(void)cpu_path;
-
-	/*
-	 * parse /proc/cpu info in the form:
-	 * L1 Icache		: 64K, 2-way, 64b line
-	 * L1 Dcache		: 64K, 2-way, 64b line
-	 * L2 cache		: n/a
-	 * L3 cache		: n/a
-	 */
-	cpu->caches = calloc(count, sizeof(*(cpu->caches)));
-	if (!cpu->caches) {
-		pr_err("failed to allocate %zu bytes for cpu caches\n",
-			count * sizeof(*(cpu->caches)));
-		return 0;
-	}
-
-	fp = fopen("/proc/cpuinfo", "r");
-	if (fp) {
-		char buffer[4096];
-
-		while ((idx < count) && fgets(buffer, sizeof(buffer), fp)) {
-			stress_cache_type_t cache_type = CACHE_TYPE_UNKNOWN;
-			uint16_t cache_level = 0;
-			char *ptr;
-			int cache_size, cache_ways, cache_line_size, n;
-
-			if (!strncmp("L1 Icache", buffer, 9)) {
-				cache_type = CACHE_TYPE_INSTRUCTION;
-				cache_level = 1;
-			} else if (!strncmp("L1 Dcache", buffer, 9))  {
-				cache_type = CACHE_TYPE_DATA;
-				cache_level = 1;
-			}
-			else if (!strncmp("L2 cache", buffer, 8)) {
-				cache_type = CACHE_TYPE_DATA;
-				cache_level = 2;
-			}
-			else if (!strncmp("L3 cache", buffer, 8)) {
-				cache_type = CACHE_TYPE_DATA;
-				cache_level = 3;
-			}
-			else
-				continue;
-			ptr = strchr(buffer, ':');
-			if (!ptr)
-				continue;
-			ptr++;
-			cache_size = 0;
-			cache_ways = 0;
-			cache_line_size = 0;
-			n = sscanf(ptr, "%dK, %d-way, %db line",
-				&cache_size, &cache_ways, &cache_line_size);
-			if (n != 3)
-				continue;
-			cpu->caches[idx].type = cache_type;
-			cpu->caches[idx].level = cache_level;
-			cpu->caches[idx].size = cache_size * 1024;
-			cpu->caches[idx].ways = cache_ways;
-			cpu->caches[idx].line_size = cache_line_size;
-			idx++;
-		}
-		(void)fclose(fp);
-	}
-
-	if (idx == 0) {
-		free(cpu->caches);
-		cpu->caches = NULL;
-		cpu->cache_count = 0;
-
-		return 0;
-	}
-	cpu->cache_count = idx;
-
-	return idx;
-}
-#endif
-
-#if defined(STRESS_ARCH_SPARC)
-/*
- *  stress_get_cpu_cache_sparc64()
- *	find cache information as provided by linux SPARC64
- *	/sys/devices/system/cpu/cpu0
- */
-static int stress_get_cpu_cache_sparc64(
-	stress_cpu_t *cpu,
-	const char *cpu_path)
-{
-	typedef struct {
-		const char *filename;			/* /sys proc name */
-		const stress_cache_type_t type;		/* cache type */
-		const uint16_t level;			/* cache level 1, 2 */
-		const cache_size_type_t size_type;	/* cache size field */
-		const size_t index;			/* map to cpu->cache array index */
-	} cache_info_t;
-
-	static const cache_info_t cache_info[] = {
-		{ "l1_dcache_line_size",	CACHE_TYPE_DATA,	1, CACHE_LINE_SIZE,	0 },
-		{ "l1_dcache_size",		CACHE_TYPE_DATA,	1, CACHE_SIZE,		0 },
-		{ "l1_icache_line_size",	CACHE_TYPE_INSTRUCTION,	1, CACHE_LINE_SIZE,	1 },
-		{ "l1_icache_size",		CACHE_TYPE_INSTRUCTION,	1, CACHE_SIZE,		1 },
-		{ "l2_cache_line_size",		CACHE_TYPE_UNIFIED,	2, CACHE_LINE_SIZE,	2 },
-		{ "l2_cache_size",		CACHE_TYPE_UNIFIED,	2, CACHE_SIZE,		2 },
-	};
-
-	const size_t count = 3;
-	size_t i;
-	bool valid = false;
-
-	cpu->caches = calloc(count, sizeof(*(cpu->caches)));
-	if (!cpu->caches) {
-		pr_err("failed to allocate %zu bytes for cpu caches\n",
-			count * sizeof(*(cpu->caches)));
-		return 0;
-	}
-
-	for (i = 0; i < SIZEOF_ARRAY(cache_info); i++) {
-		const size_t idx = cache_info[i].index;
-		uint64_t value;
-
-		if (stress_get_cpu_cache_value(cpu_path, cache_info[i].filename, &value) < 0)
-			continue;
-
-		cpu->caches[idx].type = cache_info[i].type;
-		cpu->caches[idx].level = cache_info[i].level;
-		switch (cache_info[i].size_type) {
-		case CACHE_SIZE:
-			cpu->caches[idx].size = value;
-			valid = true;
-			break;
-		case CACHE_LINE_SIZE:
-			cpu->caches[idx].line_size = (uint32_t)value;
-			valid = true;
-			break;
-		case CACHE_WAYS:
-			cpu->caches[idx].size = (uint32_t)value;
-			valid = true;
-			break;
-		default:
-			break;
-		}
-	}
-
-	if (!valid) {
-		free(cpu->caches);
-		cpu->caches = NULL;
-		cpu->cache_count = 0;
-
-		return 0;
-	}
-	cpu->cache_count = count;
-
-	return count;
-}
-#endif
-
-/*
- *  index_filter()
- *	return 1 when filename is index followed by a digit
- */
-static int index_filter(const struct dirent *d)
-{
-	return ((strncmp(d->d_name, "index", 5) == 0) && isdigit(d->d_name[5]));
-}
-
-/*
- *  index_sort()
- *	sort by index number (digits 5 onwards)
- */
-static int index_sort(const struct dirent **d1, const struct dirent **d2)
-{
-	const int i1 = atoi(&(*d1)->d_name[5]);
-	const int i2 = atoi(&(*d2)->d_name[5]);
-
-	return i1 - i2;
-}
-
-/*
- *  cpu_filter()
- *	return 1 when filename is cpu followed by a digit
- */
-static int cpu_filter(const struct dirent *d)
-{
-	return ((strncmp(d->d_name, "cpu", 3) == 0) && isdigit(d->d_name[3]));
-}
-
-/*
- *  cpu_sort()
- *	sort by CPU number (digits 3 onwards)
- */
-static int cpusort(const struct dirent **d1, const struct dirent **d2)
-{
-	const int c1 = atoi(&(*d1)->d_name[3]);
-	const int c2 = atoi(&(*d2)->d_name[3]);
-
-	return c1 - c2;
-}
-
-/*
- *  stress_get_cpu_cache_index()
- *	find cache information as provided by cache info indexes
- *	in /sys/devices/system/cpu/cpu*
- */
-static int stress_get_cpu_cache_index(
-	stress_cpu_t *cpu,
-	const char *cpu_path)
-{
-	struct dirent **namelist = NULL;
-	int n;
-	uint32_t i;
-	const size_t cpu_path_len = strlen(cpu_path);
-	char path[cpu_path_len + 128];
-
-	(void)stress_mk_filename(path, sizeof(path), cpu_path, SYS_CPU_CACHE_DIR);
-	n = scandir(path, &namelist, index_filter, index_sort);
-	if (n <= 0) {
-		cpu->caches = NULL;
-		return 0;
-	}
-	cpu->cache_count = (uint32_t)n;
-	cpu->caches = calloc(cpu->cache_count, sizeof(*(cpu->caches)));
-	if (!cpu->caches) {
-		size_t cache_bytes = cpu->cache_count * sizeof(*(cpu->caches));
-
-		pr_err("failed to allocate %zu bytes for cpu caches\n",
-			cache_bytes);
-
-		cpu->caches = NULL;
-		cpu->cache_count = 0;
-		goto list_free;
-	}
-
-	for (i = 0; i < cpu->cache_count; i++) {
-		const char *name = namelist[i]->d_name;
-		char fullpath[strlen(path) + strlen(name) + 2];
-
-		(void)memset(fullpath, 0, sizeof(fullpath));
-		(void)stress_mk_filename(fullpath, sizeof(fullpath), path, name);
-		if (stress_add_cpu_cache_detail(&cpu->caches[i], fullpath) != EXIT_SUCCESS) {
-			free(cpu->caches);
-			cpu->caches = NULL;
-			cpu->cache_count = 0;
-
-			goto list_free;
-		}
-	}
-list_free:
-	n = (int)cpu->cache_count;
-	stress_dirent_list_free(namelist, n);
-
-	return n;
-}
-
 /*
  * stress_get_cpu_cache_details()
  * @cpu: cpu to fill in.
@@ -713,19 +840,56 @@ static void stress_get_cpu_cache_details(stress_cpu_t *cpu, const char *cpu_path
 	if (stress_get_cpu_cache_auxval(cpu) > 0)
 		return;
 
-#if defined(STRESS_ARCH_SPARC)
+#if defined(__linux__) &&	\
+    defined(STRESS_ARCH_SPARC)
 	/* Try cache info for sparc CPUs */
 	if (stress_get_cpu_cache_sparc64(cpu, cpu_path) > 0)
 		return;
 #endif
 
-#if defined(STRESS_ARCH_ALPHA)
+#if defined(__linux__) &&	\
+    defined(STRESS_ARCH_M68K)
+	if (stress_get_cpu_cache_m68k(cpu) > 0)
+		return;
+#endif
+
+#if defined(__linux__) &&	\
+    defined(STRESS_ARCH_ALPHA)
 	if (stress_get_cpu_cache_alpha(cpu, cpu_path) > 0)
 		return;
 #endif
 
+#if defined(__APPLE__)
+	if (stress_get_cpu_cache_apple(cpu) > 0)
+		return;
+#endif
+
 	return;
 }
+#endif
+
+#if defined(__linux__)
+
+/*
+ *  cpu_filter()
+ *	return 1 when filename is cpu followed by a digit
+ */
+static int cpu_filter(const struct dirent *d)
+{
+	return ((strncmp(d->d_name, "cpu", 3) == 0) && isdigit(d->d_name[3]));
+}
+
+/*
+ *  cpu_sort()
+ *	sort by CPU number (digits 3 onwards)
+ */
+static int cpu_sort(const struct dirent **d1, const struct dirent **d2)
+{
+	const int c1 = atoi(&(*d1)->d_name[3]);
+	const int c2 = atoi(&(*d2)->d_name[3]);
+
+	return c1 - c2;
+}
 
 /*
  * stress_get_all_cpu_cache_details()
@@ -739,7 +903,7 @@ stress_cpus_t *stress_get_all_cpu_cache_details(void)
 	stress_cpus_t *cpus = NULL;
 	struct dirent **namelist = NULL;
 
-	cpu_count = scandir(SYS_CPU_PREFIX, &namelist, cpu_filter, cpusort);
+	cpu_count = scandir(SYS_CPU_PREFIX, &namelist, cpu_filter, cpu_sort);
 	if (cpu_count < 1) {
 		pr_err("no CPUs found in %s\n", SYS_CPU_PREFIX);
 		goto out;
@@ -758,8 +922,7 @@ stress_cpus_t *stress_get_all_cpu_cache_details(void)
 
 	for (i = 0; i < cpu_count; i++) {
 		const char *name = namelist[i]->d_name;
-		const size_t fullpath_len = strlen(SYS_CPU_PREFIX) + strlen(name) + 2;
-		char fullpath[fullpath_len];
+		char fullpath[PATH_MAX];
 		stress_cpu_t *const cpu = &cpus->cpus[i];
 
 		(void)memset(fullpath, 0, sizeof(fullpath));
@@ -769,7 +932,7 @@ stress_cpus_t *stress_get_all_cpu_cache_details(void)
 			/* 1st CPU cannot be taken offline */
 			cpu->online = 1;
 		} else {
-			char onlinepath[fullpath_len + 8];
+			char onlinepath[PATH_MAX + 8];
 			char tmp[2048];
 
 			(void)memset(onlinepath, 0, sizeof(onlinepath));
@@ -777,7 +940,7 @@ stress_cpus_t *stress_get_all_cpu_cache_details(void)
 			if (stress_get_string_from_file(onlinepath, tmp, sizeof(tmp)) < 0) {
 				/* Assume it is online, it is the best we can do */
 				cpu->online = 1;
-			} else  {
+			} else {
 				cpu->online = atoi(tmp);
 			}
 		}
@@ -789,6 +952,49 @@ out:
 	stress_dirent_list_free(namelist, cpu_count);
 	return cpus;
 }
+#elif defined(__APPLE__)
+/*
+ * stress_get_all_cpu_cache_details()
+ * Obtain information on all cpus caches on the system.
+ *
+ * Returns: dynamically-allocated stress_cpus_t object, or NULL on error.
+ */
+stress_cpus_t *stress_get_all_cpu_cache_details(void)
+{
+	int32_t i, cpu_count;
+	stress_cpus_t *cpus = NULL;
+	struct dirent **namelist = NULL;
+
+	if (stress_bsd_getsysctl("hw.physicalcpu", &cpu_count, sizeof(cpu_count)) < 0) {
+		pr_err("no CPUs found using sysctl hw.physicalcpu\n");
+		goto out;
+	}
+	cpus = calloc(1, sizeof(*cpus));
+	if (!cpus)
+		goto out;
+
+	cpus->cpus = calloc((size_t)cpu_count, sizeof(*(cpus->cpus)));
+	if (!cpus->cpus) {
+		free(cpus);
+		cpus = NULL;
+		goto out;
+	}
+	cpus->count = (uint32_t)cpu_count;
+
+	for (i = 0; i < cpu_count; i++) {
+		stress_get_cpu_cache_details(&cpus->cpus[i], "");
+	}
+
+out:
+	stress_dirent_list_free(namelist, cpu_count);
+	return cpus;
+}
+#else
+stress_cpus_t *stress_get_all_cpu_cache_details(void)
+{
+	return NULL;
+}
+#endif
 
 /*
  * stress_free_cpu_caches()
@@ -817,4 +1023,40 @@ void stress_free_cpu_caches(stress_cpus_t *cpus)
 	free(cpus);
 }
 
+/*
+ *  stress_get_llc_size()
+ * 	get Lower Level Cache size and Cache Line size (sizes in bytes)
+ *	sizes are zero if not available.
+ */
+void stress_get_llc_size(size_t *llc_size, size_t *cache_line_size)
+{
+#if defined(__linux__) ||	\
+    defined(__APPLE__)
+	uint16_t max_cache_level;
+	stress_cpus_t *cpu_caches;
+	stress_cpu_cache_t *cache = NULL;
+
+	*llc_size = 0;
+	*cache_line_size = 0;
+
+	cpu_caches = stress_get_all_cpu_cache_details();
+	if (!cpu_caches)
+		return;
+
+	max_cache_level = stress_get_max_cache_level(cpu_caches);
+	if (max_cache_level < 1)
+		goto free_cpu_caches;
+	cache = stress_get_cpu_cache(cpu_caches, max_cache_level);
+	if (!cache)
+		goto free_cpu_caches;
+
+	*llc_size = cache->size;
+	*cache_line_size = cache->line_size ? cache->line_size : 64;
+
+free_cpu_caches:
+	stress_free_cpu_caches(cpu_caches);
+#else
+	*llc_size = 0;
+	*cache_line_size = 0;
 #endif
+}
