@@ -1549,8 +1549,10 @@ void stress_rndstr(char *str, size_t len)
 	 * Be careful if expanding this alphabet, some of this function's users
 	 * use it to generate random filenames.
 	 */
-	static const char alphabet[64] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
-		"abcdefghijklmnopqrstuvwxyz" "0123456789" "-_";
+	static const char alphabet[64] = 
+		"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
+		"abcdefghijklmnopqrstuvwxyz"
+		"0123456789" "-_";
 	size_t i;
 	uint32_t r = 0;		/* Silence a gcc warning */
 	if (len == 0)
@@ -1727,6 +1729,7 @@ void pr_yaml_runinfo(FILE *yaml)
 		pr_yaml(yaml, "      machine: %s\n", uts.machine);
 	}
 #endif
+	pr_yaml(yaml, "      compiler: '%s'\n", stress_get_compiler());
 #if defined(HAVE_SYS_SYSINFO_H) &&	\
     defined(HAVE_SYSINFO)
 	(void)memset(&info, 0, sizeof(info));
@@ -2297,6 +2300,11 @@ unsigned int stress_get_cpu(void)
 const char *stress_get_compiler(void)
 {
 #if   defined(__ICC) && 		\
+      defined(__INTEL_COMPILER) &&	\
+      defined(__INTEL_COMPILER_UPDATE) && \
+      defined(__INTEL_COMPILER_BUILD_DATE)
+	static const char cc[] = "icc " XSTRINGIFY(__INTEL_COMPILER) "." XSTRINGIFY(__INTEL_COMPILER_UPDATE) " Build " XSTRINGIFY(__INTEL_COMPILER_BUILD_DATE) "";
+#elif defined(__ICC) && 		\
       defined(__INTEL_COMPILER) &&	\
       defined(__INTEL_COMPILER_UPDATE)
 	static const char cc[] = "icc " XSTRINGIFY(__INTEL_COMPILER) "." XSTRINGIFY(__INTEL_COMPILER_UPDATE) "";
@@ -2305,9 +2313,17 @@ const char *stress_get_compiler(void)
 #elif defined(__PCC__) &&			\
        defined(__PCC_MINOR__)
 	static const char cc[] = "pcc " XSTRINGIFY(__PCC__) "." XSTRINGIFY(__PCC_MINOR__) "." XSTRINGIFY(__PCC_MINORMINOR__) "";
+#elif defined(__clang_major__) &&	\
+      defined(__clang_minor__) &&	\
+      defined(__clang_patchlevel__)
+	static const char cc[] = "clang " XSTRINGIFY(__clang_major__) "." XSTRINGIFY(__clang_minor__) "." XSTRINGIFY(__clang_patchlevel__) "";
 #elif defined(__clang_major__) &&	\
       defined(__clang_minor__)
 	static const char cc[] = "clang " XSTRINGIFY(__clang_major__) "." XSTRINGIFY(__clang_minor__) "";
+#elif defined(__GNUC__) &&		\
+      defined(__GNUC_MINOR__) &&	\
+      defined(__GNUC_PATCHLEVEL__)
+	static const char cc[] = "gcc " XSTRINGIFY(__GNUC__) "." XSTRINGIFY(__GNUC_MINOR__) "." XSTRINGIFY(__GNUC_PATCHLEVEL__) "";
 #elif defined(__GNUC__) &&		\
       defined(__GNUC_MINOR__)
 	static const char cc[] = "gcc " XSTRINGIFY(__GNUC__) "." XSTRINGIFY(__GNUC_MINOR__) "";
