@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C)      2022 Colin Ian King.
+ * Copyright (C) 2022-2023 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -39,7 +39,7 @@ static const stress_help_t help[] = {
 
 /*
  *  stress_alarm
- *	stress alarm(3)
+ *	stress alarm()
  */
 static int stress_alarm(const stress_args_t *args)
 {
@@ -116,7 +116,7 @@ again:
 
 		while (keep_stressing(args) &&
 		       (stress_time_now() < t_end)) {
-			const uint64_t delay_ns = 1000 + stress_mwc32() % 10000;
+			const uint64_t delay_ns = 1000 + stress_mwc32modn(10000);
 
 			(void)kill(pid, SIGALRM);
 			shim_nanosleep_uint64(delay_ns);
