@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C)      2022 Colin Ian King.
+ * Copyright (C) 2022-2023 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -21,9 +21,9 @@
 
 static const stress_help_t help[] = {
 	{ NULL,	"brk N",	"start N workers performing rapid brk calls" },
-	{ NULL,	"brk-ops N",	"stop after N brk bogo operations" },
 	{ NULL, "brk-mlock",	"attempt to mlock newly mapped brk pages" },
 	{ NULL,	"brk-notouch",	"don't touch (page in) new data segment page" },
+	{ NULL,	"brk-ops N",	"stop after N brk bogo operations" },
 	{ NULL,	NULL,		NULL }
 };
 
@@ -34,18 +34,12 @@ typedef struct {
 
 static int stress_set_brk_mlock(const char *opt)
 {
-	(void)opt;
-	bool brk_mlock = true;
-
-	return stress_set_setting("brk-mlock", TYPE_ID_BOOL, &brk_mlock);
+	return stress_set_setting_true("brk-mlock", opt);
 }
 
 static int stress_set_brk_notouch(const char *opt)
 {
-	(void)opt;
-	bool brk_notouch = true;
-
-	return stress_set_setting("brk-notouch", TYPE_ID_BOOL, &brk_notouch);
+	return stress_set_setting_true("brk-notouch", opt);
 }
 
 static const stress_opt_set_func_t opt_set_funcs[] = {
@@ -83,10 +77,13 @@ static int stress_brk_supported(const char *name)
 
 static int stress_brk_child(const stress_args_t *args, void *context)
 {
-	uint8_t *start_ptr;
+	uint8_t *start_ptr, *unmap_ptr = NULL;
 	int i = 0;
 	const size_t page_size = args->page_size;
 	const brk_context_t *brk_context = (brk_context_t *)context;
+	double sbrk_exp_duration = 0.0, sbrk_exp_count = 0.0;
+	double sbrk_shr_duration = 0.0, sbrk_shr_count = 0.0;
+	double rate;
 
 	start_ptr = shim_sbrk(0);
 	if (start_ptr == (void *) -1) {
@@ -104,32 +101,53 @@ static int stress_brk_child(const stress_args_t *args, void *context)
 
 	do {
 		uint8_t *ptr;
+		double t;
+
+
+		/* Low memory avoidance, re-start */
+		if ((g_opt_flags & OPT_FLAGS_OOM_AVOID) && stress_low_memory(page_size))
+			VOID_RET(int, shim_brk(start_ptr));
 
 		i++;
 		if (i < 8) {
 			/* Expand brk by 1 page */
+			t = stress_time_now();
 			ptr = shim_sbrk((intptr_t)page_size);
+			if (ptr != (void *)-1) {
+				sbrk_exp_duration += stress_time_now() - t;
+				sbrk_exp_count += 1.0;
+				if (!unmap_ptr)
+					unmap_ptr = ptr;
+			}
 		} else if (i < 9) {
 			/* brk to same brk position */
 			ptr = shim_sbrk(0);
 			if (shim_brk(ptr) < 0)
 				ptr = (void *)-1;
-		} else {
+		} else if (i < 10) {
 			/* Shrink brk by 1 page */
-			i = 0;
+			t = stress_time_now();
 			ptr = shim_sbrk(0);
+			if (ptr != (void *)-1) {
+				sbrk_shr_duration += stress_time_now() - t;
+				sbrk_shr_count += 1.0;
+			}
 			ptr -= page_size;
 			if (shim_brk(ptr) < 0)
 				ptr = (void *)-1;
+		} else {
+			i = 0;
+			/* remove a page from brk region */
+			if (unmap_ptr) {
+				(void)munmap((void *)unmap_ptr, page_size);
+				unmap_ptr = NULL;
+			}
+			continue;
 		}
 
 		if (ptr == (void *)-1) {
 			if ((errno == ENOMEM) || (errno == EAGAIN)) {
-				int ignore;
-
-				ignore = shim_brk(start_ptr);
-				(void)ignore;
-
+				VOID_RET(int, shim_brk(start_ptr));
 			} else {
 				pr_fail("%s: sbrk(%d) failed: errno=%d (%s)\n",
 					args->name, (int)page_size, errno,
@@ -137,9 +155,11 @@ static int stress_brk_child(const stress_args_t *args, void *context)
 				return EXIT_FAILURE;
 			}
 		} else {
+#if !defined(__APPLE__)
 			/* Touch page, force it to be resident */
 			if (!brk_context->brk_notouch)
 				*(ptr - 1) = 0;
+#endif
 
 #if defined(HAVE_MADVISE) &&	\
     defined(MADV_MERGEABLE)
@@ -151,6 +171,11 @@ static int stress_brk_child(const stress_args_t *args, void *context)
 		inc_counter(args);
 	} while (keep_stressing(args));
 
+	rate = (sbrk_exp_count > 0.0) ? (double)sbrk_exp_duration / sbrk_exp_count : 0.0;
+	stress_metrics_set(args, 0, "nanosecs per sbrk page expand", rate * STRESS_DBL_NANOSECOND);
+	rate = (sbrk_shr_count > 0.0) ? (double)sbrk_shr_duration / sbrk_shr_count : 0.0;
+	stress_metrics_set(args, 1, "nanosecs per sbrk page shrink", rate * STRESS_DBL_NANOSECOND);
+
 	return EXIT_SUCCESS;
 }
 
