@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C)      2022 Colin Ian King.
+ * Copyright (C) 2022-2023 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -28,15 +28,11 @@ static const stress_help_t help[] = {
 	{ NULL,	NULL,		NULL }
 };
 
-#if defined(HAVE_SYSCALL_H)
+#if defined(HAVE_SYSCALL_H) &&	\
+    defined(HAVE_SYSCALL)
 
 #define HASH_SYSCALL_SIZE	(1987)
 
-#define CALL_UNDEFINED		(0)
-#define CALL_BY_SYSCALL		(1)	/* syscall() call */
-#define CALL_BY_X86_SYSCALL	(2)	/* x86 syscall opcode */
-#define CALL_BY_X86_INT80	(3)	/* x86 int $80 */
-
 #if defined(__NR_syscalls)
 #define MAX_SYSCALL	__NR_syscalls
 #else
@@ -46,19 +42,26 @@ static const stress_help_t help[] = {
 #if defined(__linux__) &&       \
     !defined(__TINYC__) &&	\
     !defined(__PCC__) &&	\
-   (defined(__x86_64__) || defined(__x86_64))
+    defined(STRESS_ARCH_X86_64)
 #define STRESS_EXERCISE_X86_SYSCALL
 #endif
 
 #if defined(__linux__) &&       \
     defined(STRESS_ARCH_X86) &&	\
-    NEED_GNUC(9,3,0)
+    (NEED_GNUC(9, 3, 0) || NEED_CLANG(9, 0, 0))
 #define STRESS_EXERCISE_X86_0X80
 #endif
 
+#define CALL_UNDEFINED		(0)
+#define CALL_BY_SYSCALL		(1)	/* syscall() call */
+#define CALL_BY_X86_SYSCALL	(2)	/* x86 syscall opcode */
+#if defined(STRESS_EXERCISE_X86_0X80)
+#define CALL_BY_X86_INT80	(3)	/* x86 int $80 */
+#endif
+
 typedef struct hash_syscall {
 	struct hash_syscall *next;
-	unsigned long	number;
+	long	number;
 } stress_hash_syscall_t;
 
 static stress_hash_syscall_t *hash_syscall_table[HASH_SYSCALL_SIZE];
@@ -82,28 +85,27 @@ static inline long x86_64_syscall6(
 	stress_call_type = CALL_BY_X86_SYSCALL;
 
 	{
-		unsigned long _arg1 = arg1;
-		unsigned long _arg2 = arg2;
-		unsigned long _arg3 = arg3;
-		unsigned long _arg4 = arg4;
-		unsigned long _arg5 = arg5;
-		unsigned long _arg6 = arg6;
+		long tmp_arg1 = arg1;
+		long tmp_arg2 = arg2;
+		long tmp_arg3 = arg3;
+		long tmp_arg4 = arg4;
+		long tmp_arg5 = arg5;
+		long tmp_arg6 = arg6;
 
-		register long __arg1 asm ("rdi") = _arg1;
-		register long __arg2 asm ("rsi") = _arg2;
-		register long __arg3 asm ("rdx") = _arg3;
-		register long __arg4 asm ("r10") = _arg4;
-		register long __arg5 asm ("r8") = _arg5;
-		register long __arg6 asm ("r9") = _arg6;
+		register long asm_arg1 __asm__ ("rdi") = tmp_arg1;
+		register long asm_arg2 __asm__ ("rsi") = tmp_arg2;
+		register long asm_arg3 __asm__ ("rdx") = tmp_arg3;
+		register long asm_arg4 __asm__ ("r10") = tmp_arg4;
+		register long asm_arg5 __asm__ ("r8")  = tmp_arg5;
+		register long asm_arg6 __asm__ ("r9")  = tmp_arg6;
 
-
-		asm volatile ("syscall\n\t"
+		__asm__ __volatile__("syscall\n\t"
 			: "=a" (ret)
-			: "0" (number), "r" (__arg1), "r" (__arg2), "r" (__arg3),
-			  "r" (__arg4), "r" (__arg5), "r" (__arg6)
+			: "0" (number), "r" (asm_arg1), "r" (asm_arg2), "r" (asm_arg3),
+			  "r" (asm_arg4), "r" (asm_arg5), "r" (asm_arg6)
 			: "memory", "cc", "r11", "cx");
 		if (ret < 0) {
-			errno = -ret;
+			errno = -(int)ret;
 			ret = -1;
 		}
 	}
@@ -121,7 +123,7 @@ static inline int x86_0x80_syscall6(
 
 	stress_call_type = CALL_BY_X86_INT80;
 
-	asm (
+	__asm__ __volatile__(
 	     "movl %6, %%eax\n"
 	     "movl %%eax, %%ebp\n"
 	     "movl %0, %%eax\n"
@@ -227,7 +229,7 @@ try_x86_syscall:
 			goto try_x86_0x80;
 		}
 		if (stress_x86syscall_available && x86_syscall_ok) {
-			ret = x86_64_syscall6(number, arg1, arg2, arg3, arg4, arg5, arg6);
+			ret = (int)x86_64_syscall6(number, arg1, arg2, arg3, arg4, arg5, arg6);
 			exit_if_child(pid);
 			if ((ret < 0) && (errno != ENOSYS))
 				enosys = true;
@@ -316,11 +318,12 @@ static const int syscall_ignore[] = {
 #endif
 };
 
-static inline bool HOT OPTIMIZE3 syscall_find(unsigned long number)
+static inline bool HOT OPTIMIZE3 syscall_find(long number)
 {
 	register stress_hash_syscall_t *h;
 	register int i;
 	register const long number16 = number & 0xffff;
+	const unsigned long index = (unsigned long)number;
 
 	/* Really make sure some syscalls are never called */
 	for (i = 0; i < (int)SIZEOF_ARRAY(syscall_ignore); i++) {
@@ -328,7 +331,7 @@ static inline bool HOT OPTIMIZE3 syscall_find(unsigned long number)
 			return true;
 	}
 
-	h = hash_syscall_table[number % HASH_SYSCALL_SIZE];
+	h = hash_syscall_table[index % HASH_SYSCALL_SIZE];
 	while (h) {
 		if (h->number == number)
 			return true;
@@ -337,9 +340,10 @@ static inline bool HOT OPTIMIZE3 syscall_find(unsigned long number)
 	return false;
 }
 
-static inline void HOT OPTIMIZE3 syscall_add(const unsigned long number)
+static inline void HOT OPTIMIZE3 syscall_add(const long number)
 {
-	const long hash = number % HASH_SYSCALL_SIZE;
+	const unsigned long index = (unsigned long)number;
+	const unsigned long hash = index % HASH_SYSCALL_SIZE;
 	stress_hash_syscall_t *newh, *h = hash_syscall_table[hash];
 
 	while (h) {
@@ -360,10 +364,9 @@ static inline void HOT OPTIMIZE3 syscall_add(const unsigned long number)
 static inline void syscall_free(void)
 {
 	size_t i;
-	stress_hash_syscall_t *h;
 
 	for (i = 0; i < HASH_SYSCALL_SIZE; i++) {
-		h = hash_syscall_table[i];
+		stress_hash_syscall_t *h = hash_syscall_table[i];
 
 		while (h) {
 			stress_hash_syscall_t *next = h->next;
@@ -374,22 +377,26 @@ static inline void syscall_free(void)
 	}
 }
 
+
 static const int sigs[] = {
 #if defined(SIGILL)
 	SIGILL,
 #endif
-#if defined(SIGTRAP)
-	SIGTRAP,
-#endif
-#if defined(SIGFPE)
-	SIGFPE,
-#endif
 #if defined(SIGBUS)
 	SIGBUS,
 #endif
 #if defined(SIGSEGV)
 	SIGSEGV,
 #endif
+};
+
+static const int exit_sigs[] = {
+#if defined(SIGTRAP)
+	SIGTRAP,
+#endif
+#if defined(SIGFPE)
+	SIGFPE,
+#endif
 #if defined(SIGIOT)
 	SIGIOT,
 #endif
@@ -1276,9 +1283,6 @@ static const long skip_syscalls[] = {
 #if defined(SYS_sched_rr_get_interval)
 	SYS_sched_rr_get_interval,
 #endif
-#if defined(SYS_sched_rr_stress_get_interval)
-	SYS_sched_rr_stress_get_interval,
-#endif
 #if defined(SYS_sched_setaffinity)
 	SYS_sched_setaffinity,
 #endif
@@ -3478,13 +3482,13 @@ static void limit_procs(const unsigned long procs)
 #endif
 }
 
-static void NORETURN MLOCKED_TEXT stress_badhandler(int signum)
+static void NORETURN MLOCKED_TEXT stress_exit_handler(int signum)
 {
 	(void)signum;
 	_exit(1);
 }
 
-static void NORETURN MLOCKED_TEXT stress_sigill_handler(int signum)
+static void NORETURN MLOCKED_TEXT stress_sig_handler(int signum)
 {
 	(void)signum;
 
@@ -3501,7 +3505,7 @@ static void NORETURN MLOCKED_TEXT stress_sigill_handler(int signum)
  */
 static inline int stress_do_syscall(
 	const stress_args_t *args,
-	const unsigned long number,
+	const long number,
 	const bool random)
 {
 	pid_t pid;
@@ -3531,14 +3535,15 @@ static inline int stress_do_syscall(
 		if (stress_drop_capabilities(args->name) < 0) {
 			_exit(EXIT_NO_RESOURCE);
 		}
+		for (i = 0; i < SIZEOF_ARRAY(exit_sigs); i++) {
+			if (stress_sighandler(args->name, exit_sigs[i], stress_exit_handler, NULL) < 0)
+				_exit(EXIT_FAILURE);
+		}
 		for (i = 0; i < SIZEOF_ARRAY(sigs); i++) {
-			if (stress_sighandler(args->name, sigs[i], stress_badhandler, NULL) < 0)
+			if (stress_sighandler(args->name, sigs[i], stress_sig_handler, NULL) < 0)
 				_exit(EXIT_FAILURE);
 		}
-		if (stress_sighandler(args->name, SIGILL, stress_sigill_handler, NULL) < 0)
-			_exit(EXIT_FAILURE);
 
-		(void)setpgid(0, g_pgrp);
 		stress_parent_died_alarm();
 		(void)sched_settings_apply(true);
 
@@ -3579,9 +3584,7 @@ static inline int stress_do_syscall(
 		ret = waitpid(pid, &status, 0);
 		if (ret < 0) {
 			(void)kill(pid, SIGKILL);
-			ret = waitpid(pid, &status, 0);
-			(void)ret;
-
+			(void)waitpid(pid, &status, 0);
 		}
 		rc = WEXITSTATUS(status);
 
@@ -3621,7 +3624,6 @@ again:
 	} else if (pid > 0) {
 		int status, ret;
 
-		(void)setpgid(pid, g_pgrp);
 		/* Parent, wait for child */
 		ret = shim_waitpid(pid, &status, 0);
 		if (ret < 0) {
@@ -3654,7 +3656,7 @@ again:
 				}
 			}
 		}
-	} else if (pid == 0) {
+	} else {
 		const unsigned long mask = ULONG_MAX;
 		ssize_t j;
 
@@ -3662,7 +3664,6 @@ again:
 		if (!keep_stressing(args))
 			_exit(0);
 
-		(void)setpgid(0, g_pgrp);
 		stress_parent_died_alarm();
 		(void)sched_settings_apply(true);
 
@@ -3675,7 +3676,7 @@ again:
 			syscall_add(skip_syscalls[j]);
 
 		do {
-			unsigned long number;
+			long number;
 
 			/* Low sequential syscalls */
 			for (number = 0; number < MAX_SYSCALL + 1024; number++) {
@@ -3743,8 +3744,9 @@ stressor_info_t stress_enosys_info = {
 };
 #else
 stressor_info_t stress_enosys_info = {
-	.stressor = stress_not_implemented,
+	.stressor = stress_unimplemented,
 	.class = CLASS_OS,
-	.help = help
+	.help = help,
+	.unimplemented_reason = "built without sys/syscall.h or syscall support"
 };
 #endif
