@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C)      2022 Colin Ian King.
+ * Copyright (C) 2022-2023 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -34,13 +34,17 @@
 
 #define MAX_EPOLL_EVENTS 	(1024)
 #define MAX_SERVERS		(4)
+#define MIN_EPOLL_SOCKETS	(64)
+#define MAX_EPOLL_SOCKETS	(100000)
+#define DEFAULT_EPOLL_SOCKETS	(4096)
 
 static const stress_help_t help[] = {
-	{ NULL,	"epoll N",	  "start N workers doing epoll handled socket activity" },
-	{ NULL,	"epoll-ops N",	  "stop after N epoll bogo operations" },
-	{ NULL,	"epoll-port P",	  "use socket ports P upwards" },
-	{ NULL,	"epoll-domain D", "specify socket domain, default is unix" },
-	{ NULL,	NULL,		  NULL }
+	{ NULL,	"epoll N",	  	"start N workers doing epoll handled socket activity" },
+	{ NULL,	"epoll-domain D", 	"specify socket domain, default is unix" },
+	{ NULL,	"epoll-ops N",	  	"stop after N epoll bogo operations" },
+	{ NULL,	"epoll-port P",	  	"use socket ports P upwards" },
+	{ NULL, "epoll-sockets N",	"specify maximum number of open sockets" },
+	{ NULL,	NULL,			NULL }
 };
 
 #if defined(HAVE_SYS_EPOLL_H) &&	\
@@ -54,9 +58,10 @@ static const stress_help_t help[] = {
 typedef void (stress_epoll_func_t)(
 	const stress_args_t *args,
 	const int child,
-	const pid_t ppid,
+	const pid_t mypid,
 	const int epoll_port,
-	const int epoll_domain);
+	const int epoll_domain,
+	const int epoll_sockets);
 
 static timer_t epoll_timerid;
 
@@ -73,9 +78,7 @@ static int stress_set_epoll_port(const char *opt)
 	int epoll_port;
 
 	stress_set_net_port("epoll-port", opt,
-		MIN_EPOLL_PORT,
-		MAX_EPOLL_PORT - (STRESS_PROCS_MAX * MAX_SERVERS),
-		&epoll_port);
+		MIN_EPOLL_PORT, MAX_EPOLL_PORT, &epoll_port);
 	return stress_set_setting("epoll-port", TYPE_ID_INT, &epoll_port);
 }
 
@@ -83,12 +86,12 @@ static int stress_set_epoll_port(const char *opt)
  *  stress_set_epoll_domain()
  *	set the socket domain option
  */
-static int stress_set_epoll_domain(const char *name)
+static int stress_set_epoll_domain(const char *opt)
 {
 	int ret, epoll_domain;
 
 	ret = stress_set_net_domain(DOMAIN_ALL, "epoll-domain",
-		name, &epoll_domain);
+		opt, &epoll_domain);
 	stress_set_setting("epoll-domain", TYPE_ID_INT, &epoll_domain);
 
 	switch (epoll_domain) {
@@ -104,9 +107,23 @@ static int stress_set_epoll_domain(const char *name)
 	return ret;
 }
 
+/*
+ *  stress_set_epoll_sockets()
+ *	set the maximum number of open sockets
+ */
+static int stress_set_epoll_sockets(const char *opt)
+{
+	int epoll_sockets;
+
+        epoll_sockets = (int)stress_get_uint32(opt);
+        stress_check_range(opt, (uint64_t)epoll_sockets, MIN_EPOLL_SOCKETS, MAX_EPOLL_SOCKETS);
+        return stress_set_setting("epoll-sockets", TYPE_ID_INT, &epoll_sockets);
+}
+
 static const stress_opt_set_func_t opt_set_funcs[] = {
 	{ OPT_epoll_domain,	stress_set_epoll_domain },
 	{ OPT_epoll_port,	stress_set_epoll_port },
+	{ OPT_epoll_sockets,	stress_set_epoll_sockets },
 	{ 0,			NULL }
 };
 
@@ -131,7 +148,8 @@ static int stress_epoll_pwait(
 	int timeout,
 	const sigset_t *sigmask)
 {
-#if defined(__NR_epoll_pwait2)
+#if defined(__NR_epoll_pwait2) &&	\
+    defined(HAVE_SYSCALL)
 	if (stress_mwc1()) {
 		struct timespec timeout_ts;
 		int64_t timeout_ns = (int64_t)timeout * 1000;
@@ -187,9 +205,10 @@ static pid_t epoll_spawn(
 	const stress_args_t *args,
 	stress_epoll_func_t func,
 	const int child,
-	const pid_t ppid,
+	const pid_t mypid,
 	const int epoll_port,
-	const int epoll_domain)
+	const int epoll_domain,
+	const int epoll_sockets)
 {
 	pid_t pid;
 
@@ -201,13 +220,11 @@ again:
 		return -1;
 	}
 	if (pid == 0) {
-		(void)setpgid(0, g_pgrp);
 		stress_parent_died_alarm();
 		(void)sched_settings_apply(true);
-		func(args, child, ppid, epoll_port, epoll_domain);
+		func(args, child, mypid, epoll_port, epoll_domain, epoll_sockets);
 		_exit(EXIT_SUCCESS);
 	}
-	(void)setpgid(pid, g_pgrp);
 	return pid;
 }
 
@@ -230,19 +247,22 @@ static int epoll_set_fd_nonblock(const int fd)
  *  epoll_recv_data()
  *	receive data on fd
  */
-static void epoll_recv_data(const int fd)
+static void epoll_recv_data(const int fd, int *fd_count)
 {
 	while (keep_stressing_flag()) {
 		char buf[8192];
 		ssize_t n;
 
 		n = recv(fd, buf, sizeof(buf), 0);
-		if (n == -1) {
-			if (errno != EAGAIN)
+		if (n < 0) {
+			if (errno != EAGAIN) {
 				(void)close(fd);
+				(*fd_count)--;
+			}
 			break;
 		} else if (n == 0) {
 			(void)close(fd);
+			(*fd_count)--;
 			break;
 		}
 	}
@@ -306,7 +326,9 @@ static int epoll_ctl_del(const int efd, const int fd)
 static int epoll_notification(
 	const stress_args_t *args,
 	const int efd,
-	const int sfd)
+	const int sfd,
+	const int epoll_sockets,
+	int *fd_count)
 {
 	const int bad_fd = stress_get_bad_fd();
 
@@ -318,6 +340,9 @@ static int epoll_notification(
 
 		if (!keep_stressing(args))
 			return -1;
+		/* Try to limit too many open fds */
+		if (*fd_count > epoll_sockets)
+			return 0;
 
 		if ((fd = accept(sfd, &saddr, &slen)) < 0) {
 			if ((errno == EAGAIN) || (errno == EWOULDBLOCK)) {
@@ -328,6 +353,10 @@ static int epoll_notification(
 				/* out of file descriptors! */
 				break;
 			}
+			if (errno == EINTR) {
+				/* interrupted */
+				break;
+			}
 			pr_fail("%s: accept failed, errno=%d (%s)\n",
 				args->name, errno, strerror(errno));
 			return -1;
@@ -341,6 +370,7 @@ static int epoll_notification(
 			(void)close(fd);
 			return -1;
 		}
+		(*fd_count)++;
 
 #if 0
 		/*
@@ -352,7 +382,7 @@ static int epoll_notification(
 		(void)epoll_ctl(efd, EPOLL_CTL_MOD, fd, &event);
 #endif
 
-		if (epoll_ctl_add(efd, fd, EPOLLIN | EPOLLET) < 0) {
+		if (epoll_ctl_add(efd, fd, EPOLLIN) < 0) {
 			pr_fail("%s: epoll_ctl_add failed, errno=%d (%s)\n",
 				args->name, errno, strerror(errno));
 			(void)close(fd);
@@ -515,7 +545,7 @@ err:
  */
 static int epoll_client(
 	const stress_args_t *args,
-	const pid_t ppid,
+	const pid_t mypid,
 	const int epoll_port,
 	const int epoll_domain)
 {
@@ -533,7 +563,7 @@ static int epoll_client(
 		int fd, saved_errno;
 		int retries = 0;
 		int ret;
-		int port = epoll_port + port_counter +
+		const int port = epoll_port + port_counter +
 				(max_servers * (int)args->instance);
 		socklen_t addr_len = 0;
 
@@ -586,8 +616,11 @@ retry:
 			return EXIT_FAILURE;
 		}
 
-		stress_set_sockaddr(args->name, args->instance, ppid,
-			epoll_domain, port, &addr, &addr_len, NET_ADDR_ANY);
+		if (stress_set_sockaddr(args->name, args->instance, mypid,
+			epoll_domain, port, &addr, &addr_len, NET_ADDR_ANY) < 0) {
+			(void)close(fd);
+			return EXIT_FAILURE;
+		}
 
 		errno = 0;
 		ret = connect(fd, addr, addr_len);
@@ -635,8 +668,10 @@ retry:
 			break;
 		}
 		(void)close(fd);
-		(void)shim_sched_yield();
 		inc_counter(args);
+		if (!keep_stressing(args))
+			break;
+		(void)shim_sched_yield();
 	} while (keep_stressing(args));
 
 #if defined(AF_UNIX) &&		\
@@ -662,9 +697,10 @@ retry:
 static void NORETURN epoll_server(
 	const stress_args_t *args,
 	const int child,
-	const pid_t ppid,
+	const pid_t mypid,
 	const int epoll_port,
-	const int epoll_domain)
+	const int epoll_domain,
+	const int epoll_sockets)
 {
 	NOCLOBBER int efd = -1, efd2 = -1, sfd = -1, rc = EXIT_SUCCESS;
 	int so_reuseaddr = 1;
@@ -673,16 +709,12 @@ static void NORETURN epoll_server(
 	struct sockaddr *addr = NULL;
 	socklen_t addr_len = 0;
 	const int bad_fd = stress_get_bad_fd();
+	int fd_count = 0;
 
 	if (stress_sighandler(args->name, SIGSEGV, stress_segv_handler, NULL) < 0) {
 		rc = EXIT_NO_RESOURCE;
 		goto die;
 	}
-
-	if (stress_sig_stop_stressing(args->name, SIGALRM) < 0) {
-		rc = EXIT_FAILURE;
-		goto die;
-	}
 	if ((sfd = socket(epoll_domain, SOCK_STREAM, 0)) < 0) {
 		pr_fail("%s: socket failed, errno=%d (%s)\n",
 			args->name, errno, strerror(errno));
@@ -697,8 +729,11 @@ static void NORETURN epoll_server(
 		goto die_close;
 	}
 
-	stress_set_sockaddr(args->name, args->instance, ppid,
-		epoll_domain, port, &addr, &addr_len, NET_ADDR_ANY);
+	if (stress_set_sockaddr(args->name, args->instance, mypid,
+		epoll_domain, port, &addr, &addr_len, NET_ADDR_ANY) < 0) {
+		rc = EXIT_FAILURE;
+		goto die_close;
+	}
 
 	if (bind(sfd, addr, addr_len) < 0) {
 		pr_fail("%s: bind failed, errno=%d (%s)\n",
@@ -860,7 +895,7 @@ static void NORETURN epoll_server(
 			saved_errno = errno;
 
 			/* Invalid epoll_wait syscall having invalid maxevents argument */
-			(void)epoll_wait(efd, events, INT_MIN, 100);
+			(void)epoll_wait(efd, events, 0, 100);
 
 		} else {
 			if (!wait_segv) {
@@ -884,7 +919,8 @@ static void NORETURN epoll_server(
 
 		}
 		if (n < 0) {
-			if (saved_errno != EINTR) {
+			if ((saved_errno != EINTR) &&
+			    (saved_errno != EINVAL)) {
 				pr_fail("%s: epoll_wait failed, errno=%d (%s)\n",
 					args->name, saved_errno, strerror(saved_errno));
 				rc = EXIT_FAILURE;
@@ -902,12 +938,13 @@ static void NORETURN epoll_server(
 				 *  for reading anymore.. so reap fd
 				 */
 				(void)close(events[i].data.fd);
+				fd_count--;
 			} else if (sfd == events[i].data.fd) {
 				/*
 				 *  The listening socket has notification(s)
 				 *  pending, so handle incoming connections
 				 */
-				if (epoll_notification(args, efd, sfd) < 0)
+				if (epoll_notification(args, efd, sfd, epoll_sockets, &fd_count) < 0)
 					break;
 				if (test_eloop(args, efd, efd2) < 0)
 					break;
@@ -919,7 +956,7 @@ static void NORETURN epoll_server(
 				/*
 				 *  The fd has data available, so read it
 				 */
-				epoll_recv_data(events[i].data.fd);
+				epoll_recv_data(events[i].data.fd, &fd_count);
 			}
 		}
 		/*
@@ -961,23 +998,53 @@ die:
  */
 static int stress_epoll(const stress_args_t *args)
 {
-	pid_t pids[MAX_SERVERS], ppid = getppid();
+	pid_t pids[MAX_SERVERS], mypid = getpid();
 	int i, rc = EXIT_SUCCESS;
-	int epoll_port = DEFAULT_EPOLL_PORT;
 	int epoll_domain = AF_UNIX;
+	int epoll_port = DEFAULT_EPOLL_PORT;
+	int epoll_sockets = DEFAULT_EPOLL_SOCKETS;
+	int start_port, end_port, reserved_port;
 
-	(void)stress_get_setting("epoll-port", &epoll_port);
 	(void)stress_get_setting("epoll-domain", &epoll_domain);
+	(void)stress_get_setting("epoll-port", &epoll_port);
+	(void)stress_get_setting("epoll-sockets", &epoll_sockets);
+
+	if (stress_sighandler(args->name, SIGPIPE, SIG_IGN, NULL) < 0)
+		return EXIT_NO_RESOURCE;
 
 	if (max_servers == 1) {
+		start_port = epoll_port + (int)args->instance;
+		reserved_port = stress_net_reserve_ports(start_port, start_port);
+		if (reserved_port < 0) {
+			pr_inf_skip("%s: cannot reserve port %d, skipping stressor\n",
+				args->name, start_port);
+			return EXIT_NO_RESOURCE;
+		}
+		/* adjust for reserved port range */
+		start_port = reserved_port;
+		end_port = reserved_port;
+		epoll_port = start_port - (int)args->instance;
+
 		pr_dbg("%s: process [%" PRIdMAX "] using socket port %d\n",
 			args->name, (intmax_t)args->pid,
 			epoll_port + (int)args->instance);
 	} else {
+		start_port = epoll_port + (max_servers * (int)args->instance);
+		end_port = start_port + max_servers - 1;
+
+		reserved_port = stress_net_reserve_ports(start_port, end_port);
+		if (reserved_port < 0) {
+			pr_inf_skip("%s: cannot reserve ports %d..%d, skipping stressor\n",
+				args->name, start_port, end_port);
+			return EXIT_NO_RESOURCE;
+		}
+		/* adjust for reserved port range */
+		start_port = reserved_port;
+		end_port = reserved_port + max_servers - 1;
+		epoll_port = start_port - (max_servers * (int)args->instance);
+
 		pr_dbg("%s: process [%" PRIdMAX "] using socket ports %d..%d\n",
-			args->name, (intmax_t)args->pid,
-			epoll_port + (max_servers * (int)args->instance),
-			epoll_port + (max_servers * (int)(args->instance + 1)) - 1);
+			args->name, (intmax_t)args->pid, start_port, end_port);
 	}
 
 	stress_set_proc_state(args->name, STRESS_STATE_RUN);
@@ -998,7 +1065,7 @@ static int stress_epoll(const stress_args_t *args)
 	 */
 	(void)memset(pids, 0, sizeof(pids));
 	for (i = 0; i < max_servers; i++) {
-		pids[i] = epoll_spawn(args, epoll_server, i, ppid, epoll_port, epoll_domain);
+		pids[i] = epoll_spawn(args, epoll_server, i, mypid, epoll_port, epoll_domain, epoll_sockets);
 		if (pids[i] < 0) {
 			pr_fail("%s: fork failed, errno=%d (%s)\n",
 				args->name, errno, strerror(errno));
@@ -1006,15 +1073,16 @@ static int stress_epoll(const stress_args_t *args)
 		}
 	}
 
-	epoll_client(args, ppid, epoll_port, epoll_domain);
+	epoll_client(args, mypid, epoll_port, epoll_domain);
 reap:
 	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
+	stress_net_release_ports(start_port, end_port);
 
 	for (i = 0; i < max_servers; i++) {
 		int status;
 
 		if (pids[i] > 0) {
-			(void)kill(pids[i], SIGKILL);
+			(void)kill(pids[i], SIGALRM);
 			if (shim_waitpid(pids[i], &status, 0) < 0) {
 				pr_dbg("%s: waitpid failed, errno=%d (%s)\n",
 					args->name, errno, strerror(errno));
@@ -1033,9 +1101,10 @@ stressor_info_t stress_epoll_info = {
 };
 #else
 stressor_info_t stress_epoll_info = {
-	.stressor = stress_not_implemented,
+	.stressor = stress_unimplemented,
 	.class = CLASS_NETWORK | CLASS_OS,
 	.opt_set_funcs = opt_set_funcs,
-	.help = help
+	.help = help,
+	.unimplemented_reason = "built without sys/epoll.h or librt or timer support"
 };
 #endif
