@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C)      2022 Colin Ian King.
+ * Copyright (C) 2022-2023 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -531,7 +531,7 @@ static int stress_exec_child(void *arg)
 			goto do_exec;
 		}
 
-		stress_strnrnd(buffer, sizeof(buffer));
+		stress_rndbuf(buffer, sizeof(buffer));
 		if (stress_mwc1()) {
 			buffer[0] = '#';
 			buffer[1] = '!';
