@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C)      2022 Colin Ian King.
+ * Copyright (C) 2022-2023 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -18,6 +18,7 @@
  *
  */
 #include "stress-ng.h"
+#include "core-cache.h"
 
 #define DEBUG_TAG_INFO		(0)
 
@@ -30,7 +31,7 @@ static const stress_help_t help[] = {
 	{ NULL,	NULL,			NULL }
 };
 
-static int stress_l1cache_set(const char *opt, const char *name, size_t max)
+static int stress_l1cache_set(const char *opt, const char *name, const size_t max)
 {
 	uint32_t val32;
 	uint64_t val64;
@@ -74,7 +75,7 @@ static const stress_opt_set_func_t opt_set_funcs[] = {
  *  stress_l1cache_ln2()
  *	calculate log base 2
  */
-static uint32_t stress_l1cache_ln2(uint32_t val)
+static uint32_t stress_l1cache_ln2(const uint32_t val)
 {
 	int ln2 = 0;
 
@@ -91,10 +92,10 @@ static uint32_t stress_l1cache_ln2(uint32_t val)
  */
 static int stress_l1cache_info_check(
 	const stress_args_t *args,
-	uint32_t ways,
-	uint32_t size,
-	uint32_t sets,
-	uint32_t line_size)
+	const uint32_t ways,
+	const uint32_t size,
+	const uint32_t sets,
+	const uint32_t line_size)
 {
 	uint64_t sz = size ? size :
 		(uint64_t)ways * (uint64_t)sets * (uint64_t)line_size;
@@ -105,7 +106,7 @@ static int stress_l1cache_info_check(
 		stress_uint64_to_str(szstr, sizeof(szstr), sz);
 
 		pr_inf("%s: l1cache: size: %s, sets: %" PRIu32
-			", ways: %" PRIu32 ", line size: %" PRIu32 "\n",
+			", ways: %" PRIu32 ", line size: %" PRIu32 " bytes\n",
 			args->name, szstr, sets, ways, line_size);
 #if DEBUG_TAG_INFO
 		pr_inf("%s: l1cache: offset size: %" PRIu32 ", index size: %" PRIu32 "\n",
@@ -186,9 +187,8 @@ static int stress_l1cache_info_ok(
 		goto bad_cache_free;
 
 	cache = stress_get_cpu_cache(cpu_caches, 1);
-	if (!cache) {
+	if (!cache)
 		goto bad_cache_free;
-	}
 
 	if (*size == 0)
 		*size = (uint32_t)cache->size;
