@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C)      2022 Colin Ian King.
+ * Copyright (C) 2022-2023 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -21,39 +21,71 @@
 #include "core-arch.h"
 #include "core-cache.h"
 #include "core-nt-store.h"
+#include "core-pthread.h"
+#include "core-target-clones.h"
+
+#if defined(HAVE_LINUX_MEMPOLICY_H) &&	\
+    defined(__NR_mbind)
+#include <linux/mempolicy.h>
+#define HAVE_MEMTHRASH_NUMA	(1)
+#endif
 
 static const stress_help_t help[] = {
 	{ NULL,	"memthrash N",		"start N workers thrashing a 16MB memory buffer" },
-	{ NULL,	"memthrash-ops N",	"stop after N memthrash bogo operations" },
 	{ NULL,	"memthrash-method M",	"specify memthrash method M, default is all" },
+	{ NULL,	"memthrash-ops N",	"stop after N memthrash bogo operations" },
 	{ NULL,	NULL,			NULL }
 };
 
 #if defined(HAVE_LIB_PTHREAD)
 
-#define MATRIX_SIZE_MAX_SHIFT	(14)
+#define MATRIX_SIZE_MAX_SHIFT	(14)	/* No more than 16 */
 #define MATRIX_SIZE_MIN_SHIFT	(10)
 #define MATRIX_SIZE		(1 << MATRIX_SIZE_MAX_SHIFT)
 #define MEM_SIZE		(MATRIX_SIZE * MATRIX_SIZE)
+#define MEM_SIZE_PRIMES		(1 + MATRIX_SIZE_MAX_SHIFT - MATRIX_SIZE_MIN_SHIFT)
+#define STRESS_CACHE_LINE_SHIFT	(6)	/* Typical 64 byte size */
+#define STRESS_CACHE_LINE_SIZE	(1 << STRESS_CACHE_LINE_SHIFT)
 
-typedef void (*stress_memthrash_func_t)(const stress_args_t *args, size_t mem_size);
 
 typedef struct {
+	const stress_args_t *args;
+	const struct stress_memthrash_method_info *memthrash_method;
+	uint32_t total_cpus;
+	uint32_t max_threads;
+#if defined(HAVE_MEMTHRASH_NUMA)
+	unsigned long numa_nodes;
+	unsigned long max_numa_nodes;
+	unsigned long *numa_node_mask;
+	size_t numa_node_mask_size;
+#endif
+} stress_memthrash_context_t;
+
+typedef void (*stress_memthrash_func_t)(const stress_memthrash_context_t *context, size_t mem_size);
+
+typedef struct stress_memthrash_method_info {
 	const char		*name;	/* human readable form of stressor */
 	stress_memthrash_func_t	func;	/* the method function */
 } stress_memthrash_method_info_t;
 
+/* Per-pthread information */
 typedef struct {
-	uint32_t total_cpus;
-	uint32_t max_threads;
-	const stress_memthrash_method_info_t *memthrash_method;
-} stress_memthrash_context_t;
+	pthread_t pthread;	/* pthread handle */
+	int ret;		/* pthread create return value */
+} stress_pthread_info_t;
+
+typedef struct {
+	size_t	mem_size;	/* memory size */
+	size_t  prime_stride;	/* prime cache sized stride */
+} stress_memthrash_primes_t;
+
 
 static const stress_memthrash_method_info_t memthrash_methods[];
 static void *mem;
 static volatile bool thread_terminate;
 static sigset_t set;
 
+static stress_memthrash_primes_t stress_memthrash_primes[MEM_SIZE_PRIMES];
 
 #if (((defined(__GNUC__) || defined(__clang__)) && 	\
        defined(STRESS_ARCH_X86)) ||			\
@@ -88,7 +120,7 @@ static inline HOT OPTIMIZE3 void stress_memthrash_random_chunk(
 		chunks = 1;
 
 	for (i = 0; !thread_terminate && (i < max); i++) {
-		const size_t chunk = stress_mwc32() % chunks;
+		const size_t chunk = stress_mwc32modn(chunks);
 		const size_t offset = chunk * chunk_size;
 		void *ptr = (void *)(((uint8_t *)mem) + offset);
 
@@ -101,53 +133,53 @@ static inline HOT OPTIMIZE3 void stress_memthrash_random_chunk(
 }
 
 static void HOT OPTIMIZE3 stress_memthrash_random_chunkpage(
-	const stress_args_t *args,
+	const stress_memthrash_context_t *context,
 	const size_t mem_size)
 {
-	stress_memthrash_random_chunk(args->page_size, mem_size);
+	stress_memthrash_random_chunk(context->args->page_size, mem_size);
 }
 
 static void HOT OPTIMIZE3 stress_memthrash_random_chunk256(
-	const stress_args_t *args,
+	const stress_memthrash_context_t *context,
 	const size_t mem_size)
 {
-	(void)args;
+	(void)context;
 
 	stress_memthrash_random_chunk(256, mem_size);
 }
 
 static void HOT OPTIMIZE3 stress_memthrash_random_chunk64(
-	const stress_args_t *args,
+	const stress_memthrash_context_t *context,
 	const size_t mem_size)
 {
-	(void)args;
+	(void)context;
 
 	stress_memthrash_random_chunk(64, mem_size);
 }
 
 static void HOT OPTIMIZE3 stress_memthrash_random_chunk8(
-	const stress_args_t *args,
+	const stress_memthrash_context_t *context,
 	const size_t mem_size)
 {
-	(void)args;
+	(void)context;
 
 	stress_memthrash_random_chunk(8, mem_size);
 }
 
 static void HOT OPTIMIZE3 stress_memthrash_random_chunk1(
-	const stress_args_t *args,
+	const stress_memthrash_context_t *context,
 	const size_t mem_size)
 {
-	(void)args;
+	(void)context;
 
 	stress_memthrash_random_chunk(1, mem_size);
 }
 
 static void stress_memthrash_memset(
-	const stress_args_t *args,
+	const stress_memthrash_context_t *context,
 	const size_t mem_size)
 {
-	(void)args;
+	(void)context;
 
 #if defined(__GNUC__)
 	(void)__builtin_memset((void *)mem, stress_mwc8(), mem_size);
@@ -157,12 +189,12 @@ static void stress_memthrash_memset(
 }
 
 static void stress_memthrash_memmove(
-	const stress_args_t *args,
+	const stress_memthrash_context_t *context,
 	const size_t mem_size)
 {
 	char *dst = ((char *)mem) + 1;
 
-	(void)args;
+	(void)context;
 #if defined(__GNUC__)
 	(void)shim_builtin_memmove((void *)dst, mem, mem_size - 1);
 #else
@@ -171,14 +203,14 @@ static void stress_memthrash_memmove(
 }
 
 static void HOT OPTIMIZE3 stress_memthrash_memset64(
-	const stress_args_t *args,
+	const stress_memthrash_context_t *context,
 	const size_t mem_size)
 {
 	register uint64_t *ptr = (uint64_t *)mem;
-	register const uint64_t *end = (uint64_t *)(((uint8_t *)mem) + mem_size);
+	register const uint64_t *end = (uint64_t *)(((uintptr_t)mem) + mem_size);
 	register uint64_t val = stress_mwc64();
 
-	(void)args;
+	(void)context;
 
 #if defined(HAVE_NT_STORE64)
 	if (stress_cpu_x86_has_sse2()) {
@@ -210,15 +242,124 @@ static void HOT OPTIMIZE3 stress_memthrash_memset64(
 	}
 }
 
-
-static void HOT OPTIMIZE3 stress_memthrash_flip_mem(
-	const stress_args_t *args,
+static void OPTIMIZE3 TARGET_CLONES stress_memthrash_swap64(
+	const stress_memthrash_context_t *context,
 	const size_t mem_size)
 {
-	(void)args;
+	uint64_t *ptr = (uint64_t *)mem;
+	register const uint64_t *end = (uint64_t *)(((uintptr_t)mem) + mem_size);
 
+	(void)context;
+
+	while (LIKELY(ptr < end)) {
+		register uint64_t r0, r1, r2, r3, r4, r5, r6, r7;
+
+		r0 = ptr[0];
+		r1 = ptr[1];
+		r2 = ptr[2];
+		r3 = ptr[3];
+		r4 = ptr[4];
+		r5 = ptr[5];
+		r6 = ptr[6];
+		r7 = ptr[7];
+		shim_mb();
+
+		ptr[0] = r4;
+		ptr[1] = r5;
+		ptr[2] = r6;
+		ptr[3] = r7;
+		ptr[4] = r0;
+		ptr[5] = r1;
+		ptr[6] = r2;
+		ptr[7] = r3;
+		shim_mb();
+		ptr += 8;
+
+		r0 = ptr[0];
+		r1 = ptr[1];
+		r2 = ptr[2];
+		r3 = ptr[3];
+		r4 = ptr[4];
+		r5 = ptr[5];
+		r6 = ptr[6];
+		r7 = ptr[7];
+		shim_mb();
+
+		ptr[0] = r4;
+		ptr[1] = r5;
+		ptr[2] = r6;
+		ptr[3] = r7;
+		ptr[4] = r0;
+		ptr[5] = r1;
+		ptr[6] = r2;
+		ptr[7] = r3;
+		shim_mb();
+		ptr += 8;
+	}
+}
+
+#if defined(HAVE_INT128_T)
+static void OPTIMIZE3 TARGET_CLONES stress_memthrash_copy128(
+	const stress_memthrash_context_t *context,
+	const size_t mem_size)
+{
+	__uint128_t *ptr = (__uint128_t *)mem;
+	register const __uint128_t *end = (__uint128_t *)(((uintptr_t)mem) + mem_size - 128);
+
+	(void)context;
+
+	while (LIKELY(ptr < end)) {
+		register __uint128_t r0, r1, r2, r3, r4, r5, r6, r7;
+
+		r0 = ptr[8];
+		r1 = ptr[9];
+		r2 = ptr[10];
+		r3 = ptr[11];
+		r4 = ptr[12];
+		r5 = ptr[13];
+		r6 = ptr[14];
+		r7 = ptr[15];
+		ptr[0] = r0;
+		ptr[1] = r1;
+		ptr[2] = r2;
+		ptr[3] = r3;
+		ptr[4] = r4;
+		ptr[5] = r5;
+		ptr[6] = r6;
+		ptr[7] = r7;
+		shim_mb();
+		ptr += 8;
+
+		r0 = ptr[8];
+		r1 = ptr[9];
+		r2 = ptr[10];
+		r3 = ptr[11];
+		r4 = ptr[12];
+		r5 = ptr[13];
+		r6 = ptr[14];
+		r7 = ptr[15];
+		ptr[0] = r0;
+		ptr[1] = r1;
+		ptr[2] = r2;
+		ptr[3] = r3;
+		ptr[4] = r4;
+		ptr[5] = r5;
+		ptr[6] = r6;
+		ptr[7] = r7;
+		shim_mb();
+		ptr += 8;
+	}
+}
+#endif
+
+static void HOT OPTIMIZE3 stress_memthrash_flip_mem(
+	const stress_memthrash_context_t *context,
+	const size_t mem_size)
+{
 	volatile uint64_t *ptr = (volatile uint64_t *)mem;
-	const uint64_t *end = (uint64_t *)(((uint8_t *)mem) + mem_size);
+	const uint64_t *end = (uint64_t *)(((uintptr_t)mem) + mem_size);
+
+	(void)context;
 
 	while (LIKELY(ptr < end)) {
 		*ptr = *ptr ^ ~0ULL;
@@ -227,15 +368,15 @@ static void HOT OPTIMIZE3 stress_memthrash_flip_mem(
 }
 
 static void HOT OPTIMIZE3 stress_memthrash_swap(
-	const stress_args_t *args,
+	const stress_memthrash_context_t *context,
 	const size_t mem_size)
 {
 	size_t i;
-	register size_t offset1 = stress_mwc32() % mem_size;
-	register size_t offset2 = stress_mwc32() % mem_size;
+	register size_t offset1 = stress_mwc32modn(mem_size);
+	register size_t offset2 = stress_mwc32modn(mem_size);
 	uint8_t *mem_u8 = (uint8_t *)mem;
 
-	(void)args;
+	(void)context;
 
 	for (i = 0; !thread_terminate && (i < 65536); i++) {
 		register uint8_t tmp;
@@ -254,17 +395,17 @@ static void HOT OPTIMIZE3 stress_memthrash_swap(
 }
 
 static void HOT OPTIMIZE3 stress_memthrash_matrix(
-	const stress_args_t *args,
+	const stress_memthrash_context_t *context,
 	const size_t mem_size)
 {
-	(void)args;
-	(void)mem_size;
-
 	size_t i, j;
 	volatile uint8_t *vmem = mem;
 
-	for (i = 0; !thread_terminate && (i < MATRIX_SIZE); i+= ((stress_mwc8() & 0xf) + 1)) {
-		for (j = 0; j < MATRIX_SIZE; j+= 16) {
+	(void)context;
+	(void)mem_size;
+
+	for (i = 0; !thread_terminate && (i < MATRIX_SIZE); i += ((stress_mwc8() & 0xf) + 1)) {
+		for (j = 0; j < MATRIX_SIZE; j += 16) {
 			size_t i1 = (i * MATRIX_SIZE) + j;
 			size_t i2 = (j * MATRIX_SIZE) + i;
 			uint8_t tmp;
@@ -277,37 +418,37 @@ static void HOT OPTIMIZE3 stress_memthrash_matrix(
 }
 
 static void HOT OPTIMIZE3 stress_memthrash_prefetch(
-	const stress_args_t *args,
+	const stress_memthrash_context_t *context,
 	const size_t mem_size)
 {
 	uint32_t i;
 	const uint32_t max = stress_mwc16();
 
-	(void)args;
+	(void)context;
 
 	for (i = 0; !thread_terminate && (i < max); i++) {
-		size_t offset = stress_mwc32() % mem_size;
+		size_t offset = stress_mwc32modn(mem_size);
 		uint8_t *const ptr = ((uint8_t *)mem) + offset;
 		volatile uint8_t *const vptr = ptr;
 
+		/* Force prefetch and then modify to thrash cache */
 		shim_builtin_prefetch(ptr, 1, 1);
-		//(void)*vptr;
 		*vptr = i & 0xff;
 	}
 }
 
 #if defined(HAVE_ASM_X86_CLFLUSH)
 static void HOT OPTIMIZE3 stress_memthrash_flush(
-	const stress_args_t *args,
+	const stress_memthrash_context_t *context,
 	const size_t mem_size)
 {
 	uint32_t i;
 	const uint32_t max = stress_mwc16();
 
-	(void)args;
+	(void)context;
 
 	for (i = 0; !thread_terminate && (i < max); i++) {
-		size_t offset = stress_mwc32() % mem_size;
+		size_t offset = stress_mwc32modn(mem_size);
 		uint8_t *const ptr = ((uint8_t *)mem) + offset;
 		volatile uint8_t *const vptr = ptr;
 
@@ -318,16 +459,16 @@ static void HOT OPTIMIZE3 stress_memthrash_flush(
 #endif
 
 static void HOT OPTIMIZE3 stress_memthrash_mfence(
-	const stress_args_t *args,
+	const stress_memthrash_context_t *context,
 	const size_t mem_size)
 {
 	uint32_t i;
 	const uint32_t max = stress_mwc16();
 
-	(void)args;
+	(void)context;
 
 	for (i = 0; !thread_terminate && (i < max); i++) {
-		size_t offset = stress_mwc32() % mem_size;
+		size_t offset = stress_mwc32modn(mem_size);
 		volatile uint8_t *ptr = ((uint8_t *)mem) + offset;
 
 		*ptr = i & 0xff;
@@ -337,15 +478,15 @@ static void HOT OPTIMIZE3 stress_memthrash_mfence(
 
 #if defined(MEM_LOCK)
 static void HOT OPTIMIZE3 stress_memthrash_lock(
-	const stress_args_t *args,
+	const stress_memthrash_context_t *context,
 	const size_t mem_size)
 {
 	uint32_t i;
 
-	(void)args;
+	(void)context;
 
 	for (i = 0; !thread_terminate && (i < 64); i++) {
-		size_t offset = stress_mwc32() % mem_size;
+		size_t offset = stress_mwc32modn(mem_size);
 		volatile uint8_t *ptr = ((uint8_t *)mem) + offset;
 
 		MEM_LOCK(ptr, 1);
@@ -354,16 +495,16 @@ static void HOT OPTIMIZE3 stress_memthrash_lock(
 #endif
 
 static void HOT OPTIMIZE3 stress_memthrash_spinread(
-	const stress_args_t *args,
+	const stress_memthrash_context_t *context,
 	const size_t mem_size)
 {
 	uint32_t i;
 	volatile uint32_t *ptr;
 	const size_t size = mem_size - (8 * sizeof(*ptr));
-	const size_t offset = (stress_mwc32() % size) & ~(size_t)3;
+	const size_t offset = stress_mwc32modn(size) & ~(size_t)3;
 
-	ptr = (uint32_t *)(((uint8_t *)mem) + offset);
-	(void)args;
+	(void)context;
+	ptr = (uint32_t *)(((uintptr_t)mem) + offset);
 
 	for (i = 0; !thread_terminate && (i < 65536); i++) {
 		(void)*ptr;
@@ -379,16 +520,16 @@ static void HOT OPTIMIZE3 stress_memthrash_spinread(
 }
 
 static void HOT OPTIMIZE3 stress_memthrash_spinwrite(
-	const stress_args_t *args,
+	const stress_memthrash_context_t *context,
 	const size_t mem_size)
 {
 	uint32_t i;
 	volatile uint32_t *ptr;
 	const size_t size = mem_size - (8 * sizeof(*ptr));
-	const size_t offset = (stress_mwc32() % size) & ~(size_t)3;
+	const size_t offset = stress_mwc32modn(size) & ~(size_t)3;
 
-	ptr = (uint32_t *)(((uint8_t *)mem) + offset);
-	(void)args;
+	(void)context;
+	ptr = (uint32_t *)(((uintptr_t)mem) + offset);
 
 	for (i = 0; !thread_terminate && (i < 65536); i++) {
 		*ptr = i;
@@ -403,9 +544,95 @@ static void HOT OPTIMIZE3 stress_memthrash_spinwrite(
 	}
 }
 
+static void HOT OPTIMIZE3 stress_memthrash_tlb(
+	const stress_memthrash_context_t *context,
+	const size_t mem_size)
+{
+	const size_t cache_lines = mem_size >> STRESS_CACHE_LINE_SHIFT;
+	const size_t mask = mem_size - 1;		/* assuming mem_size is a power of 2 */
+	const size_t offset = (size_t)stress_mwc16() & (STRESS_CACHE_LINE_SIZE - 1);
+	size_t prime_stride = 65537 * STRESS_CACHE_LINE_SIZE;	/* prime default */
+	register int i;
+	volatile uint8_t *ptr;
+	register size_t j, k;
 
-static void stress_memthrash_all(const stress_args_t *args, size_t mem_size);
-static void stress_memthrash_random(const stress_args_t *args, size_t mem_size);
+	(void)context;
+
+	/* Find size of stride for the given memory size */
+	for (i = 0; i < MEM_SIZE_PRIMES; i++) {
+		if (mem_size == stress_memthrash_primes[i].mem_size) {
+			prime_stride = stress_memthrash_primes[i].prime_stride;
+			break;
+		}
+	}
+
+	/* Stride around memory in prime cache line strides, reads */
+	for (j = 0, k = offset; j < cache_lines; j++) {
+		ptr = (volatile uint8_t *)mem + k;
+		(void)*ptr;
+		k = (k + prime_stride) & mask;
+	}
+	/* Stride around memory in prime cache line strides, writes */
+	for (j = 0, k = offset; j < cache_lines; j++) {
+		ptr = (volatile uint8_t *)mem + k;
+		*ptr = j;
+		k = (k + prime_stride) & mask;
+	}
+}
+
+static void OPTIMIZE3 TARGET_CLONES stress_memthrash_swapfwdrev(
+	const stress_memthrash_context_t *context,
+	const size_t mem_size)
+{
+	uint64_t *fwd, *rev, *end = (uint64_t *)((uintptr_t)mem + mem_size);
+
+	(void)context;
+	for (fwd = (uint64_t *)mem, rev = end - 1; fwd < end; rev--, fwd++) {
+		register uint64_t tmp;
+
+		tmp = *fwd;
+		*fwd = *rev;
+		*rev = tmp;
+	}
+	for (fwd = (uint64_t *)mem, rev = end - 1; fwd < end; rev--, fwd++) {
+		register uint64_t tmp;
+
+		tmp = *rev;
+		*rev = *fwd;
+		*fwd = tmp;
+	}
+}
+
+#if defined(HAVE_MEMTHRASH_NUMA)
+static void OPTIMIZE3 TARGET_CLONES stress_memthrash_numa(
+	const stress_memthrash_context_t *context,
+	const size_t mem_size)
+{
+	uint8_t *ptr;
+	const uint8_t *end = (uint8_t *)((uintptr_t)mem + mem_size);
+	const size_t page_size = context->args->page_size;
+	unsigned long node;
+
+	if (context->numa_nodes < 1)
+		return;
+
+	node = (unsigned long)stress_mwc32modn(context->numa_nodes);
+	(void)memset(context->numa_node_mask, 0, context->numa_node_mask_size);
+
+	for (ptr = (uint8_t *)mem; ptr < end; ptr += page_size) {
+		STRESS_SETBIT(context->numa_node_mask, node);
+
+		(void)shim_mbind((void *)ptr, page_size, MPOL_PREFERRED, context->numa_node_mask, context->max_numa_nodes, 0);
+		STRESS_CLRBIT(context->numa_node_mask, node);
+		node++;
+		if (node >= context->numa_nodes)
+			node = 0;
+	}
+}
+#endif
+
+static void stress_memthrash_all(const stress_memthrash_context_t *context, size_t mem_size);
+static void stress_memthrash_random(const stress_memthrash_context_t *context, size_t mem_size);
 
 static const stress_memthrash_method_info_t memthrash_methods[] = {
 	{ "all",	stress_memthrash_all },		/* MUST always be first! */
@@ -415,6 +642,9 @@ static const stress_memthrash_method_info_t memthrash_methods[] = {
 	{ "chunk64",	stress_memthrash_random_chunk64 },
 	{ "chunk256",	stress_memthrash_random_chunk256 },
 	{ "chunkpage",	stress_memthrash_random_chunkpage },
+#if defined(HAVE_INT128_T)
+	{ "copy128",	stress_memthrash_copy128 },
+#endif
 	{ "flip",	stress_memthrash_flip_mem },
 #if defined(HAVE_ASM_X86_CLFLUSH)
 	{ "flush",	stress_memthrash_flush },
@@ -427,20 +657,26 @@ static const stress_memthrash_method_info_t memthrash_methods[] = {
 	{ "memset",	stress_memthrash_memset },
 	{ "memset64",	stress_memthrash_memset64 },
 	{ "mfence",	stress_memthrash_mfence },
+#if defined(HAVE_MEMTHRASH_NUMA)
+	{ "numa",	stress_memthrash_numa },
+#endif
 	{ "prefetch",	stress_memthrash_prefetch },
 	{ "random",	stress_memthrash_random },
 	{ "spinread",	stress_memthrash_spinread },
 	{ "spinwrite",	stress_memthrash_spinwrite },
-	{ "swap",	stress_memthrash_swap }
+	{ "swap",	stress_memthrash_swap },
+	{ "swap64",	stress_memthrash_swap64 },
+	{ "swapfwdrev",	stress_memthrash_swapfwdrev },
+	{ "tlb",	stress_memthrash_tlb },
 };
 
-static void stress_memthrash_all(const stress_args_t *args, size_t mem_size)
+static void stress_memthrash_all(const stress_memthrash_context_t *context, size_t mem_size)
 {
 	static size_t i = 1;
 	const double t = stress_time_now();
 
 	do {
-		memthrash_methods[i].func(args, mem_size);
+		memthrash_methods[i].func(context, mem_size);
 	} while (!thread_terminate && (stress_time_now() - t < 0.01));
 
 	i++;
@@ -448,17 +684,17 @@ static void stress_memthrash_all(const stress_args_t *args, size_t mem_size)
 		i = 1;
 }
 
-static void stress_memthrash_random(const stress_args_t *args, size_t mem_size)
+static void stress_memthrash_random(const stress_memthrash_context_t *context, size_t mem_size)
 {
 	/* loop until we find a good candidate */
 	for (;;) {
-		size_t i = stress_mwc8() % SIZEOF_ARRAY(memthrash_methods);
+		size_t i = stress_mwc8modn((uint8_t)SIZEOF_ARRAY(memthrash_methods));
 		const stress_memthrash_func_t func = (stress_memthrash_func_t)memthrash_methods[i].func;
 
 		/* Don't run stress_memthrash_random/all to avoid recursion */
 		if ((func != stress_memthrash_random) &&
 		    (func != stress_memthrash_all)) {
-			func(args, mem_size);
+			func(context, mem_size);
 			return;
 		}
 	}
@@ -489,16 +725,29 @@ static int stress_set_memthrash_method(const char *name)
 	return -1;
 }
 
+static void stress_memthrash_find_primes(void)
+{
+	size_t i;
+
+	for (i = 0; i < MEM_SIZE_PRIMES; i++) {
+		const size_t mem_size = 1 << (2 * (i + MATRIX_SIZE_MIN_SHIFT));
+		const size_t cache_lines = (mem_size / STRESS_CACHE_LINE_SIZE) + 137;
+
+		stress_memthrash_primes[i].mem_size = mem_size;
+		stress_memthrash_primes[i].prime_stride =
+			(size_t)stress_get_prime64((uint64_t)cache_lines) * STRESS_CACHE_LINE_SIZE;
+	}
+}
+
 /*
  *  stress_memthrash_func()
- *	pthread that exits immediately
  */
-static void *stress_memthrash_func(void *arg)
+static void *stress_memthrash_func(void *ctxt)
 {
 	static void *nowt = NULL;
-	const stress_pthread_args_t *parg = (stress_pthread_args_t *)arg;
-	const stress_args_t *args = parg->args;
-	const stress_memthrash_func_t func = (stress_memthrash_func_t)parg->data;
+	const stress_memthrash_context_t *context = (stress_memthrash_context_t *)ctxt;
+	const stress_memthrash_func_t func = context->memthrash_method->func;
+	const stress_args_t *args = context->args;
 
 	/*
 	 *  Block all signals, let controlling thread
@@ -512,12 +761,12 @@ static void *stress_memthrash_func(void *arg)
 		for (j = MATRIX_SIZE_MIN_SHIFT; j <= MATRIX_SIZE_MAX_SHIFT &&
 		     !thread_terminate && keep_stressing(args); j++) {
 			size_t mem_size = 1 << (2 * j);
-
 			size_t i;
+
 			for (i = 0; i < SIZEOF_ARRAY(memthrash_methods); i++)
 				if (func == memthrash_methods[i].func)
 					break;
-			func(args, mem_size);
+			func(context, mem_size);
 			inc_counter(args);
 			shim_sched_yield();
 		}
@@ -572,23 +821,23 @@ static int stress_memthrash_child(const stress_args_t *args, void *ctxt)
 	stress_memthrash_context_t *context = (stress_memthrash_context_t *)ctxt;
 	const uint32_t max_threads = context->max_threads;
 	uint32_t i;
-	pthread_t pthreads[max_threads];
-	int pthreads_ret[max_threads], ret;
-	stress_pthread_args_t pargs;
+	int ret;
+	stress_pthread_info_t *pthread_info;
 
 	int flags = MAP_PRIVATE | MAP_ANONYMOUS;
 #if defined(MAP_POPULATE)
 	flags |= MAP_POPULATE;
 #endif
 
-	ret = stress_sighandler(args->name, SIGALRM, stress_memthrash_sigalrm_handler, NULL);
-	(void)ret;
+	pthread_info = calloc(max_threads, sizeof(*pthread_info));
+	if (!pthread_info) {
+		pr_inf_skip("%s: failed to allocate pthread information array, skipping stressor\n",
+			args->name);
+		return EXIT_NO_RESOURCE;
+	}
 
-	pargs.args = args;
-	pargs.data = (void *)context->memthrash_method->func;
+	VOID_RET(int, stress_sighandler(args->name, SIGALRM, stress_memthrash_sigalrm_handler, NULL));
 
-	(void)memset(pthreads, 0, sizeof(pthreads));
-	(void)memset(pthreads_ret, 0, sizeof(pthreads_ret));
 
 mmap_retry:
 	mem = mmap(NULL, MEM_SIZE, PROT_READ | PROT_WRITE, flags, -1, 0);
@@ -599,6 +848,7 @@ mmap_retry:
 		if (!keep_stressing_flag()) {
 			pr_dbg("%s: mmap failed: %d %s\n",
 				args->name, errno, strerror(errno));
+			free(pthread_info);
 			return EXIT_NO_RESOURCE;
 		}
 		(void)shim_usleep(100000);
@@ -608,15 +858,18 @@ mmap_retry:
 	}
 
 	for (i = 0; i < max_threads; i++) {
-		pthreads_ret[i] = pthread_create(&pthreads[i], NULL,
-				stress_memthrash_func, (void *)&pargs);
-		if (pthreads_ret[i]) {
+		pthread_info[i].ret = pthread_create(&pthread_info[i].pthread,
+						NULL, stress_memthrash_func,
+						(void *)context);
+		if (pthread_info[i].ret) {
+			ret = pthread_info[i].ret;
+
 			/* Just give up and go to next thread */
-			if (pthreads_ret[i] == EAGAIN)
+			if (ret == EAGAIN)
 				continue;
 			/* Something really unexpected */
 			pr_fail("%s: pthread create failed, errno=%d (%s)\n",
-				args->name, pthreads_ret[i], strerror(pthreads_ret[i]));
+				args->name, ret, strerror(ret));
 			goto reap;
 		}
 		if (!keep_stressing_flag())
@@ -628,16 +881,17 @@ mmap_retry:
 reap:
 	thread_terminate = true;
 	for (i = 0; i < max_threads; i++) {
-		if (!pthreads_ret[i]) {
-			pthreads_ret[i] = pthread_join(pthreads[i], NULL);
-			if (pthreads_ret[i] && (pthreads_ret[i] != ESRCH)) {
+		if (!pthread_info[i].ret) {
+			pthread_info[i].ret = pthread_join(pthread_info[i].pthread, NULL);
+			if (pthread_info[i].ret && (pthread_info[i].ret != ESRCH)) {
 				pr_fail("%s: pthread join failed, errno=%d (%s)\n",
-					args->name, pthreads_ret[i], strerror(pthreads_ret[i]));
+					args->name, pthread_info[i].ret, strerror(pthread_info[i].ret));
 			}
 		}
 	}
 reap_mem:
 	(void)munmap(mem, MEM_SIZE);
+	free(pthread_info);
 
 	return EXIT_SUCCESS;
 }
@@ -652,14 +906,35 @@ static int stress_memthrash(const stress_args_t *args)
 	stress_memthrash_context_t context;
 	int rc;
 
+	stress_memthrash_find_primes();
+
+	context.args = args;
 	context.total_cpus = (uint32_t)stress_get_processors_online();
-	context.max_threads = stress_memthrash_max(args->num_instances, context.total_cpus);
 	context.memthrash_method = &memthrash_methods[0];
+	context.max_threads = stress_memthrash_max(args->num_instances, context.total_cpus);
+#if defined(HAVE_MEMTHRASH_NUMA)
+	{
+		const size_t numa_bits = sizeof(unsigned long) * 8;
+		size_t numa_elements;
+
+		context.numa_nodes = stress_numa_count_mem_nodes(&context.max_numa_nodes);
+
+		numa_elements = (context.max_numa_nodes + numa_bits - 1) / numa_bits;
+		numa_elements = numa_elements ? numa_elements: 1;
+		context.numa_node_mask = calloc(context.max_numa_nodes, numa_elements);
+		context.numa_node_mask_size = (size_t)context.max_numa_nodes * numa_elements;
+		if (!context.numa_node_mask) {
+			pr_inf_skip("%s: could not allocate %zd numa elements in numa mask, skipping stressor\n",
+				args->name, numa_elements);
+			return EXIT_NO_RESOURCE;
+		}
+	}
+#endif
 
 	(void)stress_get_setting("memthrash-method", &context.memthrash_method);
 
-	pr_dbg("%s: using method '%s'\n", args->name, context.memthrash_method->name);
 	if (args->instance == 0) {
+		pr_dbg("%s: using method '%s'\n", args->name, context.memthrash_method->name);
 		pr_inf("%s: starting %" PRIu32 " thread%s on each of the %"
 			PRIu32 " stressors on a %" PRIu32 " CPU system\n",
 			args->name, context.max_threads, plural(context.max_threads),
@@ -679,6 +954,9 @@ static int stress_memthrash(const stress_args_t *args)
 	rc = stress_oomable_child(args, &context, stress_memthrash_child, STRESS_OOMABLE_NORMAL);
 
 	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
+#if defined(HAVE_MEMTHRASH_NUMA)
+	free(context.numa_node_mask);
+#endif
 
 	return rc;
 }
@@ -710,9 +988,10 @@ static const stress_opt_set_func_t opt_set_funcs[] = {
 };
 
 stressor_info_t stress_memthrash_info = {
-	.stressor = stress_not_implemented,
+	.stressor = stress_unimplemented,
 	.class = CLASS_MEMORY,
 	.opt_set_funcs = opt_set_funcs,
-	.help = help
+	.help = help,
+	.unimplemented_reason = "built without pthread support"
 };
 #endif
