@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C)      2022 Colin Ian King.
+ * Copyright (C) 2022-2023 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -27,13 +27,13 @@
 #include <utime.h>
 #endif
 
-typedef int (*stress_open_func_t)(void);
+typedef int (*stress_open_func_t)(const stress_args_t *args, const pid_t pid, double *duration, double *count);
 
 static const stress_help_t help[] = {
 	{ "o N", "open N",		"start N workers exercising open/close" },
-	{ NULL,	"open-ops N",		"stop after N open/close bogo operations" },
 	{ NULL, "open-fd",		"open files in /proc/$pid/fd" },
 	{ NULL,	"open-max N",		"specficify maximum number of files to open" },
+	{ NULL,	"open-ops N",		"stop after N open/close bogo operations" },
 	{ NULL,	NULL,			NULL }
 };
 
@@ -263,26 +263,42 @@ static int obsolete_futimes(const int fd, const struct timeval tv[2])
 	return ret;
 }
 
-static inline int open_arg2(const char *pathname, const int flags)
+static inline int open_arg2(
+	const char *pathname,
+	const int flags,
+	double *duration,
+	double *count)
 {
 	int fd;
+	double t;
 
+	t = stress_time_now();
 #if defined(__NR_open) &&	\
     defined(HAVE_SYSCALL)
 	fd = (int)syscall(__NR_open, pathname, flags);
 #else
 	fd = open(pathname, flags);
 #endif
-	if (fd >= 0)
+	if (fd >= 0) {
+		(*duration) += stress_time_now() - t;
+		(*count) += 1.0;
 		(void)obsolete_futimes(fd, NULL);
+	}
 
 	return fd;
 }
 
-static inline int open_arg3(const char *pathname, const int flags, const mode_t mode)
+static inline int open_arg3(
+	const char *pathname,
+	const int flags,
+	const int mode,
+	double *duration,
+	double *count)
 {
 	int fd;
+	double t;
 
+	t = stress_time_now();
 #if defined(__NR_open) &&	\
     defined(HAVE_SYSCALL)
 	fd = (int)syscall(__NR_open, pathname, flags, mode);
@@ -293,6 +309,9 @@ static inline int open_arg3(const char *pathname, const int flags, const mode_t
 	if (fd >= 0) {
 		struct timeval tv[2];
 
+		(*duration) += stress_time_now() - t;
+		(*count) += 1.0;
+
 		/* Exercise illegal futimes, usec too small */
 		tv[0].tv_usec = -1;
 		tv[0].tv_sec = -1;
@@ -317,7 +336,12 @@ static inline int open_arg3(const char *pathname, const int flags, const mode_t
 	return fd;
 }
 
-static int open_flag_perm(void)
+#if defined(O_CREAT)
+static int open_flag_perm(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
 {
 	static size_t index = 0;
 	int fd;
@@ -325,11 +349,13 @@ static int open_flag_perm(void)
 	const int flags = open_perms[index];
 	char filename[PATH_MAX];
 
-	(void)snprintf(filename, sizeof(filename), "stress-open-%d-%" PRIu32,
-		(int)getpid(), stress_mwc32());
+	(void)args;
+
+	(void)snprintf(filename, sizeof(filename), "stress-open-%" PRIdMAX "-%" PRIu32,
+		(intmax_t)pid, stress_mwc32());
 
 	if (UNLIKELY((open_count == 0) || (!open_perms))) {
-		fd = open_arg3(filename, O_CREAT | O_RDWR, mode);
+		fd = open_arg3(filename, O_CREAT | O_RDWR, mode, duration, count);
 		(void)shim_unlink(filename);
 		return fd;
 	}
@@ -340,18 +366,18 @@ static int open_flag_perm(void)
 		if (flags & O_DIRECTORY) {
 			(void)mkdir(filename);
 		} else
-			fd = open_arg3(filename, O_CREAT | O_RDWR, mode);
+			fd = open_arg3(filename, O_CREAT | O_RDWR, mode, duration, count);
 			if (fd >= 0)
 				(void)close(fd);
 		}
 	}
 #else
-	fd = open_arg3(filename, O_CREAT | O_RDWR, mode);
+	fd = open_arg3(filename, O_CREAT | O_RDWR, mode, duration, count);
 	if (fd >= 0)
 		void)close(fd);
 #endif
 #endif
-	fd = open_arg3(filename, flags, mode);
+	fd = open_arg3(filename, flags, mode, duration, count);
 #if defined(O_DIRECTORY)
 	if (flags & O_DIRECTORY)
 		(void)shim_rmdir(filename);
@@ -362,11 +388,19 @@ static int open_flag_perm(void)
 
 	return fd;
 }
+#endif
 
-static int open_dev_zero_rd(void)
+static int open_dev_zero_rd(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
 {
 	int flags = 0;
 
+	(void)args;
+	(void)pid;
+
 #if defined(O_ASYNC)
 	flags |= O_ASYNC;
 #endif
@@ -388,13 +422,20 @@ static int open_dev_zero_rd(void)
 	flags &= stress_mwc32();
 	flags |= O_RDONLY;
 
-	return open_arg2("/dev/zero", flags);
+	return open_arg2("/dev/zero", flags, duration, count);
 }
 
-static int open_dev_null_wr(void)
+static int open_dev_null_wr(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
 {
 	int flags = 0;
 
+	(void)args;
+	(void)pid;
+
 #if defined(O_ASYNC)
 	flags |= O_ASYNC;
 #endif
@@ -419,14 +460,21 @@ static int open_dev_null_wr(void)
 	flags &= stress_mwc32();
 	flags |= O_WRONLY;
 
-	return open_arg2("/dev/null", flags);
+	return open_arg2("/dev/null", flags, duration, count);
 }
 
 #if defined(O_TMPFILE)
-static int open_tmp_rdwr(void)
+static int open_tmp_rdwr(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
 {
 	int flags = 0;
 
+	(void)args;
+	(void)pid;
+
 #if defined(O_TRUNC)
 	flags |= O_TRUNC;
 #endif
@@ -441,64 +489,164 @@ static int open_tmp_rdwr(void)
 #endif
 	flags &= stress_mwc32();
 	flags |= O_TMPFILE | O_RDWR;
-	return open_arg3("/tmp", flags, S_IRUSR | S_IWUSR);
+	return open_arg3("/tmp", flags, S_IRUSR | S_IWUSR, duration, count);
 }
 #endif
 
 #if defined(O_TMPFILE)
-static int open_tmpfile_no_rdwr(void)
+static int open_tmpfile_no_rdwr(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
 {
+	(void)args;
+	(void)pid;
+
 	/* Force -EINVAL, need O_WRONLY or O_RDWR to succeed */
-	return open_arg3("/tmp", O_TMPFILE, S_IRUSR | S_IWUSR);
+	return open_arg3("/tmp", O_TMPFILE, S_IRUSR | S_IWUSR, duration, count);
 }
 #endif
 
 #if defined(HAVE_POSIX_OPENPT) &&	\
     defined(O_RDWR) &&			\
     defined(N_NOCTTY)
-static int open_pt(void)
+static int open_pt(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
 {
+	(void)args;
+	(void)pid;
+
 	return posix_openpt(O_RDWR | O_NOCTTY);
 }
 #endif
 
 #if defined(O_TMPFILE) &&	\
     defined(O_EXCL)
-static int open_tmp_rdwr_excl(void)
+static int open_tmp_rdwr_excl(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
 {
-	return open_arg3("/tmp", O_TMPFILE | O_EXCL | O_RDWR, S_IRUSR | S_IWUSR);
+	(void)args;
+	(void)pid;
+
+	return open_arg3("/tmp", O_TMPFILE | O_EXCL | O_RDWR, S_IRUSR | S_IWUSR, duration, count);
 }
 #endif
 
 #if defined(O_DIRECTORY)
-static int open_dir(void)
+static int open_dir(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
 {
-	return open_arg2(".", O_DIRECTORY | O_RDONLY);
+	(void)args;
+	(void)pid;
+
+	return open_arg2(".", O_DIRECTORY | O_RDONLY, duration, count);
+}
+#endif
+
+#if defined(O_DIRECTORY) &&	\
+    defined(__linux__)
+static int open_dir_proc_self_fd(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
+{
+	(void)args;
+	(void)pid;
+
+	return open_arg2("/proc/self/fd", O_DIRECTORY | O_RDONLY, duration, count);
 }
 #endif
 
 #if defined(O_PATH)
-static int open_path(void)
+static int open_path(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
 {
-	return open_arg2(".", O_DIRECTORY | O_PATH);
+	(void)args;
+	(void)pid;
+
+	return open_arg2(".", O_DIRECTORY | O_PATH, duration, count);
 }
 #endif
 
 #if defined(O_CREAT)
-static int open_create_eisdir(void)
+static int open_create_eisdir(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
 {
-	return open_arg3(".", O_CREAT, S_IRUSR | S_IWUSR);
+	(void)args;
+	(void)pid;
+
+	return open_arg3(".", O_CREAT, S_IRUSR | S_IWUSR, duration, count);
+}
+#endif
+
+#if defined(O_DIRECT) &&	\
+    defined(O_CREAT)
+static int open_direct(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
+{
+	char filename[PATH_MAX];
+	int fd;
+	double t;
+
+	(void)snprintf(filename, sizeof(filename), "stress-open-%" PRIdMAX "-%" PRIu32,
+		(intmax_t)pid, stress_mwc32());
+
+	t = stress_time_now();
+	fd = open(filename, O_RDWR | O_CREAT | O_TRUNC | O_DIRECT, S_IRUSR | S_IWUSR);
+	if (fd >= 0) {
+		(*duration) += stress_time_now() - t;
+		(*count) += 1.0;
+	} else {
+		int ret;
+		struct stat statbuf;
+
+		ret = stat(filename, &statbuf);
+		if (ret == 0) {
+			pr_inf("%s: open with O_DIRECT failed but file '%s' was created\n",
+				args->name, filename);
+		}
+	}
+	(void)shim_unlink(filename);
+	return fd;
 }
 #endif
 
 #if defined(HAVE_OPENAT) &&	\
-    defined(AT_FDCWD)
-static int open_with_openat_cwd(void)
+    defined(AT_FDCWD) &&	\
+    defined(O_CREAT)
+static int open_with_openat_cwd(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
 {
 	char cwd[PATH_MAX];
 	char filename[PATH_MAX];
 	const char *temp_path = stress_get_temp_path();
-	int fd, ret;
+	int fd;
+	double t;
+
+	(void)args;
 
 	if (!temp_path)
 		return -1;
@@ -507,13 +655,17 @@ static int open_with_openat_cwd(void)
 	if (chdir(temp_path) < 0)
 		return -1;
 
-	(void)snprintf(filename, sizeof(filename), "stress-open-%d-%" PRIu32,
-		(int)getpid(), stress_mwc32());
+	(void)snprintf(filename, sizeof(filename), "stress-open-%" PRIdMAX "-%" PRIu32,
+		(intmax_t)pid, stress_mwc32());
 
+	t = stress_time_now();
 	fd = openat(AT_FDCWD, filename, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
 	if (fd >= 0) {
 		struct timeval tv[2];
 
+		(*duration) += stress_time_now() - t;
+		(*count) += 1.0;
+
 		(void)obsolete_futimesat(AT_FDCWD, filename, NULL);
 
 		/* Exercise illegal times */
@@ -526,8 +678,7 @@ static int open_with_openat_cwd(void)
 		(void)shim_force_unlink(filename);
 	}
 
-	ret = chdir(cwd);
-	(void)ret;
+	VOID_RET(int, chdir(cwd));
 	return fd;
 }
 #endif
@@ -535,16 +686,23 @@ static int open_with_openat_cwd(void)
 #if defined(HAVE_OPENAT) &&	\
     defined(AT_FDCWD) &&	\
     defined(O_PATH) &&		\
-    defined(O_DIRECTORY)
-static int open_with_openat_dir_fd(void)
+    defined(O_DIRECTORY) &&	\
+    defined(O_CREAT)
+static int open_with_openat_dir_fd(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
 {
 	char filename[PATH_MAX];
 	int fd, dir_fd;
 
-	(void)snprintf(filename, sizeof(filename), "stress-open-%d-%" PRIu32,
-		(int)getpid(), stress_mwc32());
+	(void)args;
 
-	dir_fd = open_arg2(".", O_DIRECTORY | O_PATH);
+	(void)snprintf(filename, sizeof(filename), "stress-open-%" PRIdMAX "-%" PRIu32,
+		(intmax_t)pid, stress_mwc32());
+
+	dir_fd = open_arg2(".", O_DIRECTORY | O_PATH, duration, count);
 	if (dir_fd < 0)
 		return -1;
 
@@ -562,8 +720,13 @@ static int open_with_openat_dir_fd(void)
     defined(HAVE_LINUX_OPENAT2_H) &&	\
     defined(RESOLVE_NO_SYMLINKS) &&	\
     defined(__NR_openat2) &&		\
-    defined(HAVE_SYSCALL)
-static int open_with_openat2_cwd(void)
+    defined(HAVE_SYSCALL) &&		\
+    defined(O_CREAT)
+static int open_with_openat2_cwd(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
 {
 	static const unsigned int resolve_flags[] = {
 #if defined(RESOLVE_BENEATH)
@@ -591,25 +754,29 @@ static int open_with_openat2_cwd(void)
 	char cwd[PATH_MAX];
 	char filename[PATH_MAX];
 	const char *temp_path = stress_get_temp_path();
-	int fd = -1, ret;
+	int fd = -1;
 	size_t i = 0;
 	static size_t j;
 	struct open_how how;
 
+	(void)args;
+
 	if (!temp_path)
 		return -1;
 	if (!getcwd(cwd, sizeof(cwd)))
 		return -1;
 	if (chdir(temp_path) < 0)
 		return -1;
-	(void)snprintf(filename, sizeof(filename), "stress-open-%d-%" PRIu32,
-		(int)getpid(), stress_mwc32());
+	(void)snprintf(filename, sizeof(filename), "stress-open-%" PRIdMAX "-%" PRIu32,
+		(intmax_t)pid, stress_mwc32());
 
 	/*
 	 *  Work through resolve flags to find one that can
 	 *  open the file successfully
 	 */
 	for (i = 0; i < SIZEOF_ARRAY(resolve_flags); i++) {
+		double t;
+
 		(void)memset(&how, 0, sizeof(how));
 		how.flags = O_CREAT | O_RDWR;
 		how.mode = S_IRUSR | S_IWUSR;
@@ -626,34 +793,97 @@ static int open_with_openat2_cwd(void)
 			break;
 		}
 
+		t = stress_time_now();
 		fd = (int)syscall(__NR_openat2, AT_FDCWD, filename, &how, sizeof(how));
 		if (fd >= 0) {
+			(*duration) += stress_time_now() - t;
+			(*count) += 1.0;
 			(void)shim_unlink(filename);
 			break;
 		}
 	}
 
-	ret = chdir(cwd);
-	(void)ret;
-
+	VOID_RET(int, chdir(cwd));
 	return fd;
 }
 #endif
 
 #if defined(__linux__)
-static int open_with_open_proc_self_fd(void)
+static int open_with_open_proc_self_fd(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
 {
-	return open("/proc/self/fd/0", O_RDONLY);
+	int fd;
+	double t;
+
+	(void)args;
+	(void)pid;
+
+	t = stress_time_now();
+	fd = open("/proc/self/fd/0", O_RDONLY);
+	if (fd >= 0) {
+		(*duration) += stress_time_now() - t;
+		(*count) += 1.0;
+	}
+	return fd;
 }
 #endif
 
-static int open_dup(void)
+static int open_dup(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
 {
-	return dup(STDOUT_FILENO);
+	int fd;
+	double t;
+
+	(void)args;
+	(void)pid;
+
+	t = stress_time_now();
+	fd = dup(STDOUT_FILENO);
+	if (fd >= 0) {
+		(*duration) += stress_time_now() - t;
+		(*count) += 1.0;
+	}
+	return fd;
 }
 
+#if defined(O_CREAT) &&	\
+    defined(O_TRUNC)
+static int open_rdonly_trunc(
+	const stress_args_t *args,
+	const pid_t pid,
+	double *duration,
+	double *count)
+{
+	char filename[PATH_MAX];
+	int fd;
+
+	(void)args;
+
+	(void)snprintf(filename, sizeof(filename), "stress-open-%" PRIdMAX "-%" PRIu32,
+		(intmax_t)pid, stress_mwc32());
+
+	fd = open_arg3(filename, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR, duration, count);
+	if (fd < 0)
+		return fd;
+
+	/* undefined behaviour, will open, may truncate on some systems */
+	fd = open_arg2(filename, O_RDONLY | O_TRUNC, duration, count);
+	if (fd >= 0)
+		(void)shim_unlink(filename);
+	return fd;
+}
+#endif
+
 static stress_open_func_t open_funcs[] = {
+#if defined(O_CREAT)
 	open_flag_perm,
+#endif
 	open_dev_zero_rd,
 	open_dev_null_wr,
 #if defined(O_TMPFILE)
@@ -681,25 +911,40 @@ static stress_open_func_t open_funcs[] = {
 	open_create_eisdir,
 #endif
 #if defined(HAVE_OPENAT) &&		\
-    defined(AT_FDCWD)
+    defined(AT_FDCWD) &&		\
+    defined(O_CREAT)
 	open_with_openat_cwd,
 #endif
 #if defined(HAVE_OPENAT) &&	\
     defined(AT_FDCWD) &&	\
     defined(O_PATH) &&		\
-    defined(O_DIRECTORY)
+    defined(O_DIRECTORY) &&	\
+    defined(O_CREAT)
 	open_with_openat_dir_fd,
 #endif
 #if defined(HAVE_OPENAT2) &&		\
     defined(HAVE_LINUX_OPENAT2_H) &&	\
     defined(RESOLVE_NO_SYMLINKS) &&	\
-    defined(__NR_openat2)
+    defined(__NR_openat2) &&		\
+    defined(O_CREAT)
 	open_with_openat2_cwd,
 #endif
 #if defined(__linux__)
 	open_with_open_proc_self_fd,
 #endif
 	open_dup,
+#if defined(O_DIRECT) &&        \
+    defined(O_CREAT)
+	open_direct,
+#endif
+#if defined(O_DIRECTORY) &&	\
+    defined(__linux__)
+	open_dir_proc_self_fd,
+#endif
+#if defined(O_CREAT) &&	\
+    defined(O_TRUNC)
+	open_rdonly_trunc,
+#endif
 };
 
 /*
@@ -708,7 +953,7 @@ static stress_open_func_t open_funcs[] = {
  *	via the file names provided in /proc/$PID/fd. This
  *	ignores opens that fail.
  */
-static void stress_fd_dir(const char *path)
+static void stress_fd_dir(const char *path, double *duration, double *count)
 {
 	for (;;) {
 		struct dirent *de;
@@ -723,7 +968,7 @@ static void stress_fd_dir(const char *path)
 			int fd;
 
 			stress_mk_filename(name, sizeof(name), path, de->d_name);
-			fd = open_arg2(name, O_RDONLY);
+			fd = open_arg2(name, O_RDONLY, duration, count);
 			if (fd >= 0)
 				(void)close(fd);
 		}
@@ -747,6 +992,7 @@ static int stress_open(const stress_args_t *args)
 	struct stat statbuf;
 	bool open_fd = false;
 	int all_open_flags;
+	double duration = 0.0, count = 0.0, rate;
 
 	/*
 	 *  32 bit systems may OOM if we have too many open fds, so
@@ -772,7 +1018,7 @@ static int stress_open(const stress_args_t *args)
 	fds = (int *)mmap(NULL, sz, PROT_READ | PROT_WRITE,
 		MAP_ANONYMOUS | MAP_PRIVATE, -1, 0);
 	if (fds == MAP_FAILED) {
-		pr_inf("%s: cannot allocate %zd file descriptors\n", args->name, open_max);
+		pr_inf_skip("%s: cannot mmap %zd file descriptors\n", args->name, open_max);
 		return EXIT_NO_RESOURCE;
 	}
 
@@ -783,7 +1029,7 @@ static int stress_open(const stress_args_t *args)
 			pid = fork();
 
 			if (pid == 0) {
-				stress_fd_dir(path);
+				stress_fd_dir(path, &duration, &count);
 				_exit(0);
 			}
 		}
@@ -810,8 +1056,8 @@ static int stress_open(const stress_args_t *args)
 					goto close_all;
 				}
 
-				idx = stress_mwc32() % SIZEOF_ARRAY(open_funcs);
-				fds[i] = open_funcs[idx]();
+				idx = stress_mwc32modn(SIZEOF_ARRAY(open_funcs));
+				fds[i] = open_funcs[idx](args, mypid, &duration, &count);
 
 				if (fds[i] >= 0)
 					break;
@@ -867,6 +1113,9 @@ close_all:
 		(void)waitpid(pid, &status, 0);
 	}
 
+	rate = (count > 0.0) ? duration / count: 0.0;
+	stress_metrics_set(args, 0, "nanosecs per open", rate * 1000000000);
+
 	return EXIT_SUCCESS;
 }
 
