@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C)      2022 Colin Ian King.
+ * Copyright (C) 2022-2023 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -19,6 +19,7 @@
  */
 #include "stress-ng.h"
 #include "core-arch.h"
+#include "core-pthread.h"
 
 #if defined(HAVE_MODIFY_LDT)
 #include <asm/ldt.h>
@@ -50,8 +51,8 @@
 
 static const stress_help_t help[] = {
 	{ NULL,	"pthread N",	 "start N workers that create multiple threads" },
-	{ NULL,	"pthread-ops N", "stop pthread workers after N bogo threads created" },
 	{ NULL,	"pthread-max P", "create P threads at a time by each worker" },
+	{ NULL,	"pthread-ops N", "stop pthread workers after N bogo threads created" },
 	{ NULL,	NULL,		 NULL }
 };
 
@@ -59,11 +60,13 @@ static const stress_help_t help[] = {
 
 /* per pthread data */
 typedef struct {
-	pthread_t pthread;	/* The pthread */
-	int	  ret;		/* pthread create return */
-	uint64_t index;		/* which pthread */
+	pthread_t pthread;		/* The pthread */
+	int	  ret;			/* pthread create return */
+	uint64_t  index;		/* which pthread */
+	volatile double	t_create;	/* time when created */
+	volatile double	t_run;		/* time when thread started running */
 #if defined(HAVE_PTHREAD_ATTR_SETSTACK)
-	void 	 *stack;	/* pthread stack */
+	void 	 *stack;		/* pthread stack */
 #endif
 } stress_pthread_info_t;
 
@@ -97,7 +100,9 @@ static const stress_opt_set_func_t opt_set_funcs[] = {
 #define DEFAULT_STACK_MIN		(8 * KB)
 
 #if defined(HAVE_GET_ROBUST_LIST) &&	\
-    defined(HAVE_LINUX_FUTEX_H)
+    defined(HAVE_LINUX_FUTEX_H) &&	\
+    defined(HAVE_SYSCALL) &&		\
+    defined(__NR_get_robust_list)
 static inline long sys_get_robust_list(int pid, struct robust_list_head **head_ptr, size_t *len_ptr)
 {
 	return syscall(__NR_get_robust_list, pid, head_ptr, len_ptr);
@@ -105,7 +110,9 @@ static inline long sys_get_robust_list(int pid, struct robust_list_head **head_p
 #endif
 
 #if defined(HAVE_SET_ROBUST_LIST) &&	\
-    defined(HAVE_LINUX_FUTEX_H)
+    defined(HAVE_LINUX_FUTEX_H) &&	\
+    defined(HAVE_SYSCALL) &&		\
+    defined(__NR_set_robust_list)
 static inline long sys_set_robust_list(struct robust_list_head *head, size_t len)
 {
 	return syscall(__NR_set_robust_list, head, len);
@@ -138,7 +145,7 @@ static bool keep_running(void)
  */
 static bool keep_thread_running(void)
 {
-	return keep_running() & keep_thread_running_flag;
+	return keep_running() && keep_thread_running_flag;
 }
 
 /*
@@ -154,7 +161,8 @@ static void stress_pthread_tid_address(const stress_args_t *args)
     defined(HAVE_PRCTL) &&		\
     defined(PR_GET_TID_ADDRESS) &&	\
     defined(__NR_set_tid_address) &&	\
-    defined(HAVE_KERNEL_ULONG_T)
+    defined(HAVE_KERNEL_ULONG_T) &&	\
+    defined(HAVE_SYSCALL)
 	__kernel_ulong_t tid_addr = 0;
 
 	if (prctl(PR_GET_TID_ADDRESS, &tid_addr) == 0) {
@@ -162,8 +170,7 @@ static void stress_pthread_tid_address(const stress_args_t *args)
 			pid_t tid1, tid2;
 
 			/* Nullify */
-			tid1 = (pid_t)syscall(__NR_set_tid_address, NULL);
-			(void)tid1;
+			VOID_RET(pid_t, (pid_t)syscall(__NR_set_tid_address, NULL));
 
 			/* This always succeeds */
 			tid1 = (pid_t)syscall(__NR_set_tid_address, tid_addr);
@@ -189,10 +196,7 @@ static void *stress_pthread_func(void *parg)
 {
 	static void *nowt = NULL;
 	int ret;
-#if defined(HAVE_GET_ROBUST_LIST) &&	\
-    defined(HAVE_LINUX_FUTEX_H)
-	long lret;
-#endif
+        double t_run = stress_time_now();
 	const pid_t tgid = getpid();
 #if defined(HAVE_GETTID)
 	const pid_t tid = shim_gettid();
@@ -206,10 +210,12 @@ static void *stress_pthread_func(void *parg)
 #endif
 	const stress_pthread_args_t *spa = (stress_pthread_args_t *)parg;
 	const stress_args_t *args = spa->args;
-	const stress_pthread_info_t *pthread_info = (stress_pthread_info_t *)spa->data;
+	stress_pthread_info_t *pthread_info = (stress_pthread_info_t *)spa->data;
 	char str[16];
 
-	snprintf(str, sizeof(str), "%" PRIu64, pthread_info->index);
+	pthread_info->t_run = t_run;
+
+	(void)snprintf(str, sizeof(str), "%" PRIu64, pthread_info->index);
 	stress_set_proc_state_str(args->name, str);
 
 #if defined(HAVE_GET_ROBUST_LIST) &&	\
@@ -235,63 +241,45 @@ static void *stress_pthread_func(void *parg)
 		}
 
 		/* Exercise invalid zero length */
-		lret = sys_set_robust_list(head, 0);
-		(void)lret;
+		VOID_RET(long, sys_set_robust_list(head, 0));
 
 		/* Exercise invalid length */
-		lret = sys_set_robust_list(head, (size_t)-1);
-		(void)lret;
+		VOID_RET(long, sys_set_robust_list(head, (size_t)-1));
 #endif
 	/*
 	 *  Check get_robust_list with an invalid PID
 	 */
 	}
-	lret = sys_get_robust_list(-1, &head, &len);
-	(void)lret;
+	VOID_RET(long, sys_get_robust_list(-1, &head, &len));
 #endif
 
 	/*
 	 *  Exercise tgkill (no-op on systems that do
 	 *  not support this system call)
 	 */
-	ret = shim_tgkill(tgid, tid, 0);
-	(void)ret;
-
-	ret = shim_tgkill(-1, tid, 0);
-	(void)ret;
-
-	ret = shim_tgkill(tgid, -1, 0);
-	(void)ret;
-
-	ret = shim_tgkill(tgid, tid, -1);
-	(void)ret;
-
-	ret = shim_tgkill(stress_get_unused_pid_racy(false), tid, 0);
-	(void)ret;
-
-	ret = shim_tgkill(tgid, stress_get_unused_pid_racy(false), 0);
-	(void)ret;
+	VOID_RET(int, shim_tgkill(tgid, tid, 0));
+	VOID_RET(int, shim_tgkill(-1, tid, 0));
+	VOID_RET(int, shim_tgkill(tgid, -1, 0));
+	VOID_RET(int, shim_tgkill(tgid, tid, -1));
+	VOID_RET(int, shim_tgkill(stress_get_unused_pid_racy(false), tid, 0));
+	VOID_RET(int, shim_tgkill(tgid, stress_get_unused_pid_racy(false), 0));
 
 	/*
 	 *  Exercise tkill, this is either supported directly, emulated
 	 *  by tgkill or a no-op if systems don't support tkill or tgkill
 	 */
-	ret = shim_tkill(tid, 0);
-	(void)ret;
+	VOID_RET(int, shim_tkill(tid, 0));
+	VOID_RET(int, shim_tkill(-1, 0));
+	VOID_RET(int, shim_tkill(tid, -1));
+	VOID_RET(int, ret = shim_tkill(stress_get_unused_pid_racy(false), 0));
 
-	ret = shim_tkill(-1, 0);
-	(void)ret;
-
-	ret = shim_tkill(tid, -1);
-	(void)ret;
-
-	ret = shim_tkill(stress_get_unused_pid_racy(false), 0);
-	(void)ret;
-
-#if defined(HAVE_ASM_LDT_H) && 	\
-    defined(STRESS_ARCH_X86) &&	\
-    defined(HAVE_USER_DESC) && 	\
-    defined(HAVE_GET_THREAD_AREA)
+#if defined(HAVE_ASM_LDT_H) && 		\
+    defined(STRESS_ARCH_X86) &&		\
+    defined(HAVE_USER_DESC) && 		\
+    defined(HAVE_GET_THREAD_AREA) &&	\
+    defined(HAVE_SYSCALL) &&		\
+    defined(__NR_get_thread_area) &&	\
+    defined(__NR_set_thread_area)
 	{
 		struct user_desc u_info;
 
@@ -301,8 +289,7 @@ static void *stress_pthread_func(void *parg)
 		ret = (int)syscall(__NR_get_thread_area, &u_info);
 #if defined(HAVE_GET_THREAD_AREA)
 		if (ret == 0) {
-			ret = (int)syscall(__NR_set_thread_area, &u_info);
-			(void)ret;
+			VOID_RET(int, (int)syscall(__NR_set_thread_area, &u_info));
 		}
 #else
 		(void)ret;
@@ -404,8 +391,7 @@ yield:
 		timeout.tv_nsec = 1000000;
 
 		/* Ignore error return, just exercise the call */
-		ret = sigtimedwait(&mask, &info, &timeout);
-		(void)ret;
+		VOID_RET(int, sigtimedwait(&mask, &info, &timeout));
 	}
 #endif
 
@@ -429,6 +415,7 @@ static int stress_pthread(const stress_args_t *args)
 	int ret;
 	stress_pthread_args_t pargs = { args, NULL, 0 };
 	sigset_t set;
+	double count = 0.0, duration = 0.0, average;
 #if defined(HAVE_PTHREAD_ATTR_SETSTACK)
 	const size_t stack_size = STRESS_MAXIMUM(DEFAULT_STACK_MIN, stress_min_pthread_stack_size());
 #endif
@@ -482,10 +469,8 @@ static int stress_pthread(const stress_args_t *args)
 	ret = pthread_mutexattr_init(&mutex_attr);
 	if (ret == 0) {
 		mutex_attr_init = true;
-		ret = pthread_mutexattr_setprotocol(&mutex_attr, PTHREAD_PRIO_INHERIT);
-		(void)ret;
-		ret = pthread_mutexattr_setprioceiling(&mutex_attr, 127);
-		(void)ret;
+		VOID_RET(int, pthread_mutexattr_setprotocol(&mutex_attr, PTHREAD_PRIO_INHERIT));
+		VOID_RET(int, pthread_mutexattr_setprioceiling(&mutex_attr, 127));
 	} else {
 		mutex_attr_init = false;
 	}
@@ -536,20 +521,24 @@ static int stress_pthread(const stress_args_t *args)
 
 			ret = pthread_attr_init(&attr);
 			if (ret) {
+				(void)pthread_mutex_unlock(&mutex);
 				pr_fail("%s: pthread_attr_init failed, errno=%d (%s)\n",
 					args->name, ret, strerror(ret));
 				stop_running();
-				break;
+				goto reap;
 			}
 			ret = pthread_attr_setstack(&attr, pthreads[i].stack, stack_size);
 			if (ret) {
+				(void)pthread_mutex_unlock(&mutex);
 				pr_fail("%s: pthread_attr_setstack failed, errno=%d (%s)\n",
 					args->name, ret, strerror(ret));
 				stop_running();
-				break;
+				goto reap;
 			}
 #endif
 
+			pthreads[i].t_create = stress_time_now();
+			pthreads[i].t_run = pthreads[i].t_create;
 			pthreads[i].ret = pthread_create(&pthreads[i].pthread, NULL,
 				stress_pthread_func, (void *)&pargs);
 			if (pthreads[i].ret) {
@@ -559,10 +548,11 @@ static int stress_pthread(const stress_args_t *args)
 					break;
 				}
 				/* Something really unexpected */
+				(void)pthread_mutex_unlock(&mutex);
 				pr_fail("%s: pthread_create failed, errno=%d (%s)\n",
 					args->name, pthreads[i].ret, strerror(pthreads[i].ret));
 				stop_running();
-				break;
+				goto reap;
 			}
 			if (i + 1 > maximum)
 				maximum = i + 1;
@@ -626,8 +616,7 @@ static int stress_pthread(const stress_args_t *args)
 				continue;
 
 			(void)memset(&value, 0, sizeof(value));
-			ret = pthread_sigqueue(pthreads[j].pthread, SIGUSR1, value);
-			(void)ret;
+			VOID_RET(int, pthread_sigqueue(pthreads[j].pthread, SIGUSR1, value));
 		}
 #endif
 
@@ -642,12 +631,20 @@ reap:
 		}
 		for (j = 0; j < i; j++) {
 			if (pthreads[j].ret == 0) {
+				double t;
+
 				ret = pthread_join(pthreads[j].pthread, NULL);
 				if ((ret) && (ret != ESRCH)) {
 					pr_fail("%s: pthread_join failed (parent), errno=%d (%s)\n",
 						args->name, ret, strerror(ret));
 					stop_running();
 				}
+
+				t = pthreads[j].t_run - pthreads[j].t_create;
+				if (t > 0.0) {
+					duration += t;
+					count += 1;
+				}
 			}
 #if defined(HAVE_PTHREAD_ATTR_SETSTACK)
 			if (pthreads[j].stack != NULL)
@@ -656,6 +653,9 @@ reap:
 		}
 	} while (!locked && keep_running() && keep_stressing(args));
 
+	average = (count > 0.0) ? duration / count : 0.0;
+	stress_metrics_set(args, 0, "nanosecs to start a pthread", average * STRESS_DBL_NANOSECOND);
+
 	pr_inf("%s: maximum of %" PRIu64 " concurrent pthreads created\n",
 		args->name, maximum);
 	if (limited) {
@@ -693,9 +693,10 @@ stressor_info_t stress_pthread_info = {
 };
 #else
 stressor_info_t stress_pthread_info = {
-	.stressor = stress_not_implemented,
+	.stressor = stress_unimplemented,
 	.class = CLASS_SCHEDULER | CLASS_OS,
 	.opt_set_funcs = opt_set_funcs,
-	.help = help
+	.help = help,
+	.unimplemented_reason = "built without pthread support"
 };
 #endif
