@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C)      2022 Colin Ian King.
+ * Copyright (C) 2022-2023 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -144,9 +144,7 @@ static int exercise_renameat2(
 	 * Exercise RENAME_EXCHANGE on same file creating absolutely
 	 * no difference other than increase in kernel test coverage
 	 */
-	ret = renameat2(oldfd, old_name, oldfd, old_name,
-		RENAME_EXCHANGE);
-	(void)ret;
+	VOID_RET(int, renameat2(oldfd, old_name, oldfd, old_name, RENAME_EXCHANGE));
 
 #endif
 
@@ -251,10 +249,11 @@ restart:
 		args->name, args->pid, inst1, i++);
 
 	if ((fp = fopen(oldname, "w+")) == NULL) {
-		int rc = exit_status(errno);
+		int rc = stress_exit_status(errno);
 
-		pr_err("%s: fopen failed: errno=%d: (%s)\n",
-			args->name, errno, strerror(errno));
+		pr_err("%s: fopen failed: errno=%d: (%s)%s\n",
+			args->name, errno, strerror(errno),
+			stress_fs_type(oldname));
 		(void)stress_temp_dir_rm(args->name, args->pid, inst1);
 		(void)stress_temp_dir_rm(args->name, args->pid, inst2);
 #if defined(EXERCISE_RENAMEAT) ||	\
