@@ -1,5 +1,5 @@
 /*
- * Copyright (C)      2022 Colin Ian King.
+ * Copyright (C) 2022-2023 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -17,6 +17,7 @@
  *
  */
 #include "stress-ng.h"
+#include "core-pragma.h"
 
 #if defined(HAVE_SYS_TREE_H)
 #include <sys/tree.h>
@@ -37,7 +38,7 @@
 /* Number of items in sparse matrix */
 #define MIN_SPARSEMATRIX_ITEMS		(100)
 #define MAX_SPARSEMATRIX_ITEMS		(10000000)
-#define DEFAULT_SPARSEMATRIX_ITEMS	(5000)
+#define DEFAULT_SPARSEMATRIX_ITEMS	(10000)
 
 /* Sparse matrix X x Y size */
 #define MIN_SPARSEMATRIX_SIZE		(10)
@@ -59,6 +60,20 @@
 #define HAVE_RB_TREE	(1)
 #endif
 
+/* BSD splay tree */
+#if defined(SPLAY_HEAD) &&	\
+    defined(SPLAY_PROTOTYPE) &&	\
+    defined(SPLAY_GENERATE) &&	\
+    defined(SPLAY_ENTRY) &&	\
+    defined(SPLAY_INIT) &&	\
+    defined(SPLAY_FIND) &&	\
+    defined(SPLAY_INSERT) &&	\
+    defined(SPLAY_MIN) &&	\
+    defined(SPLAY_NEXT) &&	\
+    defined(SPLAY_REMOVE)
+#define HAVE_SPLAY_TREE
+#endif
+
 #if defined(HAVE_JUDY_H) && 	\
     defined(HAVE_LIB_JUDY) &&	\
     (ULONG_MAX > 0xffffffff)
@@ -69,11 +84,11 @@
 #define SPARSE_TEST_FAILED	(-1)
 #define SPARSE_TEST_ENOMEM	(-2)
 
-typedef void * (*func_create)(const uint32_t n, const uint32_t x, const uint32_t y);
+typedef void * (*func_create)(const uint64_t n, const uint32_t x, const uint32_t y);
 typedef void (*func_destroy)(void *handle, size_t *objmem);
-typedef int (*func_put)(void *handle, const uint32_t x, const uint32_t y, const uint64_t value);
+typedef int (*func_put)(void *handle, const uint32_t x, const uint32_t y, const uint32_t value);
 typedef void (*func_del)(void *handle, const uint32_t x, const uint32_t y);
-typedef uint64_t (*func_get)(void *handle, const uint32_t x, const uint32_t y);
+typedef uint32_t (*func_get)(void *handle, const uint32_t x, const uint32_t y);
 
 typedef struct {
 	const char              *name;  /* human readable form of sparse method */
@@ -88,43 +103,60 @@ static const stress_sparsematrix_method_info_t sparsematrix_methods[];
 
 static const stress_help_t help[] = {
 	{ NULL,	"sparsematrix N",	 "start N workers that exercise a sparse matrix" },
-	{ NULL,	"sparsematrix-ops N",	 "stop after N bogo sparse matrix operations" },
-	{ NULL,	"sparsematrix-method M", "select storage method: all, hash, judy, list or rb" },
 	{ NULL,	"sparsematrix-items N",	 "N is the number of items in the spare matrix" },
+	{ NULL,	"sparsematrix-method M", "select storage method: all, hash, hashjudt, judy, list, mmap, qhash, rb, splay" },
+	{ NULL,	"sparsematrix-ops N",	 "stop after N bogo sparse matrix operations" },
 	{ NULL,	"sparsematrix-size N",	 "M is the width and height X x Y of the matrix" },
-	{ NULL,	NULL,		 NULL }
+	{ NULL,	NULL,		 	 NULL }
 };
 
 #if defined(HAVE_RB_TREE)
 typedef struct sparse_rb {
 	uint64_t xy;		/* x,y matrix position */
-	uint64_t value;		/* value in matrix x,y */
+	uint32_t value;		/* value in matrix x,y */
 	RB_ENTRY(sparse_rb) rb;	/* red-black tree node entry */
 } sparse_rb_t;
 #else
 UNEXPECTED
 #endif
 
+#if defined(HAVE_SPLAY_TREE)
+typedef struct sparse_splay {
+	uint64_t xy;		/* x,y matrix position */
+	uint32_t value;		/* value in matrix x,y */
+	SPLAY_ENTRY(sparse_splay) splay; /* splay tree node entry */
+} sparse_splay_t;
+#else
+UNEXPECTED
+#endif
+
 typedef struct sparse_hash_node {
 	struct sparse_hash_node *next;
 	uint64_t xy;		/* x,y matrix position */
-	uint64_t value;		/* value in matrix x,y */
+	uint32_t value;		/* value in matrix x,y */
 } sparse_hash_node_t;
 
 typedef struct sparse_hash_table {
-	uint32_t n;		/* size of hash table */
+	uint64_t n;		/* size of hash table */
 	sparse_hash_node_t **table;
 } sparse_hash_table_t;
 
+#if defined(HAVE_JUDY_H)
+typedef struct sparse_hashjudy_table {
+	uint64_t n;		/* size of hash table */
+	Pvoid_t *hash_table;	/* hash table of Judy arrays */
+} sparse_hashjudy_table_t;
+#endif
+
 typedef struct sparse_qhash_node {
 	struct sparse_qhash_node *next;
 	uint64_t xy;		/* x,y matrix position */
-	uint64_t value;		/* value in matrix x,y */
+	uint32_t value;		/* value in matrix x,y */
 } sparse_qhash_node_t;
 
 typedef struct sparse_qhash_table {
-	uint32_t n;		/* size of hash table */
-	uint32_t n_nodes;	/* number of nodes */
+	uint64_t n;		/* size of hash table */
+	uint64_t n_nodes;	/* number of nodes */
 	sparse_qhash_node_t **table;
 	sparse_qhash_node_t *nodes;
 	size_t idx;
@@ -134,8 +166,8 @@ typedef struct sparse_qhash_table {
 
 typedef struct sparse_x_list_node {
 	CIRCLEQ_ENTRY(sparse_x_list_node) sparse_x_list;
+	uint32_t value;		/* value in matrix x,y */
 	uint32_t x;		/* x matrix position */
-	uint64_t value;		/* value in matrix x,y */
 } sparse_x_list_node_t;
 
 CIRCLEQ_HEAD(sparse_x_list, sparse_x_list_node);
@@ -157,18 +189,18 @@ UNEXPECTED
 
 typedef struct {
 	void *mmap;
+	uint64_t mmap_size;
 	uint32_t x;
 	uint32_t y;
-	size_t mmap_size;
 } sparse_mmap_t;
 
 typedef struct {
-	bool	skip_no_mem;	/* True if can't allocate memory */
 	size_t	max_objmem;	/* Object memory allocation estimate */
 	double	put_duration;	/* Total put duration time, seconds */
 	double	get_duration;	/* Total get duration time, seconds */
-	uint64_t put_ops;	/* Totoal put object op count */
-	uint64_t get_ops;	/* Totoal put object op count */
+	uint64_t put_ops;	/* Total put object op count */
+	uint64_t get_ops;	/* Total put object op count */
+	bool	skip_no_mem;	/* True if can't allocate memory */
 } test_info_t;
 
 /*
@@ -194,19 +226,24 @@ static int stress_set_sparsematrix_size(const char *opt)
 	uint32_t sparsematrix_size;
 
 	sparsematrix_size = stress_get_uint32(opt);
-	stress_check_range("sparsematrix-size", sparsematrix_size,
+	stress_check_range("sparsematrix-size", (uint64_t)sparsematrix_size,
 		MIN_SPARSEMATRIX_SIZE, MAX_SPARSEMATRIX_SIZE);
 	return stress_set_setting("sparsematrix-size", TYPE_ID_UINT32, &sparsematrix_size);
 }
 
+static uint32_t value_map(const uint32_t x, const uint32_t y)
+{
+	return x ^ ~y;
+}
+
 /*
  *  hash_create()
  *	create a hash table based sparse matrix
  */
-static void *hash_create(const uint32_t n, const uint32_t x, const uint32_t y)
+static void *hash_create(const uint64_t n, const uint32_t x, const uint32_t y)
 {
 	sparse_hash_table_t *table;
-	uint32_t n_prime = (size_t)stress_get_prime64((uint64_t)n);
+	uint64_t n_prime = (uint64_t)stress_get_prime64(n);
 
 	(void)x;
 	(void)y;
@@ -216,7 +253,7 @@ static void *hash_create(const uint32_t n, const uint32_t x, const uint32_t y)
 		return NULL;
 
 	table->table = (sparse_hash_node_t **)calloc((size_t)n_prime, sizeof(sparse_hash_node_t *));
-	if (!table) {
+	if (!table->table) {
 		free(table);
 		return NULL;
 	}
@@ -262,12 +299,12 @@ static void hash_destroy(void *handle, size_t *objmem)
  *  hash_put()
  *	put a value into a hash based sparse matrix
  */
-static int hash_put(void *handle, const uint32_t x, const uint32_t y, const uint64_t value)
+static int OPTIMIZE3 hash_put(void *handle, const uint32_t x, const uint32_t y, const uint32_t value)
 {
 	sparse_hash_node_t *node;
 	sparse_hash_table_t *table = (sparse_hash_table_t *)handle;
 	size_t hash;
-	uint64_t xy = ((uint64_t)x << 32) | y;
+	const uint64_t xy = ((uint64_t)x << 32) | y;
 
 	if (!table)
 		return -1;
@@ -283,7 +320,7 @@ static int hash_put(void *handle, const uint32_t x, const uint32_t y, const uint
 	}
 
 	/* Not found, allocate and add */
-	node = calloc(1, sizeof(*node));
+	node = malloc(sizeof(*node));
 	if (!node)
 		return -1;
 	node->value = value;
@@ -297,12 +334,12 @@ static int hash_put(void *handle, const uint32_t x, const uint32_t y, const uint
  *  hash_get_node()
  *	find the hash table node of a (x,y) value in a hash table
  */
-static sparse_hash_node_t *hash_get_node(void *handle, const uint32_t x, const uint32_t y)
+static sparse_hash_node_t OPTIMIZE3 *hash_get_node(void *handle, const uint32_t x, const uint32_t y)
 {
 	sparse_hash_table_t *table = (sparse_hash_table_t *)handle;
 	sparse_hash_node_t *node;
 	size_t hash;
-	uint64_t xy = ((uint64_t)x << 32) | y;
+	const uint64_t xy = ((uint64_t)x << 32) | y;
 
 	if (!table)
 		return NULL;
@@ -319,7 +356,7 @@ static sparse_hash_node_t *hash_get_node(void *handle, const uint32_t x, const u
  *  hash_get()
  *	get the (x,y) value in hash table based sparse matrix
  */
-static uint64_t hash_get(void *handle, const uint32_t x, const uint32_t y)
+static uint32_t OPTIMIZE3 hash_get(void *handle, const uint32_t x, const uint32_t y)
 {
 	sparse_hash_node_t *node = hash_get_node(handle, x, y);
 
@@ -330,7 +367,7 @@ static uint64_t hash_get(void *handle, const uint32_t x, const uint32_t y)
  *  hash_del()
  *	zero the (x,y) value in sparse hash table
  */
-static void hash_del(void *handle, const uint32_t x, const uint32_t y)
+static void OPTIMIZE3 hash_del(void *handle, const uint32_t x, const uint32_t y)
 {
 	sparse_hash_node_t *node = hash_get_node(handle, x, y);
 
@@ -342,10 +379,10 @@ static void hash_del(void *handle, const uint32_t x, const uint32_t y)
  *  qhash_create()
  *	create a hash table based sparse matrix
  */
-static void *qhash_create(const uint32_t n, const uint32_t x, const uint32_t y)
+static void *qhash_create(const uint64_t n, const uint32_t x, const uint32_t y)
 {
 	sparse_qhash_table_t *table;
-	uint32_t n_prime = (size_t)stress_get_prime64((uint64_t)n);
+	const uint64_t n_prime = stress_get_prime64(n);
 
 	(void)x;
 	(void)y;
@@ -355,7 +392,7 @@ static void *qhash_create(const uint32_t n, const uint32_t x, const uint32_t y)
 		return NULL;
 
 	table->table = (sparse_qhash_node_t **)calloc((size_t)n_prime, sizeof(sparse_qhash_node_t *));
-	if (!table) {
+	if (!table->table) {
 		free(table);
 		return NULL;
 	}
@@ -406,12 +443,12 @@ static void qhash_destroy(void *handle, size_t *objmem)
  *  qhash_put()
  *	put a value into a hash based sparse matrix
  */
-static int qhash_put(void *handle, const uint32_t x, const uint32_t y, const uint64_t value)
+static int OPTIMIZE3 qhash_put(void *handle, const uint32_t x, const uint32_t y, const uint32_t value)
 {
 	sparse_qhash_node_t *node;
 	sparse_qhash_table_t *table = (sparse_qhash_table_t *)handle;
 	size_t hash;
-	uint64_t xy = ((uint64_t)x << 32) | y;
+	const uint64_t xy = ((uint64_t)x << 32) | y;
 
 	if (!table)
 		return -1;
@@ -441,12 +478,12 @@ static int qhash_put(void *handle, const uint32_t x, const uint32_t y, const uin
  *  qhash_get_node()
  *	find the hash table node of a (x,y) value in a hash table
  */
-static sparse_qhash_node_t *qhash_get_node(void *handle, const uint32_t x, const uint32_t y)
+static sparse_qhash_node_t OPTIMIZE3 *qhash_get_node(void *handle, const uint32_t x, const uint32_t y)
 {
 	sparse_qhash_table_t *table = (sparse_qhash_table_t *)handle;
 	sparse_qhash_node_t *node;
 	size_t hash;
-	uint64_t xy = ((uint64_t)x << 32) | y;
+	const uint64_t xy = ((uint64_t)x << 32) | y;
 
 	if (!table)
 		return NULL;
@@ -463,7 +500,7 @@ static sparse_qhash_node_t *qhash_get_node(void *handle, const uint32_t x, const
  *  qhash_get()
  *	get the (x,y) value in hash table based sparse matrix
  */
-static uint64_t qhash_get(void *handle, const uint32_t x, const uint32_t y)
+static uint32_t OPTIMIZE3 qhash_get(void *handle, const uint32_t x, const uint32_t y)
 {
 	sparse_qhash_node_t *node = qhash_get_node(handle, x, y);
 
@@ -488,7 +525,7 @@ static void qhash_del(void *handle, const uint32_t x, const uint32_t y)
  *  judy_create()
  *	create a judy array based sparse matrix
  */
-static void *judy_create(const uint32_t n, const uint32_t x, const uint32_t y)
+static void *judy_create(const uint64_t n, const uint32_t x, const uint32_t y)
 {
 	static Pvoid_t PJLArray;
 
@@ -504,6 +541,8 @@ static void *judy_create(const uint32_t n, const uint32_t x, const uint32_t y)
  *  judy_destroy()
  *	destroy a judy array based sparse matrix
  */
+STRESS_PRAGMA_PUSH
+STRESS_PRAGMA_WARN_OFF
 static void judy_destroy(void *handle, size_t *objmem)
 {
 	Word_t ret;
@@ -514,20 +553,22 @@ static void judy_destroy(void *handle, size_t *objmem)
 	JLFA(ret, *(Pvoid_t *)handle);
 	(void)ret;
 }
+STRESS_PRAGMA_POP
 
 /*
  *  judy_put()
  *	put a value into a judy based sparse matrix
  */
-static int judy_put(void *handle, const uint32_t x, const uint32_t y, const uint64_t value)
+static int OPTIMIZE3 judy_put(void *handle, const uint32_t x, const uint32_t y, const uint32_t value)
 {
 	Word_t *pvalue;
+	const Word_t index = ((Word_t)x << 32) | y;
 
-	JLI(pvalue, *(Pvoid_t *)handle, ((Word_t)x << 32) | y);
+	JLI(pvalue, *(Pvoid_t *)handle, index);
 	if ((pvalue == NULL) || (pvalue == PJERR))
 		return -1;
 
-	*pvalue = (Word_t)value;
+	*pvalue = value;
 	return 0;
 }
 
@@ -535,12 +576,14 @@ static int judy_put(void *handle, const uint32_t x, const uint32_t y, const uint
  *  judy_get()
  *	get the (x,y) value in judy array based sparse matrix
  */
-static uint64_t judy_get(void *handle, const uint32_t x, const uint32_t y)
+static uint32_t OPTIMIZE3 judy_get(void *handle, const uint32_t x, const uint32_t y)
 {
-	Word_t *pvalue;
+	Word_t *pvalue, value;
+	const Word_t index = ((Word_t)x << 32) | y;
 
-	JLG(pvalue, *(Pvoid_t *)handle, ((Word_t)x << 32) | y);
-	return pvalue ? *(uint64_t *)pvalue : 0;
+	JLG(pvalue, *(Pvoid_t *)handle, index);
+	value = pvalue ? *pvalue : 0;
+	return value;
 }
 
 /*
@@ -564,28 +607,31 @@ UNEXPECTED
 static size_t rb_objmem;
 
 /*
- *  sparse_node_cmp()
+ *  sparse_rb_node_cmp()
  *	rb tree comparison function
  */
-static int sparse_node_cmp(sparse_rb_t *n1, sparse_rb_t *n2)
+static int sparse_rb_node_cmp(sparse_rb_t *n1, sparse_rb_t *n2)
 {
-	if (n1->xy == n2->xy)
+	register uint64_t n1xy = n1->xy;
+	register uint64_t n2xy = n2->xy;
+
+	if (n1xy == n2xy)
 		return 0;
-	if (n1->xy > n2->xy)
+	else if (n1xy > n2xy)
 		return 1;
 	else
 		return -1;
 }
 
 static RB_HEAD(sparse_rb_tree, sparse_rb) rb_root;
-RB_PROTOTYPE(sparse_rb_tree, sparse_rb, rb, sparse_node_cmp);
-RB_GENERATE(sparse_rb_tree, sparse_rb, rb, sparse_node_cmp);
+RB_PROTOTYPE(sparse_rb_tree, sparse_rb, rb, sparse_rb_node_cmp);
+RB_GENERATE(sparse_rb_tree, sparse_rb, rb, sparse_rb_node_cmp);
 
 /*
  *  rb_create()
  *	create a red black tree based sparse matrix
  */
-static void *rb_create(const uint32_t n, const uint32_t x, const uint32_t y)
+static void *rb_create(const uint64_t n, const uint32_t x, const uint32_t y)
 {
 	(void)n;
 	(void)x;
@@ -611,7 +657,7 @@ static void rb_destroy(void *handle, size_t *objmem)
  *  rb_put()
  *	put a value into a red black tree sparse matrix
  */
-static int rb_put(void *handle, const uint32_t x, const uint32_t y, const uint64_t value)
+static int OPTIMIZE3 rb_put(void *handle, const uint32_t x, const uint32_t y, const uint32_t value)
 {
 	sparse_rb_t node, *found;
 
@@ -620,13 +666,14 @@ static int rb_put(void *handle, const uint32_t x, const uint32_t y, const uint64
 	if (!found) {
 		sparse_rb_t *new_node;
 
-		new_node = calloc(1, sizeof(*new_node));
+		new_node = malloc(sizeof(*new_node));
 		if (!new_node)
 			return -1;
 		new_node->value = value;
 		new_node->xy = node.xy;
-		RB_INSERT(sparse_rb_tree, handle, new_node);
-		rb_objmem += sizeof(*new_node);
+		if (RB_INSERT(sparse_rb_tree, handle, new_node) != NULL)
+			free(new_node);
+		rb_objmem += sizeof(sparse_rb_t);
 	} else {
 		found->value = value;
 	}
@@ -637,7 +684,7 @@ static int rb_put(void *handle, const uint32_t x, const uint32_t y, const uint64
  *  rb_del()
  *	zero the (x,y) value in red black tree sparse matrix
  */
-static void rb_del(void *handle, const uint32_t x, const uint32_t y)
+static void OPTIMIZE3 rb_del(void *handle, const uint32_t x, const uint32_t y)
 {
 	sparse_rb_t node, *found;
 	node.xy = ((uint64_t)x << 32) | y;
@@ -654,7 +701,7 @@ static void rb_del(void *handle, const uint32_t x, const uint32_t y)
  *  rb_get()
  *	get the (x,y) value in a red back tree sparse matrix
  */
-static uint64_t rb_get(void *handle, const uint32_t x, const uint32_t y)
+static uint32_t OPTIMIZE3 rb_get(void *handle, const uint32_t x, const uint32_t y)
 {
 	sparse_rb_t node, *found;
 
@@ -668,13 +715,135 @@ static uint64_t rb_get(void *handle, const uint32_t x, const uint32_t y)
 UNEXPECTED
 #endif
 
+#if defined(HAVE_SPLAY_TREE)
+
+static size_t splay_objmem;
+
+/*
+ *  sparse_splay_node_cmp()
+ *	splay tree comparison function
+ */
+static int sparse_splay_node_cmp(sparse_splay_t *n1, sparse_splay_t *n2)
+{
+	register uint64_t n1xy = n1->xy;
+	register uint64_t n2xy = n2->xy;
+
+	if (n1xy == n2xy)
+		return 0;
+	else if (n1xy > n2xy)
+		return 1;
+	else
+		return -1;
+}
+
+static SPLAY_HEAD(sparse_splay_tree, sparse_splay) splay_root;
+SPLAY_PROTOTYPE(sparse_splay_tree, sparse_splay, splay, sparse_splay_node_cmp);
+SPLAY_GENERATE(sparse_splay_tree, sparse_splay, splay, sparse_splay_node_cmp);
+
+/*
+ *  splay_create()
+ *	create a splay tree based sparse matrix
+ */
+static void *splay_create(const uint64_t n, const uint32_t x, const uint32_t y)
+{
+	sparse_splay_t *node;
+
+	(void)n;
+	(void)x;
+	(void)y;
+
+	splay_objmem = 0;
+	SPLAY_INIT(&splay_root);
+
+	/* The following just silence warnings about unused SPLAY funcs */
+	node = SPLAY_MIN(sparse_splay_tree, &splay_root);
+	/* Node is always NULL, so this is a no-op */
+	if (node)
+		(void)SPLAY_NEXT(sparse_splay_tree, &splay_root, node);
+
+	return &splay_root;
+}
+
+/*
+ *  splay_destroy()
+ *	destroy a splay tree based sparse matrix
+ */
+static void splay_destroy(void *handle, size_t *objmem)
+{
+	*objmem = splay_objmem;
+	splay_objmem = 0;
+	(void)handle;
+}
+
+/*
+ *  splay_put()
+ *	put a value into a splay tree sparse matrix
+ */
+static int OPTIMIZE3 splay_put(void *handle, const uint32_t x, const uint32_t y, const uint32_t value)
+{
+	sparse_splay_t node, *found;
+
+	node.xy = ((uint64_t)x << 32) | y;
+	found = SPLAY_FIND(sparse_splay_tree, handle, &node);
+	if (!found) {
+		sparse_splay_t *new_node;
+
+		new_node = malloc(sizeof(*new_node));
+		if (!new_node)
+			return -1;
+		new_node->value = value;
+		new_node->xy = node.xy;
+		if (SPLAY_INSERT(sparse_splay_tree, handle, new_node) != NULL)
+			free(new_node);
+		splay_objmem += sizeof(sparse_splay_t);
+	} else {
+		found->value = value;
+	}
+	return 0;
+}
+
+/*
+ *  splay_del()
+ *	zero the (x,y) value in splay tree sparse matrix
+ */
+static void OPTIMIZE3 splay_del(void *handle, const uint32_t x, const uint32_t y)
+{
+	sparse_splay_t node, *found;
+	node.xy = ((uint64_t)x << 32) | y;
+
+	found = SPLAY_FIND(sparse_splay_tree, handle, &node);
+	if (!found)
+		return;
+
+	SPLAY_REMOVE(sparse_splay_tree, handle, found);
+	free(found);
+}
+
+/*
+ *  splay_get()
+ *	get the (x,y) value in a splay tree sparse matrix
+ */
+static uint32_t OPTIMIZE3 splay_get(void *handle, const uint32_t x, const uint32_t y)
+{
+	sparse_splay_t node, *found;
+
+	memset(&node, 0xff, sizeof(node));
+	node.xy = ((uint64_t)x << 32) | y;
+
+	found = SPLAY_FIND(sparse_splay_tree, handle, &node);
+	return found ? found->value : 0;
+}
+#else
+UNEXPECTED
+#endif
+
 #if defined(HAVE_SYS_QUEUE_CIRCLEQ)
 
 /*
  *  list_create()
  *	create a circular list based sparse matrix
  */
-static void *list_create(const uint32_t n, const uint32_t x, const uint32_t y)
+static void *list_create(const uint64_t n, const uint32_t x, const uint32_t y)
 {
 	static sparse_y_list_t y_head;
 
@@ -717,7 +886,7 @@ static void list_destroy(void *handle, size_t *objmem)
  *  list_put()
  *	put a value into a circular list based sparse matrix
  */
-static int list_put(void *handle, const uint32_t x, const uint32_t y, const uint64_t value)
+static int OPTIMIZE3 list_put(void *handle, const uint32_t x, const uint32_t y, const uint32_t value)
 {
 	sparse_y_list_t *y_head = (sparse_y_list_t *)handle;
 	sparse_y_list_node_t *y_node, *new_y_node;
@@ -731,7 +900,7 @@ static int list_put(void *handle, const uint32_t x, const uint32_t y, const uint
 			goto find_x;
 		}
 		if (y_node->y > y) {
-			new_y_node = calloc(1, sizeof(*new_y_node));
+			new_y_node = malloc(sizeof(*new_y_node));
 			if (!new_y_node)
 				return -1;
 			new_y_node->y = y;
@@ -742,7 +911,7 @@ static int list_put(void *handle, const uint32_t x, const uint32_t y, const uint
 		}
 	}
 
-	new_y_node = calloc(1, sizeof(*new_y_node));
+	new_y_node = malloc(sizeof(*new_y_node));
 	if (!new_y_node)
 		return -1;
 	new_y_node->y = y;
@@ -757,7 +926,7 @@ find_x:
 			return 0;
 		}
 		if (x_node->x > x) {
-			new_x_node = calloc(1, sizeof(*new_x_node));
+			new_x_node = malloc(sizeof(*new_x_node));
 			if (!new_x_node)
 				return -1;  /* Leaves new_y_node allocated */
 			new_x_node->x = x;
@@ -780,7 +949,7 @@ find_x:
  *	find the circular list node of a (x,y) value in a circular
  *	list based sparse matrix
  */
-static sparse_x_list_node_t *list_get_node(void *handle, const uint32_t x, const uint32_t y)
+static sparse_x_list_node_t OPTIMIZE3 *list_get_node(void *handle, const uint32_t x, const uint32_t y)
 {
 	sparse_y_list_t *y_head = (sparse_y_list_t *)handle;
 	sparse_y_list_node_t *y_node;
@@ -815,7 +984,7 @@ static void list_del(void *handle, const uint32_t x, const uint32_t y)
  *  list_get()
  *	get the (x,y) value in a circular list based sparse matrix
  */
-static uint64_t list_get(void *handle, const uint32_t x, const uint32_t y)
+static uint32_t OPTIMIZE3 list_get(void *handle, const uint32_t x, const uint32_t y)
 {
 	const sparse_x_list_node_t *x_node = list_get_node(handle, x, y);
 
@@ -825,12 +994,115 @@ static uint64_t list_get(void *handle, const uint32_t x, const uint32_t y)
 UNEXPECTED
 #endif
 
-static uint64_t value_map(const uint32_t x, const uint32_t y)
+#if defined(HAVE_JUDY)
+
+/*
+ *  hashjudy_create()
+ *	create a hash table of Judy arrays
+ */
+static void *hashjudy_create(const uint64_t n, const uint32_t x, const uint32_t y)
 {
-	return ((uint64_t)x << 32) ^ y;
-	//return ((uint64_t)~x << 11) ^ y;
+	sparse_hashjudy_table_t *table;
+
+	(void)n;
+	(void)y;
+
+	table = (sparse_hashjudy_table_t *)calloc(1, sizeof(*table));
+	if (!table)
+		return NULL;
+
+	table->n = x;
+	table->hash_table = (Pvoid_t *)calloc((size_t)x, sizeof(*table->hash_table));
+	if (!table->hash_table) {
+		free(table);
+		return NULL;
+	}
+	return (void *)table;
 }
 
+/*
+ *  hashjudy_destroy()
+ *	destroy a hash table of Judy arrays
+ */
+static void hashjudy_destroy(void *handle, size_t *objmem)
+{
+	sparse_hashjudy_table_t *table = (sparse_hashjudy_table_t *)handle;
+	uint64_t i;
+
+	*objmem = sizeof(*table) +
+		  (size_t)table->n * sizeof(*table->hash_table);
+
+	for (i = 0; i < table->n; i++) {
+		Word_t ret;
+
+		JLMU(ret, table->hash_table[i]);
+		*objmem += (size_t)ret;
+	}
+
+	free(table->hash_table);
+	table->n = 0;
+	table->hash_table = NULL;
+	free(table);
+}
+
+/*
+ *  hashjudy_put()
+ *	put a value into a hash based judy array
+ */
+static int OPTIMIZE3 hashjudy_put(void *handle, const uint32_t x, const uint32_t y, const uint32_t value)
+{
+	sparse_hashjudy_table_t *table = (sparse_hashjudy_table_t *)handle;
+	Word_t *pvalue;
+
+	if (!table)
+		return -1;
+
+	JLI(pvalue, table->hash_table[x], y);
+	if ((pvalue == NULL) || (pvalue == PJERR))
+		return -1;
+	*pvalue = (Word_t)value;
+	return 0;
+}
+
+/*
+ *  hashjudy_del()
+ *	put a value into a hash based judy array
+ */
+static void OPTIMIZE3 hashjudy_del(void *handle, const uint32_t x, const uint32_t y)
+{
+	sparse_hashjudy_table_t *table = (sparse_hashjudy_table_t *)handle;
+	Word_t *pvalue;
+
+	if (!table)
+		return;
+
+	JLG(pvalue, table->hash_table[x], y);
+	if ((pvalue == NULL) || (pvalue == PJERR))
+		return;
+	*pvalue = 0;
+}
+
+/*
+ *  hashjudy_get()
+ *	get the (x,y) value in hash judy array based sparse matrix
+ */
+static uint32_t OPTIMIZE3 hashjudy_get(void *handle, const uint32_t x, const uint32_t y)
+{
+	sparse_hashjudy_table_t *table = (sparse_hashjudy_table_t *)handle;
+	Word_t *pvalue, value;
+
+	if (!table)
+		return -1;
+
+	JLG(pvalue, table->hash_table[x], y);
+	if ((pvalue == NULL) || (pvalue == PJERR))
+		return 0;
+	value = *pvalue;
+	return (uint32_t)value;
+}
+
+#endif
+
 static int stress_sparse_method_test(
 	const stress_args_t *args,
 	const uint64_t sparsematrix_items,
@@ -857,21 +1129,20 @@ static int stress_sparse_method_test(
 
 	t1 = stress_time_now();
 	for (i = 0; keep_stressing_flag() && (i < sparsematrix_items); i++) {
-		const uint32_t x = stress_mwc32() % sparsematrix_size;
-		const uint32_t y = stress_mwc32() % sparsematrix_size;
-		uint64_t v = value_map(x, y);
-		uint64_t gv;
+		const uint32_t x = stress_mwc32modn(sparsematrix_size);
+		const uint32_t y = stress_mwc32modn(sparsematrix_size);
+		uint32_t gv, v = value_map(x, y);
 
 		if (v == 0)
-			v = ~0ULL;
+			v = ~(uint32_t)0;
 
 		gv = info->get(handle, x, y);
 		if (gv == 0) {
 			if (info->put(handle, x, y, v) < 0) {
-				pr_fail("%s: failed to put into "
+				pr_fail("%s: %s failed to put into "
 					"sparse matrix at position "
 					"(%" PRIu32 ",%" PRIu32 ")\n",
-					args->name, x, y);
+					args->name, info->name, x, y);
 				rc = SPARSE_TEST_FAILED;
 				goto err;
 			}
@@ -884,19 +1155,18 @@ static int stress_sparse_method_test(
 	stress_mwc_set_seed(w, z);
 	t1 = stress_time_now();
 	for (i = 0; keep_stressing_flag() && (i < sparsematrix_items); i++) {
-		const uint32_t x = stress_mwc32() % sparsematrix_size;
-		const uint32_t y = stress_mwc32() % sparsematrix_size;
-		uint64_t v = value_map(x, y);
-		uint64_t gv;
+		const uint32_t x = stress_mwc32modn(sparsematrix_size);
+		const uint32_t y = stress_mwc32modn(sparsematrix_size);
+		uint32_t gv, v = value_map(x, y);
 
 		if (v == 0)
-			v = ~0ULL;
+			v = ~(uint32_t)0;
 
 		gv = info->get(handle, x, y);
 		if (gv != v) {
-			pr_fail("%s: mismatch (%" PRIu32 ",%" PRIu32
-				") was %" PRIx64 ", got %" PRIx64 "\n",
-				args->name, x, y, v, gv);
+			pr_fail("%s: %s mismatch (%" PRIu32 ",%" PRIu32
+				") was %" PRIu32 ", got %" PRIu32 "\n",
+				args->name, info->name, x, y, v, gv);
 		}
 	}
 	t2 = stress_time_now();
@@ -906,8 +1176,8 @@ static int stress_sparse_method_test(
 	/* Random fetches, most probably all zero unset values */
 	t1 = stress_time_now();
 	for (i = 0; keep_stressing_flag() && (i < sparsematrix_items); i++) {
-		const uint32_t x = stress_mwc32() % sparsematrix_size;
-		const uint32_t y = stress_mwc32() % sparsematrix_size;
+		const uint32_t x = stress_mwc32modn(sparsematrix_size);
+		const uint32_t y = stress_mwc32modn(sparsematrix_size);
 
 		(void)info->get(handle, x, y);
 	}
@@ -917,9 +1187,9 @@ static int stress_sparse_method_test(
 
 	stress_mwc_set_seed(w, z);
 	for (i = 0; keep_stressing_flag() && (i < sparsematrix_items); i++) {
-		const uint32_t x = stress_mwc32() % sparsematrix_size;
-		const uint32_t y = stress_mwc32() % sparsematrix_size;
-		uint64_t v = value_map(x, y);
+		const uint32_t x = stress_mwc32modn(sparsematrix_size);
+		const uint32_t y = stress_mwc32modn(sparsematrix_size);
+		uint32_t v = value_map(x, y);
 		(void)v;
 
 		info->del(handle, x, y);
@@ -932,13 +1202,14 @@ err:
 	return rc;
 }
 
-static void *mmap_create(const uint32_t n, const uint32_t x, const uint32_t y)
+static void *mmap_create(const uint64_t n, const uint32_t x, const uint32_t y)
 {
-	const size_t page_size = stress_get_pagesize();
+	const size_t page_size = stress_get_page_size();
 	static sparse_mmap_t m;
-	size_t shmall, freemem, totalmem, freeswap, max_phys;
+	size_t shmall, freemem, totalmem, freeswap, totalswap;
+	uint64_t max_phys, total_free, max_size_t;
 
-	stress_get_memlimits(&shmall, &freemem, &totalmem, &freeswap);
+	stress_get_memlimits(&shmall, &freemem, &totalmem, &freeswap, &totalswap);
 
 	(void)n;
 
@@ -949,10 +1220,15 @@ static void *mmap_create(const uint32_t n, const uint32_t x, const uint32_t y)
 	 */
 	max_phys = n * page_size * 2;
 
-	m.mmap_size = (size_t)x * (size_t)y * sizeof(uint64_t);
-	m.mmap_size = (m.mmap_size + page_size - 1) & ~(page_size - 1);
+	m.mmap_size = (uint64_t)x * (uint64_t)y * sizeof(uint32_t);
+	m.mmap_size = (m.mmap_size + page_size - 1) & (uint64_t)~(page_size - 1);
 
-	if (max_phys > freemem + freeswap)
+	total_free = (uint64_t)freemem + (uint64_t)freeswap;
+	if (max_phys > total_free)
+		return NULL;
+
+	max_size_t = (uint64_t)(~(size_t)0);
+	if (m.mmap_size >= max_size_t)
 		return NULL;
 
 	m.x = x;
@@ -968,7 +1244,7 @@ static void *mmap_create(const uint32_t n, const uint32_t x, const uint32_t y)
 static void mmap_destroy(void *handle, size_t *objmem)
 {
 	sparse_mmap_t *m = (sparse_mmap_t *)handle;
-	const size_t page_size = stress_get_pagesize();
+	const size_t page_size = stress_get_page_size();
 	unsigned char *vec;
 	size_t pages;
 
@@ -1000,7 +1276,7 @@ static void mmap_destroy(void *handle, size_t *objmem)
 	(void)munmap(m->mmap, m->mmap_size);
 }
 
-static int mmap_put(void *handle, const uint32_t x, const uint32_t y, const uint64_t value)
+static int mmap_put(void *handle, const uint32_t x, const uint32_t y, const uint32_t value)
 {
 	sparse_mmap_t *m = (sparse_mmap_t *)handle;
 	off_t offset;
@@ -1008,8 +1284,8 @@ static int mmap_put(void *handle, const uint32_t x, const uint32_t y, const uint
 	if (m->x <= x || m->y <= y)
 		return -1;
 
-	offset = (x + ((off_t)m->y * y));
-	*((uint64_t *)(m->mmap) + offset) = value;
+	offset = (x + ((uint64_t)m->y * y));
+	*((uint32_t *)(m->mmap) + offset) = value;
 
 	return 0;
 }
@@ -1017,25 +1293,25 @@ static int mmap_put(void *handle, const uint32_t x, const uint32_t y, const uint
 static void mmap_del(void *handle, const uint32_t x, const uint32_t y)
 {
 	sparse_mmap_t *m = (sparse_mmap_t *)handle;
-	off_t offset;
+	uint64_t offset;
 
 	if (m->x <= x || m->y <= y)
 		return;
 
-	offset = (x + ((off_t)m->y * y));
-	*((uint64_t *)(m->mmap) + offset) = 0;
+	offset = (x + ((uint64_t)m->y * y));
+	*((uint32_t *)(m->mmap) + offset) = 0;
 }
 
-static uint64_t mmap_get(void *handle, const uint32_t x, const uint32_t y)
+static uint32_t mmap_get(void *handle, const uint32_t x, const uint32_t y)
 {
 	sparse_mmap_t *m = (sparse_mmap_t *)handle;
-	size_t offset;
+	uint64_t offset;
 
 	if (m->x <= x || m->y <= y)
-		return -1;
+		return (uint32_t)-1;
 
-	offset = (x + ((off_t)m->y * y));
-	return *((uint64_t *)(m->mmap) + offset);
+	offset = (x + ((uint64_t)m->y * y));
+	return *((uint32_t *)(m->mmap) + offset);
 }
 
 /*
@@ -1045,6 +1321,7 @@ static const stress_sparsematrix_method_info_t sparsematrix_methods[] = {
 	{ "all",	NULL, NULL, NULL, NULL, NULL },
 	{ "hash",	hash_create, hash_destroy, hash_put, hash_del, hash_get },
 #if defined(HAVE_JUDY)
+	{ "hashjudy",	hashjudy_create, hashjudy_destroy, hashjudy_put, hashjudy_del, hashjudy_get },
 	{ "judy",	judy_create, judy_destroy, judy_put, judy_del, judy_get },
 #endif
 #if defined(HAVE_SYS_QUEUE_CIRCLEQ)
@@ -1054,6 +1331,9 @@ static const stress_sparsematrix_method_info_t sparsematrix_methods[] = {
 	{ "qhash",	qhash_create, qhash_destroy, qhash_put, qhash_del, qhash_get },
 #if defined(HAVE_RB_TREE)
 	{ "rb",		rb_create, rb_destroy, rb_put, rb_del, rb_get },
+#endif
+#if defined(HAVE_SPLAY_TREE)
+	{ "splay",	splay_create, splay_destroy, splay_put, splay_del, splay_get },
 #endif
 	{ NULL,		NULL, NULL, NULL, NULL, NULL },
 };
@@ -1109,7 +1389,6 @@ static int stress_sparsematrix(const stress_args_t *args)
 	test_info_t test_info[SIZEOF_ARRAY(sparsematrix_methods)];
 	size_t i, begin, end;
 	size_t method = 0;	/* All methods */
-	bool lock = false;
 
 	for (i = 0; i < SIZEOF_ARRAY(test_info); i++) {
 		test_info[i].skip_no_mem = false;
@@ -1136,7 +1415,7 @@ static int stress_sparsematrix(const stress_args_t *args)
 			sparsematrix_items= MIN_SPARSEMATRIX_ITEMS;
 	}
 
-	capacity = sparsematrix_size * sparsematrix_size;
+	capacity = (uint64_t)sparsematrix_size * (uint64_t)sparsematrix_size;
 
 	if (sparsematrix_items > capacity) {
 		uint64_t new_items = capacity;
@@ -1162,7 +1441,7 @@ static int stress_sparsematrix(const stress_args_t *args)
 			for (i = 1; sparsematrix_methods[i].name; i++) {
 				if (stress_sparse_method_test(args,
 						(size_t)sparsematrix_items,
-						(size_t)sparsematrix_size,
+						sparsematrix_size,
 						&sparsematrix_methods[i],
 						&test_info[i]) == SPARSE_TEST_FAILED) {
 					stress_sparsematrix_create_failed(args, sparsematrix_methods[i].name);
@@ -1172,7 +1451,7 @@ static int stress_sparsematrix(const stress_args_t *args)
 		} else {
 			if (stress_sparse_method_test(args,
 					(size_t)sparsematrix_items,
-					(size_t)sparsematrix_size,
+					sparsematrix_size,
 					&sparsematrix_methods[method],
 					&test_info[method]) == SPARSE_TEST_FAILED) {
 				stress_sparsematrix_create_failed(args, sparsematrix_methods[method].name);
@@ -1185,14 +1464,14 @@ static int stress_sparsematrix(const stress_args_t *args)
 
 	if (method == 0) {	/* All methods */
 		begin = 1;
-		end = ~0;
+		end = SIZEOF_ARRAY(sparsematrix_methods);
 	} else {
 		begin = method;
 		end = method + 1;
 	}
 
-	pr_lock(&lock);
-	for (i = begin; i < end && sparsematrix_methods[i].name; i++) {
+	pr_lock();
+	for (i = begin; (i < end) && sparsematrix_methods[i].name; i++) {
 		char str[12];
 
 		if (test_info[i].max_objmem) {
@@ -1202,19 +1481,19 @@ static int stress_sparsematrix(const stress_args_t *args)
 		}
 
 		if (test_info[i].skip_no_mem) {
-			pr_inf_lock(&lock, "%s: %-6s skipped (out of memory)\n",
+			pr_inf("%s: %-6s skipped (out of memory)\n",
 				args->name, sparsematrix_methods[i].name);
 		} else {
-			pr_inf_lock(&lock, "%s: %-6s %8.8s %15.2f Get/s %15.2f Put/s\n",
+			pr_inf("%s: %-8s %8.8s %15.2f get/s %15.2f put/s\n",
 				args->name,
 				sparsematrix_methods[i].name, str,
 				test_info[i].get_duration > 0.0 ?
-					test_info[i].get_ops / test_info[i].get_duration : 0.0,
+					(double)test_info[i].get_ops / test_info[i].get_duration : 0.0,
 				test_info[i].put_duration > 0.0 ?
-					test_info[i].put_ops / test_info[i].put_duration : 0.0);
+					(double)test_info[i].put_ops / test_info[i].put_duration : 0.0);
 		}
 	}
-	pr_unlock(&lock);
+	pr_unlock();
 
 	rc = EXIT_SUCCESS;
 err:
