@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C)      2022 Colin Ian King.
+ * Copyright (C) 2022-2023 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -18,13 +18,15 @@
  *
  */
 #include "stress-ng.h"
+#include "core-sort.h"
 
 #if defined(HAVE_SEARCH_H)
 #include <search.h>
 #endif
 
+#define TSEARCH_SIZE_SHIFT	(22)
 #define MIN_TSEARCH_SIZE	(1 * KB)
-#define MAX_TSEARCH_SIZE	(4 * MB)
+#define MAX_TSEARCH_SIZE	(1U << TSEARCH_SIZE_SHIFT)	/* 4 MB */
 #define DEFAULT_TSEARCH_SIZE	(64 * KB)
 
 static const stress_help_t help[] = {
@@ -55,23 +57,6 @@ static const stress_opt_set_func_t opt_set_funcs[] = {
 
 #if defined(HAVE_TSEARCH)
 
-/*
- *  cmp()
- *	sort on int32 values
- */
-static int cmp(const void *p1, const void *p2)
-{
-	const int32_t *i1 = (const int32_t *)p1;
-	const int32_t *i2 = (const int32_t *)p2;
-
-	if (*i1 > *i2)
-		return 1;
-	else if (*i1 < *i2)
-		return -1;
-	else
-		return 0;
-}
-
 /*
  *  stress_tsearch()
  *	stress tsearch
@@ -81,6 +66,7 @@ static int stress_tsearch(const stress_args_t *args)
 	uint64_t tsearch_size = DEFAULT_TSEARCH_SIZE;
 	int32_t *data;
 	size_t i, n;
+	double rate, duration = 0.0, count = 0.0, sorted = 0.0;
 
 	if (!stress_get_setting("tsearch-size", &tsearch_size)) {
 		if (g_opt_flags & OPT_FLAGS_MAXIMIZE)
@@ -91,31 +77,38 @@ static int stress_tsearch(const stress_args_t *args)
 	n = (size_t)tsearch_size;
 
 	if ((data = calloc(n, sizeof(*data))) == NULL) {
-		pr_fail("%s: calloc failed, out of memory\n", args->name);
+		pr_fail("%s: calloc failed allocating %zd integers, skipping stressor\n",
+			args->name, n);
 		return EXIT_NO_RESOURCE;
 	}
 
 	stress_set_proc_state(args->name, STRESS_STATE_RUN);
 
+	stress_sort_data_int32_init(data, n);
+
 	do {
+		double t;
 		void *root = NULL;
 
+		stress_sort_data_int32_shuffle(data, n);
+
 		/* Step #1, populate tree */
 		for (i = 0; i < n; i++) {
-			data[i] = (int32_t)(((stress_mwc32() & 0xfff) << 20) ^ i);
-			if (tsearch(&data[i], &root, cmp) == NULL) {
+			if (tsearch(&data[i], &root, stress_sort_cmp_int32) == NULL) {
 				size_t j;
 
 				pr_err("%s: cannot allocate new "
 					"tree node\n", args->name);
 				for (j = 0; j < i; j++)
-					tdelete(&data[j], &root, cmp);
+					tdelete(&data[j], &root, stress_sort_cmp_int32);
 				goto abort;
 			}
 		}
 		/* Step #2, find */
+		stress_sort_compare_reset();
+		t = stress_time_now();
 		for (i = 0; keep_stressing_flag() && i < n; i++) {
-			const void **result = tfind(&data[i], &root, cmp);
+			const void **result = tfind(&data[i], &root, stress_sort_cmp_int32);
 
 			if (g_opt_flags & OPT_FLAGS_VERIFY) {
 				if (!result) {
@@ -133,9 +126,13 @@ static int stress_tsearch(const stress_args_t *args)
 				}
 			}
 		}
+		duration += stress_time_now() - t;
+		count += (double)stress_sort_compare_get();
+		sorted += (double)i;
+
 		/* Step #3, delete */
 		for (i = 0; i < n; i++) {
-			const void **result = tdelete(&data[i], &root, cmp);
+			const void **result = tdelete(&data[i], &root, stress_sort_cmp_int32);
 
 			if ((g_opt_flags & OPT_FLAGS_VERIFY) && (result == NULL)) {
 				pr_fail("%s: element %zu could not be found\n",
@@ -144,10 +141,13 @@ static int stress_tsearch(const stress_args_t *args)
 		}
 		inc_counter(args);
 	} while (keep_stressing(args));
-
 abort:
 	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 
+	rate = (duration > 0.0) ? count / duration : 0.0;
+	stress_metrics_set(args, 0, "tsearch comparisons per sec", rate);
+	stress_metrics_set(args, 1, "tsearch comparisons per item", count / sorted);
+
 	free(data);
 	return EXIT_SUCCESS;
 }
@@ -163,11 +163,12 @@ stressor_info_t stress_tsearch_info = {
 #else
 
 stressor_info_t stress_tsearch_info = {
-	.stressor = stress_not_implemented,
+	.stressor = stress_unimplemented,
 	.class = CLASS_CPU_CACHE | CLASS_CPU | CLASS_MEMORY,
 	.opt_set_funcs = opt_set_funcs,
 	.verify = VERIFY_OPTIONAL,
-	.help = help
+	.help = help,
+	.unimplemented_reason = "built without libc tsearch() support"
 };
 
 #endif
