@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C)      2022 Colin Ian King.
+ * Copyright (C) 2022-2023 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -18,6 +18,7 @@
  *
  */
 #include "stress-ng.h"
+#include "core-builtin.h"
 #include "core-cache.h"
 #include "core-target-clones.h"
 #include "core-nt-load.h"
@@ -73,12 +74,12 @@ static const stress_help_t help[] = {
 	{ NULL,	 "vm-bytes N",	 "allocate N bytes per vm worker (default 256MB)" },
 	{ NULL,	 "vm-hang N",	 "sleep N seconds before freeing memory" },
 	{ NULL,	 "vm-keep",	 "redirty memory instead of reallocating" },
-	{ NULL,	 "vm-ops N",	 "stop after N vm bogo operations" },
 #if defined(MAP_LOCKED)
 	{ NULL,	 "vm-locked",	 "lock the pages of the mapped region into memory" },
 #endif
 	{ NULL,	 "vm-madvise M", "specify mmap'd vm buffer madvise advice" },
 	{ NULL,	 "vm-method M",	 "specify stress vm method M, default is all" },
+	{ NULL,	 "vm-ops N",	 "stop after N vm bogo operations" },
 #if defined(MAP_POPULATE)
 	{ NULL,	 "vm-populate",	 "populate (prefault) page tables for a mapping" },
 #endif
@@ -213,15 +214,6 @@ do {						\
 	bit_errors += (*ptr != val);		\
 } while (0)
 
-#define ROR8(val) 				\
-do {						\
-	uint8_t tmp = val;			\
-	const uint8_t bit0 = (uint8_t)((tmp & 1) << 7);	\
-	tmp >>= 1;				\
-	tmp |= bit0;				\
-	val = tmp;				\
-} while (0)
-
 #define INC_LO_NYBBLE(val)			\
 do {						\
 	uint8_t lo = (val);			\
@@ -252,23 +244,23 @@ static void inject_random_bit_errors(uint8_t *buf, const size_t sz)
 
 	for (i = 0; i < 8; i++) {
 		/* 1 bit errors */
-		buf[stress_mwc64() % sz] ^= (1 << i);
-		buf[stress_mwc64() % sz] |= (1 << i);
-		buf[stress_mwc64() % sz] &= ~(1 << i);
+		buf[stress_mwc64modn(sz)] ^= (1 << i);
+		buf[stress_mwc64modn(sz)] |= (1 << i);
+		buf[stress_mwc64modn(sz)] &= ~(1 << i);
 	}
 
 	for (i = 0; i < 7; i++) {
 		/* 2 bit errors */
-		buf[stress_mwc64() % sz] ^= (3 << i);
-		buf[stress_mwc64() % sz] |= (3 << i);
-		buf[stress_mwc64() % sz] &= ~(3 << i);
+		buf[stress_mwc64modn(sz)] ^= (3 << i);
+		buf[stress_mwc64modn(sz)] |= (3 << i);
+		buf[stress_mwc64modn(sz)] &= ~(3 << i);
 	}
 
 	for (i = 0; i < 6; i++) {
 		/* 3 bit errors */
-		buf[stress_mwc64() % sz] ^= (7 << i);
-		buf[stress_mwc64() % sz] |= (7 << i);
-		buf[stress_mwc64() % sz] &= ~(7 << i);
+		buf[stress_mwc64modn(sz)] ^= (7 << i);
+		buf[stress_mwc64modn(sz)] |= (7 << i);
+		buf[stress_mwc64modn(sz)] &= ~(7 << i);
 	}
 }
 #else
@@ -303,12 +295,15 @@ static void stress_vm_check(const char *name, const size_t bit_errors)
  */
 static inline size_t stress_vm_count_bits8(uint8_t v)
 {
+#if defined(HAVE_BUILTIN_POPCOUNT)
+	return (size_t)__builtin_popcount((unsigned int)v);
+#else
 	size_t n;
 
 	for (n = 0; v; n++)
 		v &= v - 1;
-
 	return n;
+#endif
 }
 
 /*
@@ -317,12 +312,37 @@ static inline size_t stress_vm_count_bits8(uint8_t v)
  */
 static inline size_t stress_vm_count_bits(uint64_t v)
 {
+#if defined(HAVE_BUILTIN_POPCOUNTLL)
+	if (sizeof(unsigned long long) == sizeof(uint64_t)) {
+		return (size_t)__builtin_popcountll((unsigned long long)v);
+	}
+#endif
+#if defined(HAVE_BUILTIN_POPCOUNTL)
+	if (sizeof(unsigned long) == sizeof(uint64_t)) {
+		return (size_t)__builtin_popcountl((unsigned long)v);
+	} else if (sizeof(unsigned long) == sizeof(uint32_t)) {
+		unsigned long lo, hi;
+
+		lo = (unsigned long)(v >> 32);
+		hi = (unsigned long)(v & 0xffffffff);
+		return (size_t)__builtin_popcountl(hi) + __builtin_popcountl(lo);
+	}
+#endif
+#if defined(HAVE_BUILTIN_POPCOUNT)
+	if (sizeof(unsigned int) == sizeof(uint32_t)) {
+		unsigned int lo, hi;
+
+		lo = (unsigned int)(v >> 32);
+		hi = (unsigned int)(v & 0xffffffff);
+		return (size_t)__builtin_popcount(hi) + __builtin_popcount(lo);
+	}
+#else
 	size_t n;
 
 	for (n = 0; v; n++)
 		v &= v - 1;
-
 	return n;
+#endif
 }
 
 /*
@@ -907,7 +927,7 @@ static size_t TARGET_CLONES stress_vm_swap(
 	}
 
 	for (i = 0; i < chunks; i++) {
-		swaps[i] = (stress_mwc64() % chunks) * chunk_sz;
+		swaps[i] = (stress_mwc64modn(chunks)) * chunk_sz;
 	}
 
 	stress_mwc_set_seed(w1, z1);
@@ -1094,14 +1114,14 @@ static size_t TARGET_CLONES stress_vm_ror(
 	(void)stress_mincore_touch_pages(buf, sz);
 
 	for (ptr = (uint8_t *)buf; ptr < (uint8_t *)buf_end; ptr += chunk_sz) {
-		ROR8(*(ptr + 0));
-		ROR8(*(ptr + 1));
-		ROR8(*(ptr + 2));
-		ROR8(*(ptr + 3));
-		ROR8(*(ptr + 4));
-		ROR8(*(ptr + 5));
-		ROR8(*(ptr + 6));
-		ROR8(*(ptr + 7));
+		*(ptr + 0) = shim_ror8(*(ptr + 0));
+		*(ptr + 1) = shim_ror8(*(ptr + 1));
+		*(ptr + 2) = shim_ror8(*(ptr + 2));
+		*(ptr + 3) = shim_ror8(*(ptr + 3));
+		*(ptr + 4) = shim_ror8(*(ptr + 4));
+		*(ptr + 5) = shim_ror8(*(ptr + 5));
+		*(ptr + 6) = shim_ror8(*(ptr + 6));
+		*(ptr + 7) = shim_ror8(*(ptr + 7));
 
 		c++;
 		if (UNLIKELY(max_ops && (c >= max_ops)))
@@ -1116,7 +1136,7 @@ static size_t TARGET_CLONES stress_vm_ror(
 	stress_mwc_set_seed(w, z);
 	for (ptr = (uint8_t *)buf; ptr < (uint8_t *)buf_end; ptr += chunk_sz) {
 		uint8_t val = stress_mwc8();
-		ROR8(val);
+		val = shim_ror8(val);
 
 		bit_errors += (*(ptr + 0) != val);
 		bit_errors += (*(ptr + 1) != val);
@@ -1166,19 +1186,19 @@ static size_t TARGET_CLONES stress_vm_flip(
 		uint8_t val = stress_mwc8();
 
 		*(ptr + 0) = val;
-		ROR8(val);
+		val = shim_ror8(val);
 		*(ptr + 1) = val;
-		ROR8(val);
+		val = shim_ror8(val);
 		*(ptr + 2) = val;
-		ROR8(val);
+		val = shim_ror8(val);
 		*(ptr + 3) = val;
-		ROR8(val);
+		val = shim_ror8(val);
 		*(ptr + 4) = val;
-		ROR8(val);
+		val = shim_ror8(val);
 		*(ptr + 5) = val;
-		ROR8(val);
+		val = shim_ror8(val);
 		*(ptr + 6) = val;
-		ROR8(val);
+		val = shim_ror8(val);
 		*(ptr + 7) = val;
 		c++;
 		if (UNLIKELY(max_ops && (c >= max_ops)))
@@ -1189,7 +1209,7 @@ static size_t TARGET_CLONES stress_vm_flip(
 	(void)stress_mincore_touch_pages(buf, sz);
 
 	for (i = 0; i < 8; i++) {
-		ROR8(bit);
+		bit = shim_ror8(bit);
 		for (ptr = (uint8_t *)buf; ptr < (uint8_t *)buf_end; ptr += chunk_sz) {
 			*(ptr + 0) ^= bit;
 			*(ptr + 1) ^= bit;
@@ -1215,19 +1235,19 @@ static size_t TARGET_CLONES stress_vm_flip(
 		uint8_t val = stress_mwc8();
 
 		bit_errors += (*(ptr + 0) != val);
-		ROR8(val);
+		val = shim_ror8(val);
 		bit_errors += (*(ptr + 1) != val);
-		ROR8(val);
+		val = shim_ror8(val);
 		bit_errors += (*(ptr + 2) != val);
-		ROR8(val);
+		val = shim_ror8(val);
 		bit_errors += (*(ptr + 3) != val);
-		ROR8(val);
+		val = shim_ror8(val);
 		bit_errors += (*(ptr + 4) != val);
-		ROR8(val);
+		val = shim_ror8(val);
 		bit_errors += (*(ptr + 5) != val);
-		ROR8(val);
+		val = shim_ror8(val);
 		bit_errors += (*(ptr + 6) != val);
-		ROR8(val);
+		val = shim_ror8(val);
 		bit_errors += (*(ptr + 7) != val);
 		if (UNLIKELY(!keep_stressing_flag()))
 			break;
@@ -1326,7 +1346,7 @@ static size_t TARGET_CLONES stress_vm_galpat_zero(
 
 	for (i = 0; i < bits_bad; i++) {
 		for (;;) {
-			const size_t offset = stress_mwc64() % sz;
+			const size_t offset = stress_mwc64modn(sz);
 			const uint8_t bit = stress_mwc32() & 3;
 			register uint8_t *ptr8 = (uint8_t *)buf + offset;
 
@@ -1390,7 +1410,7 @@ static size_t TARGET_CLONES stress_vm_galpat_one(
 
 	for (i = 0; i < bits_bad; i++) {
 		for (;;) {
-			const size_t offset = stress_mwc64() % sz;
+			const size_t offset = stress_mwc64modn(sz);
 			const uint8_t bit = stress_mwc32() & 3;
 			register uint8_t *ptr8 = (uint8_t *)buf + offset;
 
@@ -2115,6 +2135,7 @@ static size_t TARGET_CLONES stress_vm_rowhammer(
 	register volatile uint32_t *addr0, *addr1;
 	register size_t errors = 0;
 	register const size_t n = sz / sizeof(*addr0);
+	uint64_t mask = ~(uint64_t)(args->page_size - 1);
 
 	(void)buf_end;
 	(void)max_ops;
@@ -2131,8 +2152,8 @@ static size_t TARGET_CLONES stress_vm_rowhammer(
 		buf32[j] = val;
 
 	/* Pick two random addresses */
-	addr0 = &buf32[(stress_mwc64() << 12) % n];
-	addr1 = &buf32[(stress_mwc64() << 12) % n];
+	addr0 = &buf32[stress_mwc64modn(n) & mask];
+	addr1 = &buf32[stress_mwc64modn(n) & mask];
 
 	/* Hammer the rows */
 	for (j = VM_ROWHAMMER_LOOPS / 4; j; j--) {
@@ -2173,20 +2194,6 @@ static size_t TARGET_CLONES stress_vm_rowhammer(
 	return bit_errors;
 }
 
-/*
- *  stress_vm_popcount()
- *	population count - count number of 1 bits in val (K & R method)
- */
-static inline unsigned int OPTIMIZE3 stress_vm_popcount(register uint8_t val)
-{
-	register unsigned int count;
-
-	for (count = 0; val; count++)
-		val &= val - 1;
-
-	return count;
-}
-
 /*
  *  stress_vm_mscan()
  *	for each byte, walk through each bit set to 0, check, set to 1, check
@@ -2222,7 +2229,7 @@ static size_t TARGET_CLONES stress_vm_mscan(
 	add_counter(args, c);
 
 	for (ptr = (volatile uint8_t *)buf; ptr < end; ptr++) {
-		bit_errors += 8 - stress_vm_popcount(*ptr);
+		bit_errors += 8 - stress_vm_count_bits8(*ptr);
 	}
 
 	for (ptr = (volatile uint8_t *)buf; ptr < (uint8_t *)buf_end; ptr++, c++) {
@@ -2243,7 +2250,7 @@ static size_t TARGET_CLONES stress_vm_mscan(
 	add_counter(args, c);
 
 	for (ptr = (volatile uint8_t *)buf; ptr < end; ptr++) {
-		bit_errors += stress_vm_popcount(*ptr);
+		bit_errors += stress_vm_count_bits8(*ptr);
 	}
 
 abort:
@@ -2691,6 +2698,188 @@ abort:
 	return bit_errors;
 }
 
+/*
+ *  stress_vm_lfsr32()
+ *	sequentially set all memory to 32 bit random values
+ *	check if they are still set correctly.
+ */
+static size_t stress_vm_lfsr32(
+	void *buf,
+	void *buf_end,
+	const size_t sz,
+	const stress_args_t *args,
+	const uint64_t max_ops)
+{
+	volatile uint32_t *ptr;
+	uint64_t c = get_counter(args);
+	size_t bit_errors = 0;
+	const size_t chunk_sz = sizeof(*ptr) * 8;
+	register uint32_t lfsr = 0xf63acb01;
+
+	(void)buf_end;
+
+	for (lfsr = 0xf63acb01, ptr = (uint32_t *)buf; ptr < (uint32_t *)buf_end; ptr += chunk_sz) {
+		*(ptr + 0) = lfsr;
+		lfsr = (lfsr >> 1) ^ (unsigned int)(-(lfsr & 1u) & 0xd0000001U);
+		*(ptr + 1) = lfsr;
+		lfsr = (lfsr >> 1) ^ (unsigned int)(-(lfsr & 1u) & 0xd0000001U);
+		*(ptr + 2) = lfsr;
+		lfsr = (lfsr >> 1) ^ (unsigned int)(-(lfsr & 1u) & 0xd0000001U);
+		*(ptr + 3) = lfsr;
+		lfsr = (lfsr >> 1) ^ (unsigned int)(-(lfsr & 1u) & 0xd0000001U);
+		*(ptr + 4) = lfsr;
+		lfsr = (lfsr >> 1) ^ (unsigned int)(-(lfsr & 1u) & 0xd0000001U);
+		*(ptr + 5) = lfsr;
+		lfsr = (lfsr >> 1) ^ (unsigned int)(-(lfsr & 1u) & 0xd0000001U);
+		*(ptr + 6) = lfsr;
+		lfsr = (lfsr >> 1) ^ (unsigned int)(-(lfsr & 1u) & 0xd0000001U);
+		*(ptr + 7) = lfsr;
+		lfsr = (lfsr >> 1) ^ (unsigned int)(-(lfsr & 1u) & 0xd0000001U);
+		c++;
+		if (UNLIKELY(max_ops && (c >= max_ops)))
+			goto abort;
+		if (UNLIKELY(!keep_stressing_flag()))
+			goto abort;
+	}
+
+	(void)stress_mincore_touch_pages(buf, sz);
+	inject_random_bit_errors(buf, sz);
+
+	for (lfsr = 0xf63acb01, ptr = (uint32_t *)buf; ptr < (uint32_t *)buf_end; ptr += chunk_sz) {
+		bit_errors += stress_vm_count_bits(*(ptr + 0) ^ lfsr);
+		lfsr = (lfsr >> 1) ^ (unsigned int)(-(lfsr & 1u) & 0xd0000001U);
+		bit_errors += stress_vm_count_bits(*(ptr + 1) ^ lfsr);
+		lfsr = (lfsr >> 1) ^ (unsigned int)(-(lfsr & 1u) & 0xd0000001U);
+		bit_errors += stress_vm_count_bits(*(ptr + 2) ^ lfsr);
+		lfsr = (lfsr >> 1) ^ (unsigned int)(-(lfsr & 1u) & 0xd0000001U);
+		bit_errors += stress_vm_count_bits(*(ptr + 3) ^ lfsr);
+		lfsr = (lfsr >> 1) ^ (unsigned int)(-(lfsr & 1u) & 0xd0000001U);
+		bit_errors += stress_vm_count_bits(*(ptr + 4) ^ lfsr);
+		lfsr = (lfsr >> 1) ^ (unsigned int)(-(lfsr & 1u) & 0xd0000001U);
+		bit_errors += stress_vm_count_bits(*(ptr + 5) ^ lfsr);
+		lfsr = (lfsr >> 1) ^ (unsigned int)(-(lfsr & 1u) & 0xd0000001U);
+		bit_errors += stress_vm_count_bits(*(ptr + 6) ^ lfsr);
+		lfsr = (lfsr >> 1) ^ (unsigned int)(-(lfsr & 1u) & 0xd0000001U);
+		bit_errors += stress_vm_count_bits(*(ptr + 7) ^ lfsr);
+		lfsr = (lfsr >> 1) ^ (unsigned int)(-(lfsr & 1u) & 0xd0000001U);
+		if (UNLIKELY(!keep_stressing_flag()))
+			break;
+	}
+	stress_vm_check("lfsr32", bit_errors);
+abort:
+	set_counter(args, c);
+
+	return bit_errors;
+}
+
+#define STRESS_VM_CHECKERBOARD_SWAP(p1, p2)	\
+do {						\
+		register uint64_t tmp;		\
+						\
+		tmp = *(p1);			\
+		shim_mb();			\
+		*(p1) = *(p2);			\
+		shim_mb();			\
+		*(p2) = tmp;			\
+		shim_mb();			\
+} while (0);
+
+/*
+ *  stress_vm_checkerboard()
+ *	fill adjecent bytes with alternative bit patterns
+ */
+static size_t TARGET_CLONES stress_vm_checkerboard(
+	void *buf,
+	void *buf_end,
+	const size_t sz,
+	const stress_args_t *args,
+	const uint64_t max_ops)
+{
+	volatile uint64_t *ptr;
+	size_t bit_errors = 0;
+	uint64_t c = get_counter(args);
+
+	const uint64_t v0 = 0x5555aaaa5555aaaaULL;
+	const uint64_t v1 = 0xaaaa5555aaaa5555ULL;
+	const uint64_t v2 = 0x5a5a5a5a5a5a5a5aULL;
+	const uint64_t v3 = 0xa5a5a5a5a5a5a5a5ULL;
+	const uint64_t v4 = 0x55aa55aa55aa55aaULL;
+	const uint64_t v5 = 0xaa55aa55aa55aa55ULL;
+	const uint64_t v6 = 0x5a5a5a5a5a5a5a5aULL;
+	const uint64_t v7 = 0xa5a5a5a5a5a5a5a5ULL;
+
+	for (ptr = (uint64_t *)buf; ptr < (uint64_t *)buf_end; ptr += 8) {
+		if (UNLIKELY(!keep_stressing_flag()))
+			return 0;
+		*(ptr + 0) = v1;
+		shim_mb();
+		*(ptr + 1) = v0;
+		shim_mb();
+		*(ptr + 2) = v3;
+		shim_mb();
+		*(ptr + 3) = v2;
+		shim_mb();
+		*(ptr + 4) = v5;
+		shim_mb();
+		*(ptr + 5) = v4;
+		shim_mb();
+		*(ptr + 6) = v7;
+		shim_mb();
+		*(ptr + 7) = v6;
+		shim_mb();
+
+		*(ptr + 0) = v0;
+		shim_mb();
+		*(ptr + 1) = v1;
+		shim_mb();
+		*(ptr + 2) = v2;
+		shim_mb();
+		*(ptr + 3) = v3;
+		shim_mb();
+		*(ptr + 4) = v4;
+		shim_mb();
+		*(ptr + 5) = v5;
+		shim_mb();
+		*(ptr + 6) = v6;
+		shim_mb();
+		*(ptr + 7) = v7;
+		shim_mb();
+	}
+
+	for (ptr = (uint64_t *)buf; ptr < (uint64_t *)buf_end; ptr += 8) {
+		if (UNLIKELY(!keep_stressing_flag()))
+			break;
+
+		STRESS_VM_CHECKERBOARD_SWAP(ptr + 0, ptr + 1);
+		STRESS_VM_CHECKERBOARD_SWAP(ptr + 2, ptr + 3);
+		STRESS_VM_CHECKERBOARD_SWAP(ptr + 4, ptr + 5);
+		STRESS_VM_CHECKERBOARD_SWAP(ptr + 6, ptr + 7);
+	}
+	(void)stress_mincore_touch_pages(buf, sz);
+	inject_random_bit_errors(buf, sz);
+
+	for (ptr = (uint64_t *)buf; ptr < (uint64_t *)buf_end; ptr += 8) {
+		if (UNLIKELY(!keep_stressing_flag()))
+			return 0;
+		bit_errors += (*(ptr + 0) != v1);
+		bit_errors += (*(ptr + 1) != v0);
+		bit_errors += (*(ptr + 2) != v3);
+		bit_errors += (*(ptr + 3) != v2);
+		bit_errors += (*(ptr + 4) != v5);
+		bit_errors += (*(ptr + 5) != v4);
+		bit_errors += (*(ptr + 6) != v7);
+		bit_errors += (*(ptr + 7) != v6);
+
+		c++;
+		if (UNLIKELY(max_ops && (c >= max_ops)))
+			break;
+	}
+
+	stress_vm_check("checkerboard", bit_errors);
+	set_counter(args, c);
+
+	return bit_errors;
+}
 
 /*
  *  stress_vm_all()
@@ -2718,6 +2907,7 @@ static const stress_vm_method_info_t vm_methods[] = {
 	{ "all",		stress_vm_all },
 	{ "cache-lines",	stress_vm_cache_lines },
 	{ "cache-stripe",	stress_vm_cache_stripe },
+	{ "checkerboard",	stress_vm_checkerboard },
 	{ "flip",		stress_vm_flip },
 	{ "fwdrev",		stress_vm_fwdrev },
 	{ "galpat-0",		stress_vm_galpat_zero },
@@ -2727,6 +2917,7 @@ static const stress_vm_method_info_t vm_methods[] = {
 	{ "rowhammer",		stress_vm_rowhammer },
 	{ "incdec",		stress_vm_incdec },
 	{ "inc-nybble",		stress_vm_inc_nybble },
+	{ "lfsr32",		stress_vm_lfsr32 },
 	{ "rand-set",		stress_vm_rand_set },
 	{ "rand-sum",		stress_vm_rand_sum },
 	{ "read64",		stress_vm_read64 },
@@ -2912,7 +3103,8 @@ static int stress_vm(const stress_args_t *args)
 
 	(void)stress_get_setting("vm-method", &context.vm_method);
 
-	pr_dbg("%s: using method '%s'\n", args->name, context.vm_method->name);
+	if (args->instance == 0)
+		pr_dbg("%s: using method '%s'\n", args->name, context.vm_method->name);
 
 	for (retries = 0; (retries < 100) && keep_stressing_flag(); retries++) {
 		context.bit_error_count = (uint64_t *)
