@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2013-2021 Canonical, Ltd.
- * Copyright (C)      2022 Colin Ian King.
+ * Copyright (C) 2022-2023 Colin Ian King.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -23,9 +23,9 @@
 
 static const stress_help_t help[] = {
 	{ NULL,	"x86syscall N",		"start N workers exercising functions using syscall" },
-	{ NULL,	"x86syscall-ops N",	"stop after N syscall function calls" },
 	{ NULL,	"x86syscall-func F",	"use just syscall function F" },
-	{ NULL,	NULL,		NULL }
+	{ NULL,	"x86syscall-ops N",	"stop after N syscall function calls" },
+	{ NULL,	NULL,			NULL }
 };
 
 /*
@@ -355,9 +355,9 @@ static int stress_x86syscall(const stress_args_t *args)
 		const uint64_t c = get_counter(args);
 		const double ns = ((dt * (double)STRESS_NANOSECOND) / (double)c) - overhead_ns;
 
-		pr_inf("%s: %.2f nanoseconds per call (excluding %.2f nanoseconds test overhead)\n",
+		pr_inf("%s: %.2f nanosecs per call (excluding %.2f nanosecs test overhead)\n",
 			args->name, ns, overhead_ns);
-		stress_misc_stats_set(args->misc_stats, 0, "nanoseconds per call", ns);
+		stress_metrics_set(args, 0, "nanosecs per call", ns);
 	}
 
 	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
@@ -374,9 +374,10 @@ stressor_info_t stress_x86syscall_info = {
 };
 #else
 stressor_info_t stress_x86syscall_info = {
-	.stressor = stress_not_implemented,
+	.stressor = stress_unimplemented,
 	.class = CLASS_OS,
 	.opt_set_funcs = opt_set_funcs,
-	.help = help
+	.help = help,
+	.unimplemented_reason = "only supported on Linux x86-64 and non-PCC compilers"
 };
 #endif
